//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: spreitze $
// $Date: 1997/02/06 07:19:25 $
// $Source: /net/watson/tmp/m/RCS/new-scheme-utils.cc,v $
// $Revision: 1.2 $
//

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "scheme-utils.h"

const SCM scheme::True          = SCM_BOOL_T;
const SCM scheme::False         = SCM_BOOL_F;
const SCM scheme::eol           = SCM_EOL;
const SCM scheme::eof           = SCM_EOF_VAL;
const SCM scheme::undefined     = SCM_UNDEFINED;
const SCM scheme::unspecified   = SCM_UNSPECIFIED;
const scheme::status scheme::Ok = 0;

extern "C" {
#include "guile/gscm.h"
}

unsigned int scheme_save_stack::_count = 0;

scheme_save_stack::scheme_save_stack()
{
  if(!scm_stack_base) {
    _old_stack_base = scm_stack_base;
    scm_stack_base = &_stack_base;
  } else {
    _old_stack_base = &_stack_base;
  }
  _count++;
}

scheme_save_stack::~scheme_save_stack()
{
  if(!--_count && _old_stack_base != &_stack_base) {
    scm_stack_base = _old_stack_base;
  }
}

#ifndef NEW_GUILE

SCM scheme::make_vector(unsigned int len, SCM val)
{
  return scm_make_vector(SCM_MAKINUM(len), val);
}

SCM* scheme::make_array(int size)
{
  return scm_mkarray(size, 1);
}

void scheme::free_array(SCM* arr)
{
  scm_free_array(arr);
}

#define scm_makfrom0str makfrom0str

#else

SCM scheme::make_vector(unsigned int len, SCM val)
{
  return scm_make_vector(SCM_MAKINUM(len), val, SCM_UNDEFINED);
}

SCM* scheme::make_array(int size)
{
  SCM vector = scm_make_vector (SCM_MAKINUM (size), SCM_BOOL_F, SCM_UNDEFINED);
  scm_permanent_object (vector); /* Make the object an eternal one */
  return SCM_VELTS (vector);
}

void scheme::free_array(SCM* arr)
{
  // nothing for now.
}

static void gscm_2_str(char ** out, int * len_out, SCM * objp)
{
  SCM_ASSERT (SCM_NIMP (*objp) && SCM_STRINGP (*objp), *objp, SCM_ARG3, "gscm_2_str");
  if (out)
    *out = SCM_CHARS (*objp);
  if (len_out)
    *len_out = SCM_LENGTH (*objp);
}

static char* crap = 0;

#define gscm_run_scm(argc, argv, stdin, stdout, stderr, init_fn, blah, cmd)\
scm_boot_guile(&crap, argc, argv, stdin, stdout, stderr, (void (*)())init_fn, "(%try-load-path \"ice-9/boot-9.scm\")")

#endif

SCM scheme::seval_str(const char* str)
{
#if 0
  scheme_save_stack ss;
  {
    scm_init_signals();
    scm_errjmp_bad = 0;
    //  scm_evstr
    SCM ret = scm_eval_string(scheme::from_string(str));
    scm_errjmp_bad = 1;
    return ret;
  }
#else
  SCM answer;
  gscm_seval_str(&answer, -1, (char*)str);
  return answer;
#endif
}

SCM scheme::seval_file(const char* str)
{
#if 0
  scheme_save_stack ss;
  {
    scm_init_signals();
    scm_errjmp_bad = 0;
    //SCM ret = scm_try_load(scheme::from_string(str));
    //int ret = scm_ldprog(str);  // loads in (program-vicinity)
    int ret = scm_ldfile((char*)str);  // loads absolute path
    //SCM ret = scheme::call("try-load", scheme::from_string(str));
    //SCM ret = scheme::call("load", scheme::from_string(str));
    scm_errjmp_bad = 1;
    scm_restore_signals();
    return ret ? scheme::True : scheme::False;
  }
#else
  SCM answer;
  gscm_seval_file(&answer, -1, (char*)str);
  return answer;
#endif
}

SCM scheme::from_string(const char* a)
{
  return scm_makfrom0str((char*)a);
}

char* scheme::to_string(SCM s)
{
  char* str = 0;
  if(s != scheme::False) {
    char* tmp;
    int len;
    gscm_2_str(&tmp, &len, &s);
    str = strncpy(new char[len+1], tmp, len);
    str[len] = 0;
  }
  return str;
}

void
scheme::set_library_path(const char* str)
{
  static char buff[2048];
  sprintf(buff, "(define library-vicinity (lambda () \"%s/\"))", str);
  scheme::seval_str(buff);
}

void
scheme::define_func(SCM (*fn)(), const char* name,
		    int req, int opt, int var, char*)
{
  static char scm_name[256];
  char* p = scm_name;
  while(*name) {
    if(name[0] == '_' && name[1] == '_') {
      *p++ = ':';
      name += 2;
    } else if(name[0] == '_') {
      *p++ = '-';
      name++;
    } else {
      *p++ = *name++;
    }
  }
  *p++ = 0;

  p = strdup(scm_name);  // this is a leak but needs to be.

#ifdef __GNUG__
  scm_make_gsubr(p, req, opt, var, (SCM (*)(...))fn);
#else
  scm_make_gsubr(p, req, opt, var, (SCM (*)())fn);
#endif
}

SCM scheme::lambda(SCM (*fn)(), int req, int opt, int varp, char* doc)
{
#ifdef __GNUG__
  return scm_make_gsubr ("*anonymous*", req, opt, varp, (SCM (*)(...))fn);
#else
  return scm_make_gsubr ("*anonymous*", req, opt, varp, fn);
#endif
}

SCM scheme::symbol(const char* name)
{
  return scm_eval(SCM_CAR(scm_intern((char*)name, strlen(name))));
}

#if 0
#define scheme_apply_fn(proc, args) scm_apply(proc, args, scheme::eol)
#else
SCM scheme_apply_fn(SCM proc, SCM slist)
{
  SCM ret;
  scheme_save_stack ss;
  {
    //SCM_ALLOW_INTS;
    ret = scm_apply(proc, slist, scheme::eol);
    //SCM_DEFER_INTS;
  }
  return ret;
}
#endif

SCM scheme::call(SCM proc)
{
  return scheme_apply_fn(proc, scheme::eol);
}

SCM scheme::call(SCM proc, SCM a)
{
  SCM args = scheme::list(a, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(SCM proc, SCM a, SCM b)
{
  SCM args = scheme::list(a, b, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(SCM proc, SCM a, SCM b, SCM c)
{
  SCM args = scheme::list(a, b, c, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(SCM proc, SCM a, SCM b, SCM c, SCM d)
{
  SCM args = scheme::list(a, b, c, d, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(SCM proc, SCM a, SCM b, SCM c, SCM d, SCM e)
{
  SCM args = scheme::list(a, b, c, d, e, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(SCM proc, SCM a, SCM b, SCM c, SCM d, SCM e, SCM f)
{
  SCM args = scheme::list(a, b, c, d, e, f, scheme::eol);
  SCM ret = scheme_apply_fn(proc, args);
  return ret;
}

SCM scheme::call(const char* pname)
{
  return scheme::call(scheme::symbol(pname));
}

SCM scheme::call(const char* pname, SCM a)
{
  return scheme::call(scheme::symbol(pname), a);
}

SCM scheme::call(const char* pname, SCM a, SCM b)
{
  return scheme::call(scheme::symbol(pname), a, b);
}

SCM scheme::call(const char* pname, SCM a, SCM b, SCM c)
{
  return scheme::call(scheme::symbol(pname), a, b, c);
}

SCM scheme::call(const char* pname, SCM a, SCM b, SCM c, SCM d)
{
  return scheme::call(scheme::symbol(pname), a, b, c, d);
}

SCM scheme::call(const char* pname, SCM a, SCM b, SCM c, SCM d, SCM e)
{
  return scheme::call(scheme::symbol(pname), a, b, c, d, e);
}

SCM scheme::call(const char* pname, SCM a, SCM b, SCM c, SCM d, SCM e, SCM f)
{
  return scheme::call(scheme::symbol(pname), a, b, c, d, e, f);
}

#define CURRY_PROC(cclo) (VELTS(cclo)[1])
#define CURRY_ARG1(cclo) (VELTS(cclo)[2])

static SCM curry_apply(SCM self, SCM rest)
{
  return scheme::call(CURRY_PROC(self),
		      scheme::cons(CURRY_ARG1(self), rest),
		      scheme::eol);
}

// probably need to save stack around this guy!
SCM scheme::curry(SCM fn, SCM arg)
{
  SCM curry_apply_fn = scheme::lambda((SCM(*)())curry_apply, 0, 0, 1, 0);
  SCM answer = scm_makcclo(curry_apply_fn, 3L);
  CURRY_ARG1(answer) = arg;
  CURRY_PROC(answer) = fn;
  return answer;
}

SCM scheme::vector(SCM elt, ...)
{
  va_list foo;
  unsigned int len = 0;

  va_start(foo, elt);
  while(elt != scheme::eol) {
    elt = va_arg(foo, SCM);
    len++;
  }

  SCM vec = scheme::make_vector(len);

  va_start (foo, elt);
  for(unsigned int i = 0; i < len; i++) {
    scheme::vset(vec, i, elt);
    elt = va_arg(foo, SCM);
  }

  va_end(foo);
  return vec;
}

SCM scheme::list(SCM elt, ...)
{
  SCM answer = scheme::eol;
  SCM* pos = &answer;
  va_list foo;

  va_start (foo, elt);
  while (elt != scheme::eol) {
    *pos = scheme::cons(elt, scheme::eol);
    pos = &CDR(*pos);
    elt = va_arg (foo, SCM);
  }

  va_end(foo);
  return answer;
}

SCM scheme::lset(SCM l, unsigned int i, SCM val)
{
  SCM tmp = l;

  while(i--)
    tmp = scheme::cdr(tmp);

  scheme::set_car(tmp, val);

  return val;
}

#ifdef __GNUG__
static scheme::status init_fn(...)
{
  return scheme::Ok;
}
#else
static scheme::status init_fn()
{
  return scheme::Ok;
}
#endif

scheme::status scheme::init(int argc,
			    char* argv[],
			    void (*init)(),
			    char* scm_file,
			    char* scm_main,
			    int scm_argc)
{
  static char buff[2048];
  scheme::status s;
  SCM answer;

  //gscm_set_init_heap_size(1024*1024);

  if(scm_argc)
    s = gscm_run_scm(argc, argv, stdin, stdout, stderr, init_fn, 0, "#t");
  else
    s = gscm_run_scm(1, argv, stdin, stdout, stderr, init_fn, 0, "#t");
  assert(s == scheme::Ok);

  // can be used to initialize any C++/scheme bindings
  if(init)
    init();

  if(scm_file) {
    scheme::seval_file(scm_file);
  }

  if(scm_main) {
    if(!strcmp(scm_main, "stand-alone-repl")) {
      scheme::seval_str("(stand-alone-repl)");
    } else {
      sprintf(buff, "(%s %d (list", scm_main, argc);
      for(int i = 0; i < argc; i++)
	sprintf(buff + strlen(buff), " \"%s\"", argv[i]);
      sprintf(buff + strlen(buff), "))");
      scheme::seval_str(buff);
    }
  }

  return scheme::Ok;
}
