//  -*- Mode: C++; -*-

// Copyright (c) 1993 Siemens Corporate Research, Inc.
// All rights reserved.
//
// $Author: bnell $
// $Date: 1996/12/18 15:46:15 $
// $Source: /home/yugo/kwuftn/CVS-repository/scheme-interface/scheme-types.h,v $
// $Revision: 1.6 $
//

#ifndef _SCHEME_TYPES_H_
#define _SCHEME_TYPES_H_

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "scheme-utils.h"

// do something about function ptrs?
// need to make tmp versions of some (all?) objects!
// make vectors so they can spit out correctly formated data
//   (e.g. not vectors of scheme_strings which have virtual funcs, etc.)

/////////////////////////////////////////////////////////////////////////

// class decls

class scheme_string;
template <class T> class scheme_value;
template <class T> class scheme_ptr;
template <class T> class scheme_vec;
template <class T> class scheme_vec0;

/////////////////////////////////////////////////////////////////////////

// base type convert functions

extern char scheme_convert(SCM s, char*);
extern unsigned char scheme_convert(SCM s, unsigned char*);
extern short scheme_convert(SCM s, short*);
extern unsigned short scheme_convert(SCM s, unsigned short*);
extern int scheme_convert(SCM s, int*);
extern unsigned int scheme_convert(SCM s, unsigned int*);
extern long scheme_convert(SCM s, long*);
extern unsigned long scheme_convert(SCM s, unsigned long*);
extern float scheme_convert(SCM s, float*);
extern double scheme_convert(SCM s, double*);
extern long double scheme_convert(SCM s, long double*);
extern char* scheme_convert(SCM s, char**);
extern scheme_string scheme_convert(SCM, scheme_string*);

template <class T> scheme_value<T> scheme_convert(SCM, scheme_value<T>*);
template <class T> scheme_ptr<T> scheme_convert(SCM, scheme_ptr<T>*);
template <class T> scheme_vec<T> scheme_convert(SCM, scheme_vec<T>*);
template <class T> scheme_vec0<T> scheme_convert(SCM, scheme_vec0<T>*);

/////////////////////////////////////////////////////////////////////////

// base class for all scm -> c++ type conversion classes

class scheme_base {
 public:
  operator SCM() { return to_scheme(); } 
  virtual SCM to_scheme() = 0;
//  virtual void* make_array(unsigned int) = 0;
  virtual ~scheme_base() { }
};

/////////////////////////////////////////////////////////////////////////

// converter back to scheme types
class scmify {
 public:
  operator SCM() { return _val; }
  scmify() : _val(scheme::False) { }
  scmify(char s) : _val(scheme::from_char(s)) { }
  scmify(unsigned char s) : _val(scheme::from_char(s)) { }
  scmify(int s) : _val(scheme::from_long(s)) { }
  scmify(unsigned int s) : _val(scheme::from_ulong(s)) { }
  scmify(short s) : _val(scheme::from_long(s)) { }
  scmify(unsigned short s) : _val(scheme::from_ulong(s)) { }
  scmify(long s) : _val(scheme::from_long(s)) { }
  scmify(unsigned long s) : _val(scheme::from_ulong(s)) { }
  scmify(float s) : _val(scheme::from_double(s)) { }
  scmify(double s) : _val(scheme::from_double(s)) { }
  scmify(long double s) : _val(scheme::from_double(s)) { }
  scmify(char* s) : _val(scheme::from_string(s)) { }
//  scmify(void* s) : _val(scheme::from_ulong((unsigned long)s)) { }
//  scmify(wchar_t* s) : _val(scheme::from_ulong((unsigned long)s)) { }
  scmify(scheme_base& b) : _val(b.to_scheme()) { }
 private:
  SCM _val;
};

/////////////////////////////////////////////////////////////////////////

// type converter classes and functions

template <class T> class schemet : public scheme_base {
 public:
  T& operator=(const T& val) { return _val = val; }
  T* operator&() { return &_val; }
  operator T() { return _val; }
  operator const T() { return _val; }
  T value() const { return _val; }
  T* valptr() { return &_val; }
  void set_value(const T& val) { _val = val; }
 protected:
  schemet(const T& val) : _val(val) { }
  schemet() { }
 private:
  T _val;
};

template <class T> class scheme_value : public schemet<T> {
 public:
  SCM to_scheme() { return scmify(value()); }
  scheme_value() : schemet<T>((T)0) { }
  scheme_value(const T& v) : schemet<T>(v) { }
  scheme_value(SCM s) : schemet<T>(scheme_convert(s, valptr())) { }
};

// add ptr own?
template <class T> class scheme_ptr : public schemet<T*> {
 public:
#ifdef __GNUG__
  operator bool() { return value() ? true : false; }
#else
  operator int() { return value() ? 1 : 0; }
#endif
  T* operator->() { return value(); }
// is this correct?
//  void operator delete(void*, size_t) { delete value(); }
// ???
//  operator void*() { return value(); }
  SCM to_scheme() { return value() ? scheme::from_ulong((unsigned int)value()) : scheme::False; }
  scheme_ptr() : schemet<T*>((T*)0) { }
  scheme_ptr(T* p) : schemet<T*>(p) { }
  scheme_ptr(SCM s)
  : schemet<T*>((T*)0)
  {
    if(s != scheme::False)
      set_value((T*)scheme::to_ulong(s));
  }
};

#include "scheme_vec.h"
#include "scheme_vec0.h"

class scheme_string_base : public schemet<char*> {
 public:
  operator char*() { return value(); }
  operator const char*() { return value(); }
  SCM to_scheme() { return value() ? scheme::from_string(value()) : scheme::False; }
  scheme_string_base() : schemet<char*>((char*)0) { }
  scheme_string_base(const char* s) : schemet<char*>((char*)s) { }
  scheme_string_base(SCM s) : schemet<char*>((char*)0)
  {
    set_value(scheme::to_string(s));
  }
};

class scheme_string : public scheme_string_base {
 public:
  scheme_string() { }
  scheme_string(const char* s) : scheme_string_base((char*)s) { }
  scheme_string(SCM s) : scheme_string_base(s) { }
  scheme_string(const scheme_string& s)
    : scheme_string_base(strcpy(new char[strlen(s.value())+1], s.value())) { }
  ~scheme_string() { delete[] value(); }
};

class scheme_tmp_string : public scheme_string_base {
 public:
  scheme_tmp_string() { }
  scheme_tmp_string(const char* s) : scheme_string_base((char*)s) { }
  scheme_tmp_string(const scheme_string& s)
    : scheme_string_base(strcpy(new char[strlen(s.value())+1], s.value())) { }
  scheme_tmp_string(const scheme_tmp_string& s)
    : scheme_string_base(strcpy(new char[strlen(s.value())+1], s.value())) { }
  scheme_tmp_string(SCM s) : scheme_string_base(s) { }
};

/////////////////////////////////////////////////////////////////////////

// type convertion functions for scheme_base classes

template <class T>
inline scheme_value<T> scheme_convert(SCM s, scheme_value<T>*)
{
  return scheme_value<T>(s);
}

template <class T>
inline scheme_ptr<T> scheme_convert(SCM s, scheme_ptr<T>*)
{
  return scheme_ptr<T>(s);
}

// problems with deletions here!!!
template <class T>
inline scheme_vec<T> scheme_convert(SCM s, scheme_vec<T>*)
{
  return scheme_vec<T>(s);
}

template <class T>
inline scheme_vec0<T> scheme_convert(SCM s, scheme_vec0<T>*)
{
  return scheme_vec0<T>(s);
}

//////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
//
//   // to add new type: scheme_obj
//   class scheme_obj : public schemet<obj> {
//    public:
//     SCM to_scheme() { convert obj back into SCM representation }
//     scheme_obj() { set value of object to some default value }
//     scheme_obj(SCM) : scheme<obj>(convert_scm_to_obj_type()) { }
//     ~scheme_obj() { any destructor business }
//   };
//
//   // write a function (beware of multiple deletes!)
//   scheme_obj scheme_convert(SCM s, scheme_obj*)
//   {
//     return scheme_obj(s);
//   }
// 
/////////////////////////////////////////////////////////////////////////

#endif
