//  -*- Mode: C++; -*-

// Copyright (c) 1993 Siemens Corporate Research, Inc.
// All rights reserved.
//
// $Author: bnell $
// $Date: 1996/12/18 15:46:17 $
// $Source: /home/yugo/kwuftn/CVS-repository/scheme-interface/scheme-utils.h,v $
// $Revision: 1.6 $
//

#ifndef _SCHEME_UTILS_H_
#define _SCHEME_UTILS_H_

#include <stdio.h>
#include <stdlib.h>

//#define NEW_GUILE

#if defined(__CLCC__) && defined(_HPUX_SOURCE)
#define __STDC__
#endif

extern "C" {
#ifdef NEW_GUILE
#include "libguile.h"
#else
#include "guile/_scm.h"
#endif
#ifndef scm_stack_base
extern SCM_STACKITEM* scm_stack_base;
#endif
}

class scheme_save_stack {
 public:
  scheme_save_stack();
  ~scheme_save_stack();
 private:
  SCM_STACKITEM* _old_stack_base;
  SCM_STACKITEM _stack_base;
  static unsigned int _count;
};

typedef SCM (*SCMfn)();

#define scheme_define_func(func, req, op, var) \
scheme::define_func((SCMfn)func, #func, req, op, var, 0)

class scheme {
 public:
  typedef int status;

  // scheme constants
  static const SCM True;
  static const SCM False;
  static const SCM eol;
  static const SCM eof;
  static const SCM undefined;
  static const SCM unspecified;
  static const status Ok;

  // define C<->scheme function
  static void define_func(SCM (*func)(), const char*, int, int, int, char* =0);
  static SCM lambda(SCM (*func)(), int, int, int, char* =0);

  // lookup symbol value
  static SCM symbol(const char* name);

  // call scheme procs
  static SCM call(SCM proc);
  static SCM call(SCM proc, SCM a);
  static SCM call(SCM proc, SCM a, SCM b);
  static SCM call(SCM proc, SCM a, SCM b, SCM c);
  static SCM call(SCM proc, SCM a, SCM b, SCM c, SCM d);
  static SCM call(SCM proc, SCM a, SCM b, SCM c, SCM d, SCM e);
  static SCM call(SCM proc, SCM a, SCM b, SCM c, SCM d, SCM e, SCM f);
  static SCM call(const char* pname);
  static SCM call(const char* pname, SCM a);
  static SCM call(const char* pname, SCM a, SCM b);
  static SCM call(const char* pname, SCM a, SCM b, SCM c);
  static SCM call(const char* pname, SCM a, SCM b, SCM c, SCM d);
  static SCM call(const char* pname, SCM a, SCM b, SCM c, SCM d, SCM e);
  static SCM call(const char* pname, SCM a, SCM b, SCM c, SCM d, SCM e, SCM f);

  static SCM curry(SCM fn, SCM arg);

  // basic conversions
  static SCM from_bool(int b)            { return b ? True : False; }
  static SCM from_char(char a)           { return SCM_MAKICHR(a); }
  static SCM from_long(long a)           { return scm_long2num(a); }
  static SCM from_ulong(unsigned long a) { return scm_ulong2num(a); }
  static SCM from_ptr(void* p)           { return from_ulong((unsigned long)p); }
  static SCM from_double(double a)       { return scm_makdbl(a, 0.0); }
  static SCM from_string(const char* a);

  static int  to_bool(SCM b)   { return b == True ? 1 : 0; }
  static char to_char(SCM a)   { return SCM_ICHR(a); }
  static long to_long(SCM a)
    { return scm_num2long(a, (char*)SCM_ARG1, "scheme::to_long"); }
  static unsigned long to_ulong(SCM a)
    { return scm_num2ulong((a), (char*)SCM_ARG1, "scheme::to_ulong"); }
  static void* to_ptr(SCM a)   { return (void*)from_ulong(a); }
  static double to_double(SCM a) { return scm_num2dbl(a,"scheme::to_double"); }
  static char* to_string(SCM a);

  static SCM string_p(SCM l) { return scm_string_p(l); }

  // define persistent scheme data
  static SCM* make_array(int size);
  static void free_array(SCM*);

  // common functions
  static unsigned int obj_length(SCM o) { return scm_obj_length(o); }

  // vector functions
  static SCM vector_p(SCM v) { return scm_vector_p(v); }
  static SCM make_vector(unsigned int len, SCM init = scheme::False);
  static SCM vector(SCM a, ...);
  static SCM vset(SCM vec, unsigned int i, SCM val)
    { return scm_vector_set_x(vec, SCM_MAKINUM(i), val); }
  static SCM vref(SCM vec, unsigned int i)
    { return scm_vector_ref(vec, SCM_MAKINUM(i)); }

  // list functions
  static SCM list_p(SCM l) { return scm_pair_p(l); }
  static SCM list(SCM a, ...);
  static SCM lref(SCM l, unsigned int i)
    { return scm_list_ref(l, SCM_MAKINUM(i)); }
  static SCM lset(SCM l, unsigned int i, SCM val);

  static SCM car(SCM s)
    { return ((SCM_NIMP(s) && SCM_CONSP(s))
	      ? SCM_CAR(s)
	      : scm_wta (s, (char*)SCM_ARG1, "scheme::car")); }

  static SCM set_car(SCM s, SCM val)
    { return ((SCM_NIMP(s) && SCM_CONSP(s))
	      ? (SCM_CAR(s) = val)
	      : scm_wta (s, (char *)SCM_ARG1, "scheme::set_car!")); }

  static SCM cdr(SCM s)
    { return ((SCM_NIMP(s) && SCM_CONSP(s))
	      ? SCM_CDR(s)
	      : scm_wta (s, (char*)SCM_ARG1, "scheme::cdr")); }

  static SCM set_cdr(SCM s, SCM val)
    { return ((SCM_NIMP(s) && SCM_CONSP(s))
	      ? (SCM_CDR(s) = val)
	      : scm_wta (s, (char *)SCM_ARG1, "scheme::set_cdr!")); }

  static SCM cons(SCM a, SCM b) { return scm_cons(a, b); }

  // p must be a scheme port
  static FILE* stream(SCM p) { return SCM_STREAM(p); }

  static void set_library_path(const char* str);

  // eval strings or files
  static SCM seval_str(const char*);
  static SCM seval_file(const char*);

  // call to initiailize scheme and possibly pass control to it
  // if scm_file and/or scm_main are passed.
  static status init(int argc,
		     char* argv[],
		     void (*init)() = 0,
		     char* scm_file = 0,
		     char* scm_main = 0,
		     int scm_argv = 0);

  // don't let anyone create a scheme object!
 protected:
  scheme() { }
  ~scheme() { }
};

#endif
