/* Ilu.java */
/* Chris Jacobi, January 22, 1997 2:00 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: Ilu.java,v 1.10 1997/01/22 22:19:28 jacobi Exp $ */

 
package xerox.ilu;
import xerox.ilu.IluClassRep;
import xerox.ilu.IluInit;
import xerox.ilu.IluObject0;
import xerox.ilu.IluObject;
import xerox.ilu.IluSystemException;
import xerox.ilu.IluRT0;
import xerox.ilu.IluServer;

/**
 * Main application interface for ILU.
 */
public final class Ilu {
        
    /**
     * Don't create instances 
     */
    private Ilu(){};
    
    /** 
     * Importing an object 
     */
    public static IluObject 
    objectFromSBH (
            String sbh, 
            IluClassRep iluClass
            ) throws IluSystemException {
        return IluRT0.objectFromSBH(sbh, iluClass);
    }

    /**
     * Create true objects.
     * OBSOLETE.
     */
    public static void 
    registerTrueObject(
            String ih,
            IluObject0 tobj, 
            IluServer server, 
            IluClassRep iluClass
            ) throws IluSystemException {
        registerTrueObject(ih, tobj, server, iluClass, remember, null);
    }

    /** 
     * Literals for the lifetime argument of registerTrueObject.
     */
    
    /** Requires ilu set a value depending on other arguments */
    public static final int unspec = IluRT0.unspec;
    /** Remember obj to prevent java garbage collection. */
    public static final int remember = IluRT0.remember;
    /** Forget obj to subject it to java garbage collection. */
    public static final int forget = IluRT0.forget;

    /**
     * Create true objects. <p>
     * Life time is important and must be determined.  Normally the ilu
     * object stays alive until the intermediate object is garbage
     * collected.  Use any of a large number of schemes to prevent
     * garbage collection. <p>
     * 
     * Return value is intermediate object.  Caller may prevent this
     * from being garbage collected. <p>
     *
     * @see IluAnchorSupport.
     */
    public static Object 
    registerTrueObject(
            String ih,
            IluObject0 tobj, 
            IluServer server, 
            IluClassRep iluClass,
            int lifetime, //0 gives default behaviour
            xerox.ilu.IluAnchorSupport extraAnchor //null is ok
            ) throws IluSystemException {
        return IluRT0.registerTrueObject(ih, tobj, server, iluClass, lifetime, extraAnchor);
    }
    
    
    /** 
     * Returns the string binding handle.
     * May return NIL if the object's server isn't exported through any
     * port; may return an invalid SBH if the cached one references a
     * closed port.
     */
    public static String 
    sbhOfObject(IluObject0 obj) {
        return IluRT0.sbhOfObject(obj);
    }
  
    /** 
     * Destroy ILU-ness of true object.
     * Only for impatient applications,
     * Normal applications leave this to the garbage collector.
     * Do NOT use from applets; error behavior is not yet defined.
     * Do not use on surrogates.
     */
    public static void 
    destroyObject(IluObject0 obj) {
        IluRT0.destroyObject(obj);
    }
    
    
    /**
     * Generates a string that's unique over space and time.
     * A server with nothing better to use might call this to 
     * get an ID.
     */
    public static String 
    inventID() {
        return IluRT0.inventID();
    }
    

    /**
     * Makes sure Ilu is loaded.
     */
    public static void init(){};
    
    static {
       IluInit.init();
    }
  
} // Ilu

