/* IluCall.java */
/* Chris Jacobi, January 31, 1997 11:04 am PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

/* $Id: IluCall.java,v 1.38 1997/01/31 19:04:40 jacobi Exp $ */
/* See native code in IluJava_IluCall.c */ 

package xerox.ilu;

/**
 * Data structures and operations to perform ilu calls.<p>
 * Class is used from stubs only; clients have no need
 * to know this class.<p>
 *
 * The types as part of procedure names refer to ILU-types; not 
 * Java types.<br>
 * 	The parameters of course are the Java representation for  
 * 	the named ILU-type.
 */
 
public final
class IluCall {

    /* Pointer to the real information used by the native side only */
    private long yCall;
    private long yPtr;
    private long yCallSurrConn;
    /*friendly*/ xerox.ilu.IluObject jObj;
    /*friendly*/ xerox.ilu.IluOInt jOIntDisc;//keep em trace-able
    /*friendly*/ Object jObjTracable;//make it trace-able
    /*friendly*/ xerox.ilu.IluMethodRep jMethodRep;	//native read+write!!
    /*friendly*/ xerox.ilu.IluClassRep jClassRep;	//native read+write!!
    /*friendly*/ boolean jMustFork = false;	//native write
    private boolean jNeedsSizing = false;      	//native write
    /*friendly*/ int needFinish = 0;	//native read+write
    private int selfCheck = 0;	//native read+write
    
    /* 
     * Used by the native side to find the class.
     * This is necessary because it limits the navel show 
     * required otherwise. 
     */
    private native void givePrototype();
    static {
        (new IluCall()).givePrototype();
    }
    
    static void someException(String s) throws IluSystemException {
        IluDebug.clientPreError(s);
        throw new IluSomeSystemException(s);
    }
    
    static void badChar() throws IluSystemException {
        IluDebug.clientPreError("bad character");
        throw new xerox.ilu.IluCharacterRangeException();
    }
    
    static void badLength() throws IluSystemException {
        IluDebug.clientPreError("bad length");
        throw new xerox.ilu.IluBadParameterException("bad length");
    }
    
    /* 
     * Fake constructor; in reality we are constructing IluCall's on 
     * the native side (oops: only surrogate calls...)
     */
    private IluCall () {
    }
    
    /* Real finalizer */
    private final native void nativeFinalize() throws Throwable;
    
    /** Finalizer to give the native side a chance for cleanup */
    protected void finalize() throws Throwable {
        nativeFinalize(); 
    }
    
    
    /**
     * Stub visible "constructor" for SURROGATE calls
     * Either allocates or reuses a call.<br>
     */
    public static IluCall startCall(
    	xerox.ilu.IluMethodRep methodRep,
    	xerox.ilu.IluSurrogateObject obj, 
        xerox.ilu.IluPassport passport
    		) throws IluSystemException
    {
        IluCall call = null;
        if (IluDebug.callsFlag > 0) {
            IluDebug.log.println("! startCall " + obj + " " + methodRep);
        }
        try {
            call = nativeStartCall(methodRep, obj, passport);
        } finally {
            if (call.jMustFork) {
                IluSurrogateConnection.forkSC(call.yCallSurrConn);
            }
        }
        return call;
    } //startCall
    
    	
    private static native IluCall nativeStartCall(
    	xerox.ilu.IluMethodRep methodRep,
    	xerox.ilu.IluSurrogateObject obj, 
        xerox.ilu.IluPassport passport
    	) throws IluSystemException;
	
    /**
     * native side for makeCallShell.
     */
    private native void initCallShell() throws IluSystemException;
    	
    /**
     * Gives a new reusable IluCall.
     * This is aimed for reading server (true) connections; 
     */
    public static IluCall makeCallShell() throws IluSystemException {
       IluCall cl = new IluCall();
       cl.initCallShell();
       return cl;
    } //makeCallShell
    
    public native void finishCall();
    
    /**
     * Marshalling routines.
     * All name of types are as defined by ilu.
     * Arguments are as used in Java.
     *
     * The IluClassRep parameters are as specified in the class which did 
     * define the method; the actual object passed can be from a 
     * further subclass.
     */
    
    public native int  
    szObjectx(IluObject obj, 
        boolean discriminant, IluClassRep staticClass
        ) throws IluSystemException;
    
    public int  
    szObject(IluObject obj, 
        boolean discriminant, IluClassRep staticClass
        ) throws IluSystemException {
        return szObjectx((IluObject)obj, discriminant, staticClass);
    }
    
    public native void 
    outObjectx(IluOInt oi, 
        boolean discriminant, IluClassRep staticClass
        ) throws IluSystemException;
        
    public void 
    outObject(IluObject0 obj, 
        boolean discriminant, IluClassRep staticClass
        ) throws IluSystemException 
    {
        IluOInt oi = null;
        if (IluDebug.callsFlag > IluDebug.basic) {
            IluDebug.log.println("! outObject " + obj);
        }
        if (obj != null) {
            oi = IluOInt.nPeekWithCall((IluObject) obj, this);
            if (oi==null) {
                //object unknown; try to register it right here...
                IluClassRep iluClass = IluClassAccess.iluClassFromJavaClass(obj.getClass());
                if (iluClass != null & staticClass.isA(iluClass)) {
                    IluRT0.registerTrueObject(
                        null,	//ih
                        obj, 
                        null,	//server
                        iluClass,
                        0,	//lifetime
                        null	//extraAnchor
                        );
                }
                oi = IluOInt.nPeekWithCall((IluObject) obj, this);
                if (oi==null) {
                    someException("object not registerd with Ilu " + obj);
                }
            }
        }
        outObjectx(oi, discriminant, staticClass);
    } //outObject
    
    public IluObject inObject(
        	boolean discriminant, IluClassRep staticClass
        	) throws IluSystemException {
        IluObject realObject = null;
        IluOInt oi;
        //I'd love to have this outside any locks !?
        //Compare to IluRT0
        if (IluDebug.callsFlag > IluDebug.basic) {
            IluDebug.log.println("! IluCall_inObject: enter");
        }
        oi = inOI(discriminant, staticClass);
        if (oi==null) {
            if (!discriminant && staticClass.isOptional()) {
                if (IluDebug.traceCalls() > IluDebug.dont) {
                    IluDebug.log.println("! IluCall_inObject: null");
                }
                return null;
            }
            someException("inObject failed (null)");
        }
        realObject = oi.getRealObject();
        if (realObject==null) {
            oi.createSurrogateForOI();
            realObject = oi.getRealObject();
            if (realObject==null) {
                someException("inObject failed (local)");
            }
        }
        if (IluDebug.traceCalls() > IluDebug.dont) {
            IluDebug.log.println("! IluCall_inObject got: " + realObject);
        }
        return realObject;
    } //inObject
    
    
    /* may be "un-finished" */
    private native IluOInt 
    inOI(boolean dis, IluClassRep staticClass) throws IluSystemException;
    
    public IluObject getCallSingleton() throws IluSystemException {
        IluObject realObject = null;
        if (IluDebug.traceCalls() > IluDebug.basic) {
            IluDebug.log.println("! IluCall_getCallSingleton: enter");
        }
        IluOInt oi = getCallSingletonOIx();
        if (oi != null) {
            realObject = oi.getRealObject();
        }
        if (realObject==null) {
            someException("getCallSingleton failed");
        }
        if (IluDebug.traceCalls() > IluDebug.dont) {
            IluDebug.log.println("! IluCall_getCallSingleton got: " + realObject);
        }
        return realObject;
    } //getCallSingleton
    
    private native IluOInt 
    getCallSingletonOIx() throws IluSystemException;
    
    public native int  szInt64(long val) throws IluSystemException;
    public native void outInt64(long val) throws IluSystemException;
    public native long inInt64() throws IluSystemException;
    
    public native int  szInt32(int val) throws IluSystemException;
    public native void outInt32(int val) throws IluSystemException;
    public native int  inInt32() throws IluSystemException;
    
    public native int  szInt16(short val) throws IluSystemException;
    public native void outInt16(short val) throws IluSystemException;
    public native short inInt16() throws IluSystemException;
    
    //bytes preserve the bit pattern, not the numeric value 
    public native int  szByte(byte val) throws IluSystemException;
    public native void outByte(byte val) throws IluSystemException; 
    public native byte inByte() throws IluSystemException; 
    
    public native int  szBool(boolean val) throws IluSystemException;
    public native void outBool(boolean val) throws IluSystemException;
    public native boolean inBool() throws IluSystemException;
    
    public native int  szChar16(char val) throws IluSystemException;
    public native void outChar16x(char val) throws IluSystemException;
    public void outChar16(char val) throws IluSystemException {
        if (val=='\u0000') badChar();
        outChar16x(val);
    }
    public native char inChar16() throws IluSystemException;
    
    public native int  szChar8(char val) throws IluSystemException;
    public native void outChar8x(char val) throws IluSystemException;
    public void outChar8(char val) throws IluSystemException {
        if (val == '\u0000' || val > '\u00ff' ) badChar();
        outChar8x(val);
    }
    public native char inChar8() throws IluSystemException;

    public native int  szOptional(boolean stat) throws IluSystemException;
    public native void outOptional(boolean stat) throws IluSystemException;
    public native boolean inOptional() throws IluSystemException;
    
    public native int  szArray(int elementCount) throws IluSystemException;
    public native void outArray(int elementCount) throws IluSystemException;
    public native void inArray() throws IluSystemException;
    public native void endArray() throws IluSystemException;
    
    public native int  szRecord() throws IluSystemException;
    public native void outRecord() throws IluSystemException;
    public native void inRecord() throws IluSystemException;
    public native void endRecord() throws IluSystemException;
    
    public native int  
    	szSequence(int leng, int limit) throws IluSystemException;
    public native void 
    	outSequence(int leng, int limit) throws IluSystemException;
    public native void 
    	outSequenceMark(int extent) throws IluSystemException;
    public native int  inSequence(int limit) throws IluSystemException;
    public native void inSequenceMark(int extent) throws IluSystemException;
    public native void endSequence() throws IluSystemException;
    
    public native int  
    	szUnion(int discrim, int discrimSize) throws IluSystemException;
    public native void 
    	outUnion(int discrim, int discrimSize) throws IluSystemException;
    public native int  
    	inUnion(int discrimSize) throws IluSystemException; 
    	//returns discriminator...
    public native void 
    	endUnion() throws IluSystemException;
    
    private native int  szReal128Buff(byte[] val) throws IluSystemException;
    private native void outReal128Buff(byte[] val) throws IluSystemException;
    private native void inReal128Buff(byte[] buf) throws IluSystemException;
    public int  szReal128(xerox.ilu.types.float128 val) throws IluSystemException {
        return szReal128Buff(val.toByte16());
    }
    public void outReal128(xerox.ilu.types.float128 val) throws IluSystemException {
        outReal128Buff(val.toByte16());
    }
    public xerox.ilu.types.float128 inReal128() throws IluSystemException {
        byte[] b16 = new byte[16];
        inReal128Buff(b16);
        return xerox.ilu.types.float128.fromByte16(b16);
    }
    
    public native int  szReal64(double val) throws IluSystemException;
    public native void outReal64(double val) throws IluSystemException;
    public native double inReal64() throws IluSystemException;
    
    public native int   szReal32(float val) throws IluSystemException;
    public native void  outReal32(float val) throws IluSystemException;
    public native float inReal32() throws IluSystemException;
    
    public native int  szEnum(int val) throws IluSystemException;
    public native void outEnum(int val) throws IluSystemException;
    public native int  inEnum() throws IluSystemException;
    
    public native int  szCard16(short val) throws IluSystemException;
    public native void outCard16(short val) throws IluSystemException;
    public native short  inCard16() throws IluSystemException;
    
    /* preserve bit pattern not value */
    public native int  szCard32(int val) throws IluSystemException;
    public native void outCard32(int val) throws IluSystemException;
    public native int  inCard32() throws IluSystemException;
    
    /* preserve bit pattern not value */
    public native int  szCard64(long val) throws IluSystemException;
    public native void outCard64(long val) throws IluSystemException;
    public native long inCard64() throws IluSystemException;
    
    /* Variable-length array of short character (String) */
    public native int  
    	szString8(String s, int limit) throws IluSystemException;
    public native void 
    	outString8x(String s, int limit) throws IluSystemException;
    
    public void 
        outString8(String s, int limit) throws IluSystemException {
        int n = s.length();
        if ((limit < n) && (limit != 0)) badLength();
        for (int i = 0; i<n; i++) {
            char c = s.charAt(i);
            if ( c == '\u0000' || c > '\u00ff' ) badChar();
        }
        outString8x(s, limit);
    } //outString8
    
    public native String 
    	inString8(int limit) throws IluSystemException;
    
    /* Fix-length array of short character (StringVec) */
    public native int  
    	szChar8Array_old(String s, int limit) throws IluSystemException;
    public native void 
    	outChar8Array_old(String s, int limit) throws IluSystemException;
    public native String 
    	inChar8Array_old(int limit) throws IluSystemException;
    public native int  
    	szChar8Array(char[] a, int limit) throws IluSystemException;
    public native void 
    	outChar8Arrayx(char[] a, int limit) throws IluSystemException;
    public void 
    	outChar8Array(char[] a, int limit) throws IluSystemException {
        if (limit != a.length) badLength();
        for (int i = 0; i<limit; i++) {
            if ( a[i] == '\u0000' || a[i] > '\u00ff' ) badChar();
        }
        outChar8Arrayx(a, limit);
    }
    public native char[] 
    	inChar8Array(int limit) throws IluSystemException;

    /* Variable-length array of character (WString) */
    public native int  
    	szString16(String s, int limit) throws IluSystemException;
    public native void 
    	outString16x(String s, int limit) throws IluSystemException;
    
    public void 
    	outString16(String s, int limit) throws IluSystemException {
        int n = s.length();
        if ((limit < n) && (limit != 0)) badLength();
        for (int i = 0; i<n; i++) {
            if ( s.charAt(i) == '\u0000' ) badChar();
        }
        outString16x(s, limit);
    } //outString16
    
    public native String 
    	inString16(int limit) throws IluSystemException;

    /* Fix-length array of character (WStringVec) */
    public native int  
    	szChar16Array_old(String s, int limit) throws IluSystemException;
    public native void 
    	outChar16Array_old(String s, int limit) throws IluSystemException;
    public native String 
    	inChar16Array_old(int limit) throws IluSystemException;
    public native int  
    	szChar16Array(char[] a, int limit) throws IluSystemException;
    public native void 
    	outChar16Arrayx(char[] a, int limit) throws IluSystemException;
    
    public void 
    	outChar16Array(char[] a, int limit) throws IluSystemException {
        if (limit != a.length) badLength();
        for (int i = 0; i<limit; i++) {
            if ( a[i] == '\u0000' ) badChar();
        }
        outChar16Arrayx(a, limit);
    } //outChar16Array
    
    public native char[] 
    	inChar16Array(int limit) throws IluSystemException;


    /* 
     * Variable-length array of byte (Bytes) 
     */
    public/*not used by stubber*/ native int  
    	szBytesSx(byte[] bytes, int start, int len, int limit
    	) throws IluSystemException; 
    public final int  
    	szBytesS(byte[] bytes, int limit
    	) throws IluSystemException {
    	return szBytesSx(bytes, 0, bytes.length, limit);
    }
    public/*not used by stubber*/ native void 
    	outBytesSx(byte[] bytes, int start, int len, int limit
    	) throws IluSystemException; 
    public final void 
    	outBytesS(byte[] bytes, int limit
    	) throws IluSystemException {
    	outBytesSx(bytes, 0, bytes.length, limit);
    }
    public native byte[] 
    	inBytesS(int limit) throws IluSystemException;
    
    /* 
     * Fixed-length array of byte (Opaque) 
     * bytes.length()#limit because 
     * limit from isl and bytes.length from application which
     * might be buggy.
     */
    public/*not used by stubber*/ native int  
    	szBytesAx(byte[] bytes, int start, int limit
    	) throws IluSystemException;
    
    public final int  
    	szBytesA(byte[] bytes, int limit
    	) throws IluSystemException {
    	return szBytesAx(bytes, 0, limit);
    } //szBytesA
    
    public/*not used by stubber*/ native void 
    	outBytesAx(byte[] bytes, int start, int limit
    	) throws IluSystemException;
    public final void 
    	outBytesA(byte[] bytes, int limit
    	) throws IluSystemException {
    	outBytesAx(bytes, 0, limit);
    }
    public native byte[] 
    	inBytesA(int limit) throws IluSystemException;
    
    /** true object utilities */
    public native int  startReadRequest() throws IluSystemException;
    public native boolean doneReadRequest() throws IluSystemException;
    public native boolean noReply() throws IluSystemException;
    public native int beginSizingReply() throws IluSystemException;
    public native boolean startWriteReply(int sz) throws IluSystemException;
    public native boolean doneWriteReply() throws IluSystemException;
    public native int 
    	beginSizingException(int eindex) throws IluSystemException;
    public native boolean 
    	startWriteException(int evalue, int sz) throws IluSystemException;
    public native boolean doneWriteException() throws IluSystemException;
   
    /** surrogate object utilities */
    public native void startWriteRequest(int sz) 
    	throws IluSystemException;
    public native void doneWriteRequest() 
    	throws IluSystemException;
    public native int  
    	startReadReply() throws IluSystemException; 
    	//returns errcode
    public native void doneReadReply() throws IluSystemException;
    
    public IluUserException readException(IluMethodRep method, int errcode)
            throws IluSystemException {
         if ((errcode<=0) || (errcode>method.jExceptions.length)) {
             return null;
         } else {
             IluExceptionRep exRep = method.jExceptions[errcode-1];
             IluUserException ex = exRep.allocInstance();
             ex.readException(this);
             return ex;
         }
     } //readException
    
    private native void nativeMarkCallFailure(int protocolException);
    
    /** for server side only */
    public void unexpectedException(Throwable e) 
            throws xerox.ilu.IluUnexpectedException { 
        nativeMarkCallFailure(IluSystemException.protocolException_Unknown);
        throw new xerox.ilu.IluUnexpectedException(e);
    } //unexpectedException
        
    /** acessors */
    public final boolean needsSizing() {
        return jNeedsSizing;
    }
    
    public final boolean mustFork() {
        return jMustFork;
    }
    
    /** 
     * Debugging aid.
     * Do not use in real applications.
     */
     public final int getSelfCheck() {
        return selfCheck;
    }
        
} //IluCall

