/* IluClassRep.java */
/* Chris Jacobi, January 26, 1997 3:24 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

/* $Id: IluClassRep.java,v 1.17 1997/01/26 23:25:45 jacobi Exp $ */

/* 
 * See also IluJava_IluClassRep.c
 */

package xerox.ilu;

/**
 * An IluClassRep represents the ilu-class of an ILU object.<p>
 *
 * To applications an IluClassRep is immutable.  
 * IluClassRep's are assembled by stubs.
 */
public class IluClassRep {
    private long yIluClass = 0;
        //No garbage collection necessary as ILU will never free classes anyway
        
    private String javaRefIfName; //Interface name for object references
    private String jIluClassName;
    private String brand = null;
    private String uuid = null;
    private String singleton = null;
    private boolean optional = false;
    private boolean collectible = false;
    private String authentication = null;
    private int method_count = 0;
    private IluClassRep[] jIluSuperClasses = null;
    /*friends*/ IluMethodRep[] jMethods = null;
    /*friends*/ Class jSurrClass = null; 
    /*friends*/ IluFactory factory = null;
    
    /*friends*/ static java.util.Hashtable registrations = 
        new java.util.Hashtable();
        
    /** 
     * Prevent creation of bogus IluClassRep's
     */
    private IluClassRep(){
    }
    
    /** 
     * "Constructor" to be used by stubs only.<p>
     *
     * This sets IluClassRep up but does not yet talk to the ilu kernel
     * Java stubs will collect the data but the kernel call will be made
     * when everything is ready only
     * (so stubs don't need to acquire the kernel monitor otmu)<br>
     *
     * Stubs then use IluMethodRep's to continue defining this class 
     * and finally use finishClass to finalize the construction.
     */ 
    //going to prepare iluxport.h: ilu_DefineObjectType
    public static IluClassRep setupClass (
            String javaRefIfName, //interface name (in java's name space)
            String iluClassName, //class name (in ilu's name space)
            String uuid,
            int method_count) throws IluSystemException {
        IluClassRep cls = new IluClassRep();
        cls.javaRefIfName = javaRefIfName;
        cls.jIluClassName = iluClassName;
        cls.uuid = uuid;
        cls.method_count = method_count;
        cls.jMethods = new IluMethodRep[method_count];
        return cls;
    }

    /** 
     * Talks to the ilu kernel and sets up the real data
     */ 
    private native void reallyFinish() throws IluSystemException;
    
    /** 
     * Stub use only. Finishes the registration of the class;
     * at this time the registration is forwarded all the way
     * to the Ilu kernel.
     */ 
    public void finishClass() throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        for (int i = 0 ; i<method_count ; i++) {
            if (this.jMethods[i]==null) {
                IluDebug.clientPreError("methods of class inconsistent");
                throw new IluSomeSystemException();
            }
        }
       if (this.jIluSuperClasses == null) {
           this.jIluSuperClasses = new IluClassRep[0];
       }
       //Prevent garbage collection as we need to hang on to the
       //methods (used for dispatching in true objects) at least
       //as long as the kernel does.
       //xerox.ilu.IluForEver.remember(this); 
       // ... not necessary because of registrations hashtable
       reallyFinish(); 
       registrations.put(jIluClassName, this);
    }
    
    
    /**
     * Defines the surrogate class. 
     * Returns self 
     */
    public IluClassRep setSurrClass(String name) 
            throws IluSystemException {
        if (yIluClass != 0 & !IluEnvironment.isPriviledged(2)) {
            throw new IluSomeSystemException();
        }
        try {
            this.jSurrClass = java.lang.Class.forName(name);
        } catch (java.lang.ClassNotFoundException e) {
            IluDebug.clientPreError("bad java class " + name);
            throw new IluSomeSystemException("bad java class " + name);
        }
        return this;
    } //setSurrClass

    
    /**
     * Defines the surrogate factory. 
     * Returns self 
     */
    public IluClassRep setSurrFactory(xerox.ilu.IluFactory factory) 
            throws IluSystemException {
        if (yIluClass != 0 & !IluEnvironment.isPriviledged(2)) {
            throw new IluSomeSystemException();
        }
        this.factory = factory;
        return this;
    } //setSurrFactory
    
    
    /**
     * Defines the authentication. 
     * Returns self 
     */
    public IluClassRep setAuthentication(String authentication) 
            throws IluSystemException {
        if (yIluClass!=0) throw new IluSomeSystemException();
        this.authentication = authentication;
        return this;
    } //setAuthentication
    

    /* Defines the class to be collectible. 
     * Returns self 
     */
    public IluClassRep setCollectable() 
            throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        xerox.ilu.IluGCClient.init();
        this.collectible = true;
        return this;
    } //setCollectable
    
    
    /**
     * Makes this optional. 
     * Returns self 
     */
    public IluClassRep setOptional() 
            throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        this.optional = true;
        return this;
    } //setOptional
    
    
    /**
     * Makes this a singleton. 
     * Returns self 
     */
    public IluClassRep setSingleton(String singleton) 
            throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        this.singleton = singleton;
        return this;
    } //setSingleton
    
    
    /**
     * Sets a brand name. 
     * Returns self 
     */
    public IluClassRep setBrand(String brand) 
            throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        this.brand = brand;
        return this;
    } //setBrand
    

    /**
     * Defines the super classes. 
     * Returns self 
     */
    public IluClassRep setSuperClasses(xerox.ilu.IluClassRep[] iluSuperClasses) 
            throws IluSystemException {
        if (yIluClass != 0) throw new IluSomeSystemException();
        this.jIluSuperClasses = iluSuperClasses;
        return this;
    } //setSuperClasses


    /**
     * Returns class name (in ilu name space) 
     */
    public final String iluClassName() {
        return jIluClassName;
    }


    /**
     * Returns name of reference interface (in java's name space) 
     */
    public final String referenceIfName() {
        return this.javaRefIfName;
    }
    
    
    /**
     * Returns IluClassRep given its name in ilu name space.
     * Or null if not registered. 
     */
    public static IluClassRep fromIluClassName(String iluClassName) {
        return (IluClassRep) registrations.get(iluClassName);
    }
    
    
    /**
     * Returns IluClassRep given an object.
     */
    public static IluClassRep fromIluObject(xerox.ilu.IluObject0 obj)
        throws xerox.ilu.IluSystemException 
    {
        IluOInt oi = IluOInt.peek((xerox.ilu.IluObject)obj);
        if (oi == null) {
            throw new IluSomeSystemException("bad object");
        }
        return oi.jClassRep;
    }
    
    
    /**
     * Returns whether "this" is an a, or, a subclass of a.
     */
    public boolean isA(IluClassRep a)
        throws xerox.ilu.IluSystemException 
    {
        if (this==a) return true;
        for (int i = 0; i<this.jIluSuperClasses.length; i++) {
            if (this.isA(this.jIluSuperClasses[i])) return true;
        } 
        return false;
    }
    
    
    /** Returns the direct Ilu superclasses of an Ilu class */
    public IluClassRep[] superClasses() {
        IluClassRep[] safeCopy = new IluClassRep[this.jIluSuperClasses.length];
        System.arraycopy(this.jIluSuperClasses, 0, safeCopy, 0, safeCopy.length);
        return safeCopy;
    }
    
    
    /** 
     * Returns the number of super classes.
     */
    public int superClassCnt() {
        int cnt = 0;
        for (int i = 0; i<this.jIluSuperClasses.length; i++) {
            cnt = cnt + 1 + this.jIluSuperClasses[i].superClassCnt();
        }
        return cnt;
    } //superClassCnt
    
    
    /** Accessor; returns whether null is a legal value */
    public boolean isOptional() {
        return optional;
    }
    
    
    /** 
     * Accessor; returns whether whether this class is subject to
     * distributed GC.
     */
    public boolean isCollectible() {
        return collectible;
    }
    
    
    /** returns readable representation for debugging */
    public java.lang.String toString() {
       return super.toString() + ":" + this.jIluClassName;
    }
    
    
    static {
        IluInit.init();
    }

} //IluClassRep

