/* IluExceptionRep.java */
/* Chris Jacobi, January 27, 1997 3:15 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluExceptionRep.java,v 1.5 1997/01/27 23:39:45 jacobi Exp $ */
 
/*
 * See IluJava_IluExceptionRep.c
 */

package xerox.ilu;
import java.util.Hashtable;

/**
 * IluExceptionRep represents an Ilu exception.<p>
 *
 * Stub visible; not likely to be used by applications. <p>
 * This is NOT an exception class but something to 
 * represent an exception class; IluExceptionRep's are not
 * raised and caught but passed around. <br>
 * (Unlike real java exceptions, IluExceptionRep's are 
 * created without capturing a stack frame). <p>
 */
public class IluExceptionRep {
    private long yIluException = 0; //native defined
    /*friendly*/ Class exClass;
    private String i;	//native used
    private String e;	//native used
    
    private static Hashtable tab = new Hashtable(); //java used only
    
    /** 
     * Private to prevent creation of bogus instances.
     * Use defineException to create instances.
     */
    private IluExceptionRep(Class exClass, String i, String e) {
        this.exClass = exClass;
        this.i = i; 
        this.e = e;
    } //IluExceptionRep
    
    /** 
     * Stub accessible. <br>
     * Fetches or defines an IluExceptionRep and its ilu_Exception.
     *
     * @param i  ilu interface name
     * @param e  ilu exception name
     */
    public static IluExceptionRep 
    defineException (String className, String i, String e) {
       IluExceptionRep exRep = null;
       String key = (i==null) ? ("#" + e) : (i + "#" + e);
       Object val = tab.get(key);
       if (val!=null) {
           return ((IluExceptionRep) val);
       }
       synchronized (tab) {
           val = tab.get(key);
           if (val!=null) {
               exRep = ((IluExceptionRep) val);
           } else {
               Class exClass = null;
               try {
                   exClass = Class.forName(className);
               } catch (ClassNotFoundException cnfex) {
                   throw new IluRuntimeException("couldn't define exception");
               }
               exRep = new IluExceptionRep(exClass, i, e);
               exRep.registerException();
               tab.put(key, exRep);
           }
       } 
       return exRep;
    } //defineException
    
    /* called by defineException; called within monitor lock */
    private native void registerException();
    
    public java.lang.String toString() {
       if (i==null) {
           return super.toString() + ":#" + this.e;
       } else {
           return super.toString() + ":" + this.i + "#" + this.e;
       }
    } //toString
    
    /*friendly*/ IluUserException
    allocInstance() throws IluSystemException {
        try {
            IluUserException ex = (IluUserException) exClass.newInstance();
            return ex;
        } catch (Exception e) {
            throw new IluSystemException("couldn't allocate exception");
        }
    } //allocInstance
    
    static {
        IluInit.init();
    }
     
} //IluExceptionRep
