/* IluInit2.java */
/* Chris Jacobi, January 30, 1997 3:48 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluInit2.java,v 1.8 1997/01/31 19:04:53 jacobi Exp $ */
 
/*
 * See IluJava_IluInit.c
 */


package xerox.ilu;

/** 
 * Helper class which takes care of initialization of ILU classes and
 * libraries in the right order. <br>
 * Not for applications.
 */
 
public class IluInit2 {
    
    /* Flag to prevent recursive initializations */
    private static boolean initialized = false;
    private static boolean anySupport = false;

    /* Invokes initialization procedures for various native packages */
    private native static void initializeLibrary();
    
    /* 
     * Monitor to initialize just once.
     * While this makes sure initialization happens only once
     * applications can still fail if they don't initialize
     * ILU before forking threads; that is rare and requires 
     * applications to fork before initializing classes
     * which reference ILU stuff.
     */
    private static final synchronized boolean mustInit() {
       boolean wasinit = initialized;
       initialized = true;
       return ! wasinit;
    } //mustInit
    
    private static final synchronized boolean mustInitAnySupport() {
       boolean wasinit = anySupport;
       anySupport = true;
       return ! wasinit;
    } //mustInitAnySupport
    
    private static void log(String s) {
        if (IluDebug.initsFlag>0) {
            System.err.println("IluInit2 did " + s);
        }
    } //log
    
    /** Initialization procedures.  Idempotent */
    public static final void init () {
        if (! initialized) {
            if (mustInit()) { 
                try {
                    System.loadLibrary("IluJava");
                    xerox.ilu.IluDebug.init();
                    if (IluDebug.initsFlag>0) log("IluDebug");
                    xerox.ilu.IluDebugHooks.init();
                    if (IluDebug.initsFlag>0) log("IluDebugHooks");
                    xerox.ilu.IluAlarmTech.init();
                    if (IluDebug.initsFlag>0) log("IluAlarmTech");
                    xerox.ilu.IluRT0.init();
                    if (IluDebug.initsFlag>0) log("IluRT0");
                    initializeLibrary();
                    if (IluDebug.initsFlag>0) log("library");
                    xerox.ilu.IluSurrogateConnection.init();
                    if (IluDebug.initsFlag>0) log("IluSurrogateConnection");
                    xerox.ilu.IluDebugHooks.defineDebugHook(null);
                    if (IluDebug.initsFlag>0) log("def DH");
                    xerox.ilu.IluDebugHooks.defineCheckFailure(null);
                    if (IluDebug.initsFlag>0) log("def CF");
                } catch (UnsatisfiedLinkError e) {
                    System.err.println("** Unsatisfied link  " +  e);
                    System.err.println("** Is the LD_LIBRARY_PATH set?");
                    e.printStackTrace();
                    //stay alive for debugging...
                    for (;;) {}
                } catch (Exception e) {
                    System.err.println("** Failed initializing Ilu " +  e);
                    e.printStackTrace();
                    //stay alive for debugging...
                    for (;;) {}
                }
            }
        }
    } //init
    
    /** Initialization procedures for the "Any" support.  Idempotent */
    public static final void initAnySupport() {
        if (! anySupport) {
            init();
            if (mustInitAnySupport()) { 
                try {
                    System.loadLibrary("IluJavaAnySupport");
                    if (IluDebug.initsFlag>0) log("IluJavaAnySupport");
                } catch (UnsatisfiedLinkError e) {
                    System.err.println("** Unsatisfied link  " +  e);
                    System.err.println("** Failed loading 'Any' support");
                    e.printStackTrace();
                } catch (Exception e) {
                    System.err.println("** Failed loading 'Any' support " +  e);
                    e.printStackTrace();
                }
            }
        }
    } //initAnySupport
    
    static {
        init();
    }

} // IluInit2

