/* IluJava_IluAlarmTech.c */
/* Chris Jacobi, February 5, 1997 4:14 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* 
 * $Id: IluJava_IluAlarmTech.c,v 1.5 1997/02/06 00:31:29 jacobi Exp $ 
 *
 *
 * IluAlarmTech implementation for Ilu-Java.
 * Corresponding Java side in xerox.ilu.IluAlarmTech.java
 */
 
#include <StubPreamble.h>
#include "xerox_ilu_IluAlarmTech.h"
#include <iluxport.h>
#include <IluJava.h>

typedef void (*alarmProc)(ilu_private rock);


struct pair_s {
  ilu_integer f1;	
  ilu_integer f2;	
};
typedef struct pair_s pair;


/* Use dynamic_method simply because I couldn't get
 * execute_java_static_method to work correctly.
 */ 
static struct Hxerox_ilu_IluAlarmTech * _ilujava_jalarm;

EXPORTJAVA void 
xerox_ilu_IluAlarmTech_giveInstance(struct Hxerox_ilu_IluAlarmTech * jobj)
{
    _ilujava_jalarm = jobj;
}


INTERNALONLY ilu_refany 
_ilujava_CreateAlarm()
/* provided to the ilu kernel */
{
    long jx;
    struct execenv *ee = EE(); /*Java execution environment*/
    if (_ilujava_jalarm==0) {
        _iluJava_PANIC("$ IluAlarmTech CreateAlarm: BAD INSTANCE");
    }    
    if (_ilujava_alarmFlag>0) {
        ilu_DebugPrintf("$ IluAlarmTech CreateAlarm\n");
    }
    jx = execute_java_dynamic_method(ee, 
        (HObject *) _ilujava_jalarm,
    	"createAlarm", 
    	"()Lxerox/ilu/IluAlarmTech;"
    	);
    if (ee->exceptionKind !=0) {
        _iluJava_PANIC("$ IluAlarmTech CreateAlarm: JAVA EXCEPTION\n");
    }
    return (ilu_refany) jx;
}


EXPORTJAVA void 
xerox_ilu_IluAlarmTech_wakeIlu(struct Hxerox_ilu_IluAlarmTech * jobj, 
	long x1, long x2, long x3, long x4
	)
/* Called from Java to invoke the alarm */
{
    union tag_proc64 {
        pair p;
        void (*proc)(ilu_private rock);
    } proc64;
    union tag_rock64 {
        pair p;
        ilu_private rock;
    } rock64;
    proc64.p.f1 = (ilu_integer) x1;
    proc64.p.f2 = (ilu_integer) x2;
    rock64.p.f1 = (ilu_integer) x3;
    rock64.p.f2 = (ilu_integer) x4;
    if (_ilujava_alarmFlag>1) {
        ilu_DebugPrintf("$ IluAlarmTech wakeIlu\n");
    }
    proc64.proc(rock64.rock);
}


INTERNALONLY void 
_ilujava_SetAlarm(ilu_refany alarm,
	ilu_FineTime t,
	void (*proc)(ilu_private rock),
	ilu_private rock
	)
/* Procedure provided to the ilu kernel */
{
    ilu_cardinal milli;
    long unused;
    Hxerox_ilu_IluAlarmTech * jalarm = (Hxerox_ilu_IluAlarmTech *) alarm;
    union tag0_proc64 {
        pair p;
        void (*proc)(ilu_private rock);
    } proc64;
    union tag0_rock64 {
        pair p;
        ilu_private rock;
    } rock64;
    /* convert ilu-time to duration in milliseconds */
    ilu_FineTime now = ilu_FineTime_Now();
    ilu_FineTime length;
    length = ilu_FineTime_Sub(t, now);
    /* simplistic way to handle negative times */
    if (ilu_FineTime_Cmp(now, t)>=0) {
        length.ft_s = 0; milli = 0;
    } else {
        milli = ilu_rescale(t.ft_t, ilu_FineTimeRate, 1000);
    }
    /* convert arguments and call into java */
    proc64.proc = proc;
    rock64.rock = rock;
    if (_ilujava_alarmFlag>1) {
        ilu_DebugPrintf("$ IluAlarmTech SetAlarm %d %d\n", (int) length.ft_s, milli);
    }
    unused = execute_java_dynamic_method(0, 
        (HObject *) jalarm, 
        "setAlarm", "(IIIIII)V", 
        (long) length.ft_s, (long) milli,
        (long) proc64.p.f1, (long) proc64.p.f2, 
        (long) rock64.p.f1, (long) rock64.p.f2
        );
}


INTERNALONLY void 
_ilujava_UnsetAlarm(ilu_refany alarm)
/* Procedure provided to the ilu kernel */
{
    Hxerox_ilu_IluAlarmTech * jalarm;
    jalarm = (Hxerox_ilu_IluAlarmTech *) alarm;
    if (_ilujava_alarmFlag>1) {
    	ilu_DebugPrintf("$ IluAlarmTech UnsetAlarm\n");
    }
    execute_java_dynamic_method(0, 
    	(HObject *) jalarm, "unsetAlarm", "()V"
    	);
}

INTERNALONLY void 
_ilujava_DestroyAlarm(ilu_refany alarm)
/* Procedure provided to the ilu kernel */
{
    Hxerox_ilu_IluAlarmTech * jalarm;
    jalarm = (Hxerox_ilu_IluAlarmTech *) alarm;
    if (_ilujava_alarmFlag>1) {
    	ilu_DebugPrintf("$ IluAlarmTech provided\n");
    }
    execute_java_dynamic_method(0, 
    	(HObject *) jalarm, "destroyAlarm", "()V"
    	);
}


static ilu_MainLoop alarmData;

EXPORTLIMITED void 
_ilujava_AlarmTechInit()
/* call this to initialize this file */
{
    alarmData.ml_run         	= 0;
    alarmData.ml_exit         	= 0;
    alarmData.ml_register_input	= 0;
    alarmData.ml_unregister_input	= 0;
    alarmData.ml_register_output	= 0;
    alarmData.ml_unregister_output	= 0;
    alarmData.ml_create_alarm	= _ilujava_CreateAlarm;
    alarmData.ml_set_alarm    	= _ilujava_SetAlarm;
    alarmData.ml_unset_alarm	= _ilujava_UnsetAlarm;
    alarmData.ml_destroy_alarm	= _ilujava_DestroyAlarm;
    ilu_SetMainLoop(&alarmData);
}

