/* IluJava_IluCall.c */
/* Chris Jacobi, February 5, 1997 4:16 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluCall.c,v 1.38 1997/02/06 00:32:09 jacobi Exp $ */

/*
 * C side for IluCall.java
 */

 
#include <StubPreamble.h>
#include <interpreter.h>
#include <javaString.h>

#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"
#include "IluJava_JStubs.h"
#include "IluJava_JOps.h"

#define JCALL  struct Hxerox_ilu_IluCall *



static void OptionalPrintE(ilu_Error err, char * t) {
    if (_ilujava_iluDebuggingFlag>0) {
        ilu_DebugPrintf("$ **IluJava_IluCall.c: %s %s %s \n", 
           t, ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
    }
}

static void OptionalPrint0(char * t) {
    if (_ilujava_iluDebuggingFlag>0) {
        ilu_DebugPrintf("$ **IluJava_IluCall.c: %s \n", t);
    }
}
   
#define IluCallException "xerox/ilu/IluCallException"


static int ilujava_nullBuffer = 0; 
    /* placeholder where empty pointers can point to */

static Hjava_lang_String * ilujava_jEmptyString;
    /* Replace null strings by this empty string 
     * Initialized in "givePrototype"
     */

static void ilujava_NullPointerException()
{
    OptionalPrint0("NullPointerException");
    SignalError(EE(), "java/lang/NullPointerException", 0);
}

static void ilujava_SignalInconsistency()
{
    ilu_DebugPrintf("$ IluCall inconsistency\n");
    SignalError(EE(), "xerox/ilu/IluInconsistentCallException", "IluCall");
}






/* Self checking codes to enforce legal transitions */
#define CALL_INACTIVE 0
#define CALL_FINISHED 0
    /* Not yet initialized, or, done for now.
     * The point is that these codes must not be encountered at random
     * times. 
     */
#define CALL_IN 1
    /* Input operations are expected. 
     */
#define CALL_SZ 2
    /* Sizing operations are expected. 
     */
#define CALL_OUT 3
    /* Output operations are expected. 
     */
#define CALL_INBETWEEN 4
    /* Done with writing or reading.
     * Controlled state change expected. 
     */
#define CALL_FINALIZED 99
    /* Marked to be finalized.  We set this state but NEVER want 
     * to encounter it. 
     */
     
#define EXPECT_UNFINISHED_V \
    switch (unhand(jcall)->selfCheck) {\
    case CALL_SZ:\
    case CALL_OUT:\
    case CALL_IN:\
    case CALL_INBETWEEN:\
        break;\
    default:\
        ilu_DebugPrintf("$ IluCall_EXPECT_UNFINISHED_V: %d\n",\
            unhand(jcall)->selfCheck);\
	ilujava_SignalInconsistency();\
        return;\
    }
    
#define EXPECT_UNFINISHED_V_QUIET \
    switch (unhand(jcall)->selfCheck) {\
    case CALL_SZ:\
    case CALL_OUT:\
    case CALL_IN:\
    case CALL_INBETWEEN:\
        break;\
    default:\
        ilu_DebugPrintf("$ EXPECT_UNFINISHED_V_QUIET: %d\n",\
            unhand(jcall)->selfCheck);\
        return;\
    }
    
#define EXPECT_LIFE_V \
    switch (unhand(jcall)->selfCheck) { \
    case CALL_SZ: \
    case CALL_OUT: \
    case CALL_IN: \
        break; \
    default:  \
        ilu_DebugPrintf("$IluCall_EXPECT_LIFE_V\n"); \
	ilujava_SignalInconsistency(); \
        return;\
    }
        
#define EXPECT_SZ_OR_INBETWEEN_1 \
    switch (unhand(jcall)->selfCheck) { \
    case CALL_SZ: \
    case CALL_INBETWEEN: \
        break; \
    default:  \
        ilu_DebugPrintf("$IluCall_EXPECT_SZ_OR_INBETWEEN_1\n"); \
	ilujava_SignalInconsistency(); \
        return 0;\
    }
    

#define EXPECT_SZ_OR_INBETWEEN_V \
    switch (unhand(jcall)->selfCheck) { \
    case CALL_SZ: \
    case CALL_INBETWEEN: \
        break; \
    default:  \
        ilu_DebugPrintf("$IluCall_EXPECT_SZ_OR_INBETWEEN_V\n"); \
	ilujava_SignalInconsistency(); \
        return;\
    }
    

#define EXPECT_V(XX) \
    if (unhand(jcall)->selfCheck != XX) { \
        ilu_DebugPrintf("$IluCall_EXPECT_V (should %d was %d) %d\n",\
            XX, unhand(jcall)->selfCheck, __LINE__); \
	ilujava_SignalInconsistency();  \
        return; \
    }

#define EXPECT_1(XX) \
    if (unhand(jcall)->selfCheck != XX) { \
        ilu_DebugPrintf("$IluCall_EXPECT_1 (should %d was %d) %d\n",\
            XX, unhand(jcall)->selfCheck, __LINE__); \
	ilujava_SignalInconsistency();  \
        return 0; \
    }

#define TRANSITION(XX) unhand(jcall)->selfCheck = XX;



#define PROLOG_1(XX) \
    ILU_ERRS((IoErrs, bad_param)) ioerrs; \
    ilu_Call cCall = Access(jcall, yCall, ilu_Call); \
    EXPECT_1(XX)

#define PROLOG_V(XX) \
    ILU_ERRS((IoErrs, bad_param)) ioerrs; \
    ilu_Call cCall = Access(jcall, yCall, ilu_Call); \
    EXPECT_V(XX)
    
#define ERRCHECK_1 \
    if (ILU_ERRNOK(ioerrs)) {\
        OptionalPrintE(ioerrs, "IluCall");\
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx100"); \
        return 0; \
    }

#define ERRCHECK_V \
    if (ILU_ERRNOK(ioerrs)) { \
        OptionalPrintE(ioerrs, "IluCall"); \
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx101"); \
        return; \
    }






static /*interpreter.h*/ ClassClass * ilujava_javaClassIluCall;

/* called from static block to initialize the J-class */
EXPORTJAVA void 
xerox_ilu_IluCall_givePrototype(JCALL jcall)
{
    ilujava_javaClassIluCall = obj_classblock(jcall);
    ilujava_jEmptyString = makeJavaString((char *)&ilujava_nullBuffer, 0);
}


/* Create an object */
/* Made native because I want to add more fields as well as add caching */
/* Consider making this avvailable to java ? */
static JCALL 
ilujava_constructIluCall(JCALL unused)
{
    JCALL jcall;
    jcall = (JCALL) ObjAlloc(ilujava_javaClassIluCall, 0);
    	/* ObjAlloc does not execute Java constructor */
    xerox_ilu_IluCall_initCallShell(jcall);
    return jcall;
}


/* Called by the java finalize procedure */
EXPORTJAVA void 
xerox_ilu_IluCall_nativeFinalize(JCALL jcall)
{
    ilu_Call cCall;
    TRANSITION(CALL_FINALIZED);
    cCall = Access(jcall, yCall, ilu_Call);
    Access(jcall, yCall, ilu_Call) = 0; 
    ilu_free(cCall);
}


EXPORTJAVA void
xerox_ilu_IluCall_initCallShell(JCALL jcall)
{
    ilu_Call cCall = ilu_must_malloc(sizeof(struct _ilu_Call_s));
    TRANSITION(CALL_INACTIVE);
    Access(jcall, yCall, ilu_Call) = cCall;
}

EXPORTJAVA JCALL 
xerox_ilu_IluCall_nativeStartCall(
	JCALL junused,
	struct Hxerox_ilu_IluMethodRep * jMethodRep,
	struct Hxerox_ilu_IluSurrogateObject * jObj,
	struct Hxerox_ilu_IluPassport * jPassport
	)
/* (surrogate) stub initiating a call */
{
    ILU_ERRS((IoErrs)) errs;
    ilu_Connection newConnection;
    ilu_boolean success;
    JCALL jcall;
    ilu_Call cCall;
    struct Hxerox_ilu_IluOInt * joi;
    struct Hxerox_ilu_IluObject * jObjIf = 
    	(struct Hxerox_ilu_IluObject *) jObj;
    struct Hxerox_ilu_IluClassRep * jStaticClass = 
    	unhand(jMethodRep)->jClassRep; /*methods carry static class*/
    
    jcall = ilujava_constructIluCall(junused);
    cCall = Access(jcall, yCall, ilu_Call);
    /* GC of stuffing the following java objects into the C heap is ok:
     * jObj will be held on by stub;
     * jMethodRep, jClassRep are not subject to GC
     */
    
    EXPECT_1(CALL_INACTIVE)
    unhand(jcall)->jObj = jObjIf;
    unhand(jcall)->jMethodRep = jMethodRep;
    unhand(jcall)->jClassRep = jStaticClass;
    unhand(jcall)->jMustFork = 0;
    joi = xerox_ilu_IluOInt_nPeekWithCall(0, jObjIf, jcall);
    unhand(jcall)->jOIntDisc = joi;
    success = ilu_StartCall(
    	cCall,
    	Access(joi, yServer, ilu_Server),
        Access(jStaticClass, yIluClass, ilu_Class),
	Access(jMethodRep, yIluMethod, ilu_Method),
	JAVALANGIDX,
	_iluJava_getIluPassport((ilu_refany) jPassport), 
	&newConnection, 
	&errs
    	);
    if (newConnection) {
        Access(jcall, yCallSurrConn, ilu_Connection) = newConnection;
        unhand(jcall)->jMustFork = 1;
    }
    if (success) {
        TRANSITION(CALL_SZ)
        unhand(jcall)->needFinish = 1;
        unhand(jcall)->jNeedsSizing = (long) ilu_CallNeedsSizing(cCall);
        return jcall;
    } else {
        unhand(jcall)->needFinish = 0;
        xerox_ilu_IluCall_nativeFinalize(jcall);
        _iluJava_IluErrorToJavaException(errs, "xxx_2");
        return jcall;
    }
} /* nativeStartCall */




EXPORTJAVA void 
xerox_ilu_IluCall_finishCall(JCALL jcall)
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    if (unhand(jcall)->needFinish == 0) return;  /*seems already done*/
    EXPECT_UNFINISHED_V
    ILU_CLER(err);
    ilu_FinishCall(Access(jcall, yCall, ilu_Call), &err);
    unhand(jcall)->needFinish = 0;
    unhand(jcall)->jObjTracable = 0;
    unhand(jcall)->jOIntDisc = 0;
    TRANSITION(CALL_FINISHED)
    if (ILU_ERRNOK(err)) {
        _iluJava_IluErrorToJavaException(err, "finishCall"); 
    }
}

static void 
finisCall_noErrs(JCALL jcall)
/* Like xerox_ilu_IluCall_finishCall but for internal use
 * Does NOT report errors into java as this is already a 
 * consequence of an error to be reported
 *
 * FinishCall is conditional because we don't really
 * trust the stubs (in case of applets).
 */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    if (unhand(jcall)->needFinish) {
        EXPECT_UNFINISHED_V_QUIET
        ILU_CLER(err);
        ilu_FinishCall(Access(jcall, yCall, ilu_Call), &err);
        unhand(jcall)->needFinish = 0;
        unhand(jcall)->jObjTracable = 0;
        unhand(jcall)->jOIntDisc = 0;
        TRANSITION(CALL_FINISHED)
        if (ILU_ERRNOK(err)) {
            /* do nothing */ 
        }
    }
}


static void 
finisCall_noCheck(JCALL jcall)
/* Like finisCall_noErrs but doesn't deal with
 * self checking.
 *
 * FinishCall is conditional because we don't really
 * trust the stubs (in case of applets).
 */
{
    if (unhand(jcall)->needFinish) {
        ILU_ERRS((bad_locks, IoErrs)) err;
        ILU_CLER(err);
        ilu_FinishCall(Access(jcall, yCall, ilu_Call), &err);
        unhand(jcall)->needFinish = 0;
        if (ILU_ERRNOK(err)) {
            /* do nothing */ 
        }
    }
}


EXPORTJAVA void
xerox_ilu_IluCall_startWriteRequest(JCALL jcall, long sz)
    /* used by surrogates */
{
    ILU_ERRS((IoErrs)) err;
    EXPECT_SZ_OR_INBETWEEN_V
    if (!ilu_StartRequest(Access(jcall, yCall, ilu_Call), sz, &err)) {
        /* error */
        finisCall_noErrs(jcall);
        _iluJava_IluErrorToJavaException(err, "startWriteRequest");
        return;
    }
    TRANSITION(CALL_OUT)
}


EXPORTJAVA void
xerox_ilu_IluCall_doneWriteRequest(JCALL jcall)
    /* used by surrogates */
{
    ILU_ERRS((IoErrs)) err;
    EXPECT_V(CALL_OUT)
    if (!ilu_FinishRequest(Access(jcall, yCall, ilu_Call), &err)) {
        /* Error. 
         * Call is over; call into java anyway in case data structures
         * would need to be returned.
         */
        finisCall_noErrs(jcall);
        _iluJava_IluErrorToJavaException(err, "doneWriteRequest");
        return;
    }
    TRANSITION(CALL_INBETWEEN)
    /* ??? Should finishCall if the method is ASYNCHRONOUS */
    return;
}


EXPORTJAVA long
xerox_ilu_IluCall_startReadReply(JCALL jcall)
    /* used by surrogates */
{
    ilu_cardinal errorStat;
    ilu_ProtocolException protoErr;
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_INBETWEEN)
    protoErr = ilu_GetReply(cCall, &errorStat, &err);
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_startReadReply proto: %d user: %d\n", (int) protoErr, (int) errorStat);
    }
    if (protoErr == ilu_ProtocolException_Success) {
        TRANSITION(CALL_IN)
        return errorStat;
    } else {
        int asint = (int) protoErr;
        finisCall_noErrs(jcall);
        asint = - asint; /*this should make it negative*/
        TRANSITION(CALL_FINISHED)
        if (protoErr == ilu_ProtocolException_Not) {
            _iluJava_IluErrorToJavaException(err, "ProtocolException");
        }
        return (long) asint;
    }
} /*startReadReply*/


EXPORTJAVA void
xerox_ilu_IluCall_doneReadReply(JCALL jcall)
/* used by surrogates */
{
    ILU_ERRS((IoErrs)) err;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    if (unhand(jcall)->needFinish) {
        EXPECT_V(CALL_IN)
        if ( ! ilu_ReplyRead(cCall, &err)) {
            /* Error; Call finishCall anyway. */
            OptionalPrintE(err, "doneReadReply"); 
            finisCall_noErrs(jcall);
            _iluJava_IluErrorToJavaException(err, "doneReadReply");
            return;
        }
        finisCall_noErrs(jcall);
    }
    TRANSITION(CALL_INBETWEEN)
}


EXPORTJAVA long
xerox_ilu_IluCall_startReadRequest(JCALL jcall)
    /* used by true objects */
{
    TRANSITION(CALL_IN)
    /* I think there is nothing to do */
    return 0;
}


EXPORTJAVA long
xerox_ilu_IluCall_doneReadRequest(JCALL jcall)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_boolean     ans;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_IN)
    ans = ilu_RequestRead(cCall, &err);
    if (ILU_ERRNOK(err)) {
        finisCall_noErrs(jcall);
    }
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


#define hasExceptions(jcall) ((unhand(unhand(jcall)->jMethodRep)->jExceptions)!=0)

EXPORTJAVA long
xerox_ilu_IluCall_beginSizingReply(JCALL jcall)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_cardinal ret; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_INBETWEEN)
    ret = ilu_BeginSizingReply(cCall, hasExceptions(jcall), &err);
    TRANSITION(CALL_SZ)
    return (long) ret;
}


EXPORTJAVA long
xerox_ilu_IluCall_startWriteReply(JCALL jcall, long argSize)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_SZ_OR_INBETWEEN_1
    ans = ilu_BeginReply(cCall, hasExceptions(jcall), (ilu_cardinal) argSize, &err);
    if (! ans) {
        finisCall_noCheck(jcall);
    }
    TRANSITION(CALL_OUT)
    return (ans!=0);
}

EXPORTJAVA long
xerox_ilu_IluCall_beginSizingException(JCALL jcall, long eindex)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_cardinal ret; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_INBETWEEN)
    ret = ilu_BeginSizingException(cCall, (ilu_integer) eindex, &err);
    TRANSITION(CALL_SZ)
    return (long) ret;
}


EXPORTJAVA long
xerox_ilu_IluCall_startWriteException(JCALL jcall, long evalu, long argSize)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall;
    cCall = Access(jcall, yCall, ilu_Call);
    
    EXPECT_SZ_OR_INBETWEEN_1
    TRANSITION(CALL_OUT)
    ans = ilu_BeginException(cCall, (ilu_integer) evalu, (ilu_cardinal) argSize, &err);
    if (! ans) {
        finisCall_noCheck(jcall);
    }
    return (ans!=0);
}


EXPORTJAVA long
xerox_ilu_IluCall_noReply(JCALL jcall)
    /* used by true objects */
{
    ILU_ERRS((bad_param, bad_locks, broken_locks)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_INBETWEEN)
    ans = ilu_NoReply(cCall, &err);
    finisCall_noCheck(jcall);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA long
xerox_ilu_IluCall_doneWriteReply(JCALL jcall)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_OUT)
    ans = ilu_FinishReply(cCall, &err);
    finisCall_noCheck(jcall);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA long
xerox_ilu_IluCall_doneWriteException(JCALL jcall)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_1(CALL_OUT)
    ans = ilu_FinishException(cCall, &err);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA long
xerox_ilu_IluCall_szInt32(JCALL jcall, long ival)
{
    ilu_cardinal sz; 
    ilu_integer i = ival;
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfInteger(cCall, i, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt32(JCALL jcall,
	long ival)
{
    ilu_integer i = ival;
    PROLOG_V(CALL_OUT)
    ilu_OutputInteger(cCall, i, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inInt32(JCALL jcall)
{
    ilu_integer value;
    PROLOG_1(CALL_IN)
    ilu_InputInteger(cCall, &value, &ioerrs);
    ERRCHECK_1
    return value;
}


EXPORTJAVA long
xerox_ilu_IluCall_szBool(JCALL jcall, long ival)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfBoolean(cCall, (ival != 0), &ioerrs);
    ERRCHECK_1
    return (long) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outBool(JCALL jcall, long ival)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputBoolean(cCall, (ival != 0), &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inBool(JCALL jcall)
{
    ilu_boolean value;
    PROLOG_1(CALL_IN)
    ilu_InputBoolean(cCall, &value, &ioerrs);
    ERRCHECK_1
    return (long) value;
}

EXPORTJAVA long
xerox_ilu_IluCall_szInt16(JCALL jcall, short ival)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfShortInteger(cCall, (ilu_shortinteger) ival, &ioerrs);
    ERRCHECK_1
    return (long) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt16(JCALL jcall, short ival)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputShortInteger(cCall, (ilu_shortinteger) ival, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA short
xerox_ilu_IluCall_inInt16(JCALL jcall)
{
    ilu_shortinteger value;
    PROLOG_1(CALL_IN)
    ilu_InputShortInteger(cCall, &value, &ioerrs);
    ERRCHECK_1
    return (short) value;
}


EXPORTJAVA long
xerox_ilu_IluCall_szByte(JCALL jcall, char ival)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfByte(cCall, (ilu_byte) ival, &ioerrs);
    ERRCHECK_1
    return (long) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outByte(JCALL jcall, char ival)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputByte(cCall, (ilu_byte) ival, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA char
xerox_ilu_IluCall_inByte(JCALL jcall)
{
    ilu_byte value;
    PROLOG_1(CALL_IN)
    ilu_InputByte(cCall, &value, &ioerrs);
    ERRCHECK_1
    return (char) value;
}


EXPORTJAVA long
xerox_ilu_IluCall_szChar16(JCALL jcall, unicode ival)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfCharacter(cCall, (ilu_character) ival, &ioerrs);
    ERRCHECK_1
    return (long) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outChar16x(JCALL jcall, unicode ival)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputCharacter(cCall, (ilu_character) ival, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA unicode
xerox_ilu_IluCall_inChar16(JCALL jcall)
{
    ilu_character value;
    PROLOG_1(CALL_IN)
    ilu_InputCharacter(cCall, &value, &ioerrs);
    ERRCHECK_1
    return (unicode) value;
}


EXPORTJAVA long
xerox_ilu_IluCall_szChar8(JCALL jcall, unicode ival)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfShortCharacter(cCall, (ilu_shortcharacter) ival, &ioerrs);
    ERRCHECK_1
    return (long) sz;
}

EXPORTJAVA void
xerox_ilu_IluCall_outChar8x(JCALL jcall, unicode ival)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputShortCharacter(cCall, (ilu_shortcharacter) ival, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA unicode
xerox_ilu_IluCall_inChar8(JCALL jcall)
{
    ilu_shortcharacter value;
    PROLOG_1(CALL_IN)
    ilu_InputShortCharacter(cCall, &value, &ioerrs);
    ERRCHECK_1
    return (unicode) value;
}


EXPORTJAVA long
xerox_ilu_IluCall_szObjectx(JCALL jcall,
	struct Hxerox_ilu_IluObject * jobj, 
	/*boolean*/ long discriminant,
	struct Hxerox_ilu_IluClassRep * jStaticClass
	)
{  
    struct Hxerox_ilu_IluOInt * joi;
    struct Hxerox_ilu_IluClassRep * jActualClass;
    ilu_Object cObject = 0;
    ilu_Class cStaticClass = 0;
    ilu_Class cActualClass = 0;
    ilu_cardinal sz;
    ilu_Server cServer;
    PROLOG_1(CALL_SZ)
    if (jStaticClass==0) {
        SignalError(EE(), IluCallException, "xxx-21");
        return 0;
    }
    cStaticClass = Access(jStaticClass, yIluClass, ilu_Class);
    if (jobj) {
        joi = xerox_ilu_IluOInt_nPeekWithCall(0, jobj, jcall); 
        if (joi==0) {
            OptionalPrintE(ioerrs, "IluCall.szObject: problem");
            SignalError(EE(), IluCallException, "xxx-21");
            return 0;
        }
        /* Object's server may differ from call's server */
        cServer = Access(joi, yServer, ilu_Server);
        jActualClass = unhand(joi)->jClassRep;
        if (jActualClass==0) {
            SignalError(EE(), IluCallException, "xxx-23a");
            return 0;
        }
        cActualClass = Access(jActualClass, yIluClass, ilu_Class);
        ilu_EnterServer(cServer, cActualClass); 
        	/*actual or static?  the c runtime uses actual*/
        cObject = Access(joi, yIluKernelObject, ilu_Object); 
        if (cObject==0) {
            /*we freed the object right before entering the monitor*/
            ilu_ExitServer(cServer, cActualClass);
            SignalError(EE(), IluCallException, "xxx-23b");
            return 0;
        }
    }
    sz = ilu_SizeOfObjectID(cCall, cObject, discriminant!=0, cStaticClass, &ioerrs);
    if (jobj) ilu_ExitServer(cServer, cActualClass); 
    ERRCHECK_1
    return (long) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outObjectx(JCALL jcall,
	struct Hxerox_ilu_IluOInt * joi, 
	/*boolean*/ long discriminant,
	struct Hxerox_ilu_IluClassRep * jStaticClass)
{
    struct Hxerox_ilu_IluClassRep * jActualClass;
    ilu_Object cObject = 0;
    ilu_Class cStaticClass = 0;
    ilu_Class cActualClass = 0;
    ilu_Server cServer;
    PROLOG_V(CALL_OUT)
    if (jStaticClass==0) {
        OptionalPrintE(ioerrs, "IluCall.outObject: no static class");
        SignalError(EE(), IluCallException, "xxx-21");
    }
    if (joi) {
        jActualClass = unhand(joi)->jClassRep;
        if (jActualClass==0) {
            SignalError(EE(), IluCallException, "xxx-23d");
            return;
        }
        cActualClass = Access(jActualClass, yIluClass, ilu_Class);
        /* Use Object's server; not call's server */
        cServer = Access(joi, yServer, ilu_Server); 
        if (cServer==0) {
            SignalError(EE(), IluCallException, "xxx-23f bad server");
            return;
        }
        ilu_EnterServer(cServer, cActualClass);
        cObject = Access(joi, yIluKernelObject, ilu_Object);
        if (cObject==0) {
            /* we freed the object right before entering the monitor*/
            ilu_ExitServer(cServer, cActualClass);
            SignalError(EE(), IluCallException, 
                "xxx-23e Object has been unregistered");
            return;
        }
    }
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ IluCall_outObjectx c<%x>\n", cObject);
    }
    cStaticClass = Access(jStaticClass, yIluClass, ilu_Class);
    ilu_OutputObjectID(cCall, cObject, discriminant!=0, cStaticClass, &ioerrs);
    /* ?? locking for collectible not understood */
    ERRCHECK_V
}


EXPORTJAVA struct Hxerox_ilu_IluOInt *
xerox_ilu_IluCall_inOI(JCALL jcall,
	long discriminant, 
	struct Hxerox_ilu_IluClassRep * jStaticClass)
{
    struct Hxerox_ilu_IluOInt * joi; 
    ilu_Server cServer;
    ilu_Object cIluObject;
    ilu_Class cStaticClass;
    PROLOG_1(CALL_IN)
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_inOI\n");
    }
    if (jStaticClass==0) {
        OptionalPrintE(ioerrs, "IluCall.inOI: no static class");
        SignalError(EE(), IluCallException, "xxx-21");
    }
    cStaticClass = Access(jStaticClass, yIluClass, ilu_Class);
    ilu_InputObjectID(cCall, &cIluObject, discriminant!=0, cStaticClass, &ioerrs);
    /*o!=NIL => Inside(*o's server...); ILU_ERRNOK(*err) => *o==NIL*/ 
    ERRCHECK_1
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ IluCall_inOI got c<%x>\n", cIluObject);
    }
    if (cIluObject==0) {
        return 0;
    }
    joi = (struct Hxerox_ilu_IluOInt *) ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    if (_ilujava_callsFlag>0) {
        ilu_DebugPrintf("$ IluCall_inOI: got j<%x>\n", joi);
    }
    if (joi == 0) {
        /* Object is not/no-more known; create a surrogate */
        if (_ilujava_callsFlag>0) {
            ilu_DebugPrintf("$ IluCall_inOI: needs to create ob\n");
        }
        cServer = ilu_ServerOfObject(cIluObject);
        if (ilu_TrueServerP(cServer)) {
            ilu_ExitServer(cServer, cStaticClass);
            OptionalPrint0("xxx-True object has disappeared");
            SignalError(EE(), IluCallException, "xxx-TOHD");
            return 0;
        } else {
            joi = (struct Hxerox_ilu_IluOInt *) _iluJava_creatSurrOIFromRegistry(
            	cIluObject, (ilu_refany) jStaticClass);
        }
    } else {
        if (unhand(joi)->retained != 0) {unhand(joi)->ghost = 1;}
    }
    unhand(jcall)->jObjTracable = (HObject*) joi; /*make traceable*/
    /* Need object's server; not call's server */
    cServer = Access(joi, yServer, ilu_Server);
    if (cServer==0) {
        BUGPRINT("$ SERVER MISSING; substituted...");
        cServer = cCall->ca_server;
    }
    ilu_ExitServer(cServer, cStaticClass); /*iluxport.h says: "static_type"*/
    return joi;
}

EXPORTJAVA Hxerox_ilu_IluOInt *
xerox_ilu_IluCall_getCallSingletonOIx(JCALL jcall)
{
    ilu_Server cServer;
    ilu_Object cIluObject;
    struct Hxerox_ilu_IluOInt * joi = 0; 
    PROLOG_1(CALL_IN);
    cIluObject = ilu_GetCallSingleton(cCall, &ioerrs);
    ERRCHECK_1
    if (cIluObject!=0) {
        cServer = ilu_ServerOfObject(cIluObject);
        joi = (struct Hxerox_ilu_IluOInt*)
    		ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
        if (joi == 0) {
            ilu_ExitServer(cServer, ilu_IntroTypeOfCall(cCall));
            OptionalPrint0("yyy-2");
            SignalError(EE(), IluCallException, "yyy-2");
            return 0;
        }
        unhand(jcall)->jObjTracable = (HObject*) joi; /*make traceable*/
        ilu_ExitServer(cServer, ilu_IntroTypeOfCall(cCall));
        /*see ilu.c to figure out whether to use static or actual class...*/
        }
    return joi;
}


EXPORTJAVA long
xerox_ilu_IluCall_szOptional(JCALL jcall, long stat)
{
    ilu_cardinal sz; 
    ilu_boolean istat = (stat != 0);
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfOptional(cCall, istat, &ioerrs);
    ERRCHECK_1
    return sz;
}

EXPORTJAVA void
xerox_ilu_IluCall_outOptional(JCALL jcall, long stat)
{
    ilu_boolean istat = (stat != 0);
    PROLOG_V(CALL_OUT)
    ilu_OutputOptional(cCall, istat, &ioerrs);
    ERRCHECK_V
}

EXPORTJAVA long
xerox_ilu_IluCall_inOptional(JCALL jcall)
{
    ilu_boolean optionalStatus;
    PROLOG_1(CALL_IN)
    ilu_InputOptional(cCall, &optionalStatus, &ioerrs);
    ERRCHECK_1
    return (long) optionalStatus;
}


EXPORTJAVA long
xerox_ilu_IluCall_szArray(JCALL jcall, long elementCount)
{
    ilu_cardinal sz; 
    ilu_cardinal length = elementCount;
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfArray(cCall, length, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outArray(JCALL jcall, long elementCount)
{
    ilu_cardinal length = elementCount;
    PROLOG_V(CALL_OUT)
    ilu_OutputArray(cCall, length, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_inArray(JCALL jcall)
{
    PROLOG_V(CALL_IN)
    ilu_InputArray(cCall, &ioerrs);
    ERRCHECK_V
}

EXPORTJAVA void
xerox_ilu_IluCall_endArray(JCALL jcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_LIFE_V
    x = ilu_EndArray(cCall, &ioerrs);
    ERRCHECK_V
}

EXPORTJAVA long
xerox_ilu_IluCall_szRecord(JCALL jcall)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfRecord(cCall, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outRecord(JCALL jcall)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputRecord(cCall, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_inRecord(JCALL jcall)
{
    PROLOG_V(CALL_IN)
    ilu_InputRecord(cCall, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_endRecord(JCALL jcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_LIFE_V
    x = ilu_EndRecord(cCall, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_szSequence(JCALL jcall, long leng, long limit)
{
    ilu_cardinal sz; 
    ilu_cardinal len = leng;
    ilu_cardinal lim = limit;
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfSequence(cCall, len, lim, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outSequence(JCALL jcall, long leng, long limit)
{
    ilu_cardinal sz; 
    ilu_cardinal len = leng;
    ilu_cardinal lim = limit;
    PROLOG_V(CALL_OUT)
    ilu_OutputSequence(cCall, len, lim, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_outSequenceMark(JCALL jcall, long extent)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputSequenceMark(cCall, (ilu_cardinal) extent, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inSequence(JCALL jcall, long limit)
{
    ilu_cardinal len;
    ilu_cardinal lim = limit;
    PROLOG_1(CALL_IN)
    ilu_InputSequence(cCall, &len, lim, &ioerrs);
    ERRCHECK_1
    return (long) len;
}


EXPORTJAVA void
xerox_ilu_IluCall_inSequenceMark(JCALL jcall, long extent)
{
    PROLOG_V(CALL_IN)
    ilu_InputSequenceMark(cCall, (ilu_cardinal) extent, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_endSequence(JCALL jcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_LIFE_V
    x = ilu_EndSequence(cCall, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_szUnion(JCALL jcall, long discrim, long discrimSize)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfUnion(cCall, (ilu_cardinal) discrim, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outUnion(JCALL jcall, long discrim, long discrimSize)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputUnion(cCall, (ilu_cardinal) discrim, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inUnion(JCALL jcall, long discrimSize)
{
    ilu_cardinal discriminator;
    PROLOG_1(CALL_IN)
    ilu_InputUnion(cCall, &discriminator, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK_1
    return (long) discriminator;
}


EXPORTJAVA void
xerox_ilu_IluCall_endUnion(JCALL jcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    EXPECT_LIFE_V
    x = ilu_EndUnion(cCall, &ioerrs);
    ERRCHECK_V
}


#define TOIluInt64(i) ( * (ilu_longinteger *) &(i))
#define TOJavaIC64(i) ( * (int64_t *) &(i))
#define TOIluCard64(i) ( * (ilu_longcardinal *) &(i))

EXPORTJAVA long
xerox_ilu_IluCall_szInt64(JCALL jcall, int64_t ival)
{
    ilu_cardinal sz; 
    ilu_longinteger i = TOIluInt64(ival);
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfLongInteger(cCall, i, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt64(JCALL jcall, int64_t ival)
{
    ilu_longinteger i = TOIluInt64(ival);
    PROLOG_V(CALL_OUT)
    ilu_OutputLongInteger(cCall, i, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA int64_t
xerox_ilu_IluCall_inInt64(JCALL jcall)
{
    ilu_longinteger value;
    PROLOG_1(CALL_IN)
    ilu_InputLongInteger(cCall, &value, &ioerrs);
    ERRCHECK_1
    return TOJavaIC64(value);
}


EXPORTJAVA long
xerox_ilu_IluCall_szCard64(JCALL jcall, int64_t v)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfLongCardinal(cCall, TOIluCard64(v), &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard64(JCALL jcall, int64_t v)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputLongCardinal(cCall, TOIluCard64(v), &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA int64_t
xerox_ilu_IluCall_inCard64(JCALL jcall)
{
    ilu_longcardinal value;
    PROLOG_1(CALL_IN)
    ilu_InputLongCardinal(cCall, &value, &ioerrs);
    ERRCHECK_1
    return TOJavaIC64(value);
}


EXPORTJAVA long
xerox_ilu_IluCall_szReal64(JCALL jcall, double val)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfReal(cCall, val, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal64(JCALL jcall, double val)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputReal(cCall, val, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA double
xerox_ilu_IluCall_inReal64(JCALL jcall)
{
    double value;
    PROLOG_1(CALL_IN)
    ilu_InputReal(cCall, &value, &ioerrs);
    ERRCHECK_1
    return value;
}

EXPORTJAVA long
xerox_ilu_IluCall_szReal128Buff(JCALL jcall, HArrayOfByte * lr)
{
    char *buf;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (lr==0  | obj_length(lr)!=sizeof(ilu_longreal)) {
        SignalError(EE(), IluCallException, "xxx_xxx101");
        return 0;
    }
    buf = (char*) unhand(lr)->body;
    /* what about endian-ness? */
    sz = ilu_SizeOfLongReal(cCall, * (ilu_longreal*) buf, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal128Buff(JCALL jcall, HArrayOfByte * lr)
{
    char *buf;
    PROLOG_V(CALL_OUT)
    if (lr==0  | obj_length(lr)!=sizeof(ilu_longreal)) {
        SignalError(EE(), IluCallException, "xxx_xxx103");
        return;
    }
    buf = (char*) unhand(lr)->body;
    /* what about endian-ness? */
    ilu_OutputLongReal(cCall, * (ilu_longreal*) buf, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA void
xerox_ilu_IluCall_inReal128Buff(JCALL jcall, HArrayOfByte * lr)
{
    char *buf;
    PROLOG_V(CALL_IN)
    if (lr==0  | obj_length(lr)!=sizeof(ilu_longreal)) {
        SignalError(EE(), IluCallException, "xxx_xxx104");
        return;
    };
    buf = (char*) unhand(lr)->body;
    /* what about endian-ness? */
    ilu_InputLongReal(cCall, (ilu_longreal*) &buf, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_szReal32(JCALL jcall, float f)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfShortReal(cCall, f, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal32(JCALL jcall, float f)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputShortReal(cCall, f, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA float
xerox_ilu_IluCall_inReal32(JCALL jcall)
{
    float f;
    PROLOG_1(CALL_IN)
    ilu_InputShortReal(cCall, &f, &ioerrs);
    ERRCHECK_1
    return f;
}

EXPORTJAVA long
xerox_ilu_IluCall_szEnum(JCALL jcall, long v)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfEnum(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outEnum(JCALL jcall, long v)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputEnum(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inEnum(JCALL jcall)
{
    ilu_shortcardinal v;
    PROLOG_1(CALL_IN)
    ilu_InputEnum(cCall, &v, &ioerrs);
    ERRCHECK_1
    return v;
}


EXPORTJAVA long
xerox_ilu_IluCall_szCard16(JCALL jcall, short v)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfShortCardinal(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard16(JCALL jcall, short v)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputShortCardinal(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA short
xerox_ilu_IluCall_inCard16(JCALL jcall)
{
    ilu_shortcardinal v;
    PROLOG_1(CALL_IN)
    ilu_InputShortCardinal(cCall, &v, &ioerrs);
    ERRCHECK_1
    return (short) v;
}

EXPORTJAVA long
xerox_ilu_IluCall_szCard32(JCALL jcall, long v)
{
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    sz = ilu_SizeOfCardinal(cCall, (ilu_cardinal) v, &ioerrs);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard32(JCALL jcall, long v)
{
    PROLOG_V(CALL_OUT)
    ilu_OutputCardinal(cCall, (ilu_cardinal) v, &ioerrs);
    ERRCHECK_V
}


EXPORTJAVA long
xerox_ilu_IluCall_inCard32(JCALL jcall)
{
    ilu_cardinal v;
    PROLOG_1(CALL_IN)
    ilu_InputCardinal(cCall, &v, &ioerrs);
    ERRCHECK_1
    return (long) v;
}



EXPORTJAVA long
xerox_ilu_IluCall_szString8(JCALL jcall, Hjava_lang_String * s, long limit)
    /* Variable-length array of short character (String) */
{
    int len = javaStringLength(s);
    char * cString;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if ((len > limit) && (limit != 0)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return 0;
    }
    cString = allocCString(s); 
    	/*(javaString.h) incl. unicode to char conversion */
    sz = ilu_SizeOfString(cCall, cString, 
    	(ilu_cardinal) len, (ilu_cardinal) limit, &ioerrs);
    java_free(cString);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outString8x(JCALL jcall, Hjava_lang_String * s, long limit)
    /* Variable-length array of short character (String) */
{
    int len = javaStringLength(s);
    char * cString;
    PROLOG_V(CALL_OUT)
    if ((len > limit) && (limit != 0)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return;
    }
    cString = allocCString(s); 
    	/*(javaString.h) incl. unicode to char conversion */
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ outString8x len=%d  s=<%s> \n", len, cString);
    }
    ilu_OutputString(cCall, cString, (ilu_cardinal) len, 
    	(ilu_cardinal) limit, &ioerrs);
    java_free(cString);
    ERRCHECK_V
}


EXPORTJAVA Hjava_lang_String *
xerox_ilu_IluCall_inString8(JCALL jcall, long limit)
{
    Hjava_lang_String * s;
    char * cString = 0;
    ilu_cardinal len = 0;
    PROLOG_1(CALL_IN)
    ilu_InputString(cCall, &cString, &len, (ilu_cardinal) limit, &ioerrs);
    if (ILU_ERRNOK(ioerrs)) {
        ilu_free(cString);
        OptionalPrintE(ioerrs, "inString8");\
        _iluJava_IluErrorToJavaException(ioerrs, "xxx-55");
        return 0;
    }
    if (_ilujava_callsFlag>0) {
        ilu_DebugPrintf("$ inString8 len=%d  s=<%s> \n", len, cString);
    }
    s = makeJavaString(cString, (int) len);
    ilu_free(cString);
    return s;
}


EXPORTJAVA long
xerox_ilu_IluCall_szChar8Array_old(JCALL jcall, Hjava_lang_String * s, long limit)
{
    char * buff = 0;  
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (s==0 | limit==0 ) {
        OptionalPrintE(ioerrs, "szChar8Array");\
        SignalError(EE(), IluCallException, "xxx-59");
        return 0;
    }
    buff = ilu_malloc(limit);
    IluJava_JString_toC8(s, buff, limit);
    sz = ilu_SizeOfStringVec(cCall, buff, (ilu_cardinal) limit, &ioerrs);
    ilu_free(buff);
    ERRCHECK_1
    return sz;
} /*szChar8Array_old*/


EXPORTJAVA void
xerox_ilu_IluCall_outChar8Array_old(JCALL jcall, Hjava_lang_String * s, long limit)
{
    char * buff = 0;  
    ilu_cardinal sz; 
    PROLOG_V(CALL_OUT)
    if (s==0 | limit==0 ) {
        OptionalPrintE(ioerrs, "outChar8Array");\
        SignalError(EE(), IluCallException, "xxx-59");
        return;
    }
    buff = ilu_malloc(limit);
    IluJava_JString_toC8(s, buff, limit);
    ilu_OutputStringVec(cCall, buff, (ilu_cardinal) limit, &ioerrs);
    java_free(buff);
    ERRCHECK_V
} /*outChar8Array_old*/


EXPORTJAVA Hjava_lang_String *
xerox_ilu_IluCall_inChar8Array_old(JCALL jcall, long limit)
{
    Hjava_lang_String * s;
    char * cString = 0;
    ilu_cardinal len;
    PROLOG_1(CALL_IN)
    ilu_InputStringVec(cCall, &cString, (ilu_cardinal) limit, &ioerrs);
    if (ILU_ERRNOK(ioerrs)) {
        ilu_free(cString);
        OptionalPrintE(ioerrs, "inChar8Array"); \
        _iluJava_IluErrorToJavaException(ioerrs, "xxx-58");
        return 0;
    }
    s = makeJavaString(cString, (int) limit);
    ilu_free(cString);
    return s;
} /*inChar8Array_old*/


EXPORTJAVA long
xerox_ilu_IluCall_szBytesSx(JCALL jcall, 
    HArrayOfByte * bytes, 
    long start, long len, long limit
    )
    /* Variable-length array of byte */
{
    char *data;
    long dataLen;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if ((len > limit) && (len != 0)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return 0;
    }
    if (bytes == 0) {
        data = (char *) &ilujava_nullBuffer;
	dataLen = 0;
	start  = 0;
    } else {
        data = (char *) unhand(bytes)->body;
        dataLen = obj_length(bytes);
        
    }
    if (start>dataLen || len>(dataLen-start)) {
        /* data structure too small to contain len bytes */
        SignalError(EE(), IluCallException, "xxx-55");
        return 0;
    }
    sz = ilu_SizeOfBytes(cCall, (ilu_bytes) data+start, 
    	(ilu_cardinal)len, (ilu_cardinal)limit, 
    	&ioerrs
    	);
    ERRCHECK_1
    return sz;
}



EXPORTJAVA void
xerox_ilu_IluCall_outBytesSx(JCALL jcall, 
    HArrayOfByte * bytes, 
    long start, long len, long limit
    )
    /* Variable-length array of byte */
{
    char *data;
    long dataLen;
    ilu_cardinal sz; 
    PROLOG_V(CALL_OUT)
    if ((len > limit) && (limit != 0)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return;
    }
    if (bytes == 0) {
        data = (char *) &ilujava_nullBuffer;
	dataLen = 0;
	start  = 0;
    } else {
        data = (char *) unhand(bytes)->body;
        dataLen = obj_length(bytes);
        
    }
    if (start>dataLen || len>(dataLen-start)) {
        /* data structure too small to contain len bytes */
        SignalError(EE(), IluCallException, "xxx-55");
        return;
    }
    ilu_OutputBytes(cCall, (ilu_bytes) data+start, 
    	(ilu_cardinal)len, (ilu_cardinal)limit, 
    	&ioerrs
    	);
    ERRCHECK_V
}


EXPORTJAVA HArrayOfByte *
xerox_ilu_IluCall_inBytesS(JCALL jcall, long limit)
    /* Variable-length array of byte */
{
    HArrayOfByte * bytes = 0;
    ilu_bytes data = 0;
    ilu_cardinal len;
    PROLOG_1(CALL_IN)
    ilu_InputBytes(cCall, &data, 
        &len, (ilu_cardinal) limit, 
    	&ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        OptionalPrintE(ioerrs, "IluCall");
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx107"); 
        return 0; 
    }
    if (data) {
        bytes = IluJava_JAoB_from8((char *)data, len);
        ilu_free(data);
    }
    return bytes;
}



EXPORTJAVA long
xerox_ilu_IluCall_szBytesAx(JCALL jcall, 
    HArrayOfByte * bytes, 
    long start, long limit
    )
    /* Fixed-length array of byte */
{
    char *data;
    long dataLen;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (bytes == 0) {
	if (limit>0) {
	    ilujava_NullPointerException();
	    return 0;
	}
	dataLen = 0;
	data = (char *) &ilujava_nullBuffer;
    } else {
        dataLen = obj_length(bytes);
        data = (char *) unhand(bytes)->body;
    }
    if (start>dataLen || limit>(dataLen-start)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return 0;
    }
    sz = ilu_SizeOfOpaque(cCall, (ilu_opaque) data+start, 
    	(ilu_cardinal)limit,  
    	&ioerrs
    	);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outBytesAx(JCALL jcall, 
    HArrayOfByte * bytes, 
    long start, long limit
    )
    /* Fixed-length array of byte */
{
    char *data;
    long dataLen;
    ilu_cardinal sz; 
    PROLOG_V(CALL_OUT)
    if (bytes == 0) {
	if (limit>0) {
	    ilujava_NullPointerException();
	    return;
	}
	dataLen = 0;
	data = (char *) &ilujava_nullBuffer;
    } else {
        dataLen = obj_length(bytes);
        data = (char *) unhand(bytes)->body;
    }
    if (start>dataLen || limit>(dataLen-start)) {
        SignalError(EE(), IluCallException, "xxx-55");
        return;
    }
    ilu_OutputOpaque(cCall, (ilu_opaque) data+start, 
    	(ilu_cardinal) limit, 
    	&ioerrs
    	);
    ERRCHECK_V
}


EXPORTJAVA HArrayOfByte *
xerox_ilu_IluCall_inBytesA(JCALL jcall, long len)
    /* Fixed-length array of byte */
{
    HArrayOfByte * bytes;
    ilu_opaque data = 0;
    PROLOG_1(CALL_IN)
    ilu_InputOpaque(cCall, &data, 
        (ilu_cardinal) len, &ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        OptionalPrintE(ioerrs, "IluCall");
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx108"); 
        return 0; 
    }
    bytes = IluJava_JAoB_from8((char *)data, len);
    ilu_free(data);
    return bytes;
}


EXPORTJAVA long
xerox_ilu_IluCall_szString16(JCALL jcall, 
    struct Hjava_lang_String * jstring, 
    long limit
    )
    /* Variable-length array of 16 bit character (WString) */
{
    ilu_wstring cw;
    long len;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (jstring == 0) {
	len = 0;
    } else {
        len = javaStringLength(jstring);
    }
    if ((len>limit) && (limit!=0)) len = limit;
    cw = ilu_malloc(len*sizeof(ilu_character)+2/*safety*/);
    IluJava_JString_toC16(jstring, (char *) cw, len);
    sz = ilu_SizeOfWString(cCall, cw, 
    	(ilu_cardinal) len, (ilu_cardinal) limit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outString16x(JCALL jcall, 
    struct Hjava_lang_String * jstring, 
    long limit
    )
    /* Variable-length array of 16 bit character (WString) */
{
    ilu_wstring cw;
    long len;
    PROLOG_V(CALL_OUT)
    if (jstring == 0) jstring = ilujava_jEmptyString;
    len = javaStringLength(jstring);
    if ((len>limit) && (limit!=0)) len = limit;
    cw = ilu_malloc(len*sizeof(ilu_character)+2/*safety*/);
    IluJava_JString_toC16(jstring, (char *) cw, len);
    ilu_OutputWString(cCall, cw, 
    	(ilu_cardinal)len, (ilu_cardinal)limit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_V
}


EXPORTJAVA struct Hjava_lang_String *
xerox_ilu_IluCall_inString16(JCALL jcall, long limit)
    /* Variable-length array of character (WString) */
{
    struct Hjava_lang_String * jstring;
    ilu_wstring data;
    ilu_cardinal len;
    PROLOG_1(CALL_IN)
    ilu_InputWString(cCall, &data, 
        &len, (ilu_cardinal) limit, 
    	&ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        OptionalPrintE(ioerrs, "IluCall");
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx109"); 
        return 0; 
    }
    jstring = IluJava_JString_from16((char *)data, len);
    ilu_free(data);
    return jstring;
}



EXPORTJAVA long
xerox_ilu_IluCall_szChar16Array_old(JCALL jcall, 
    struct Hjava_lang_String * jstring, 
    long limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    long len;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (jstring == 0) jstring = ilujava_jEmptyString;
    cw = ilu_malloc(limit*sizeof(ilu_character)+2/*safety*/);
    /*javaString.h*/javaString2unicode(jstring, (unicode *) cw, len);
    sz = ilu_SizeOfWStringVec(cCall, cw, 
    	(ilu_cardinal)limit,  
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_1
    return sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outChar16Array_old(JCALL jcall, 
    struct Hjava_lang_String * jstring, 
    long limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    long len;
    ilu_cardinal sz; 
    PROLOG_V(CALL_OUT)
    if (jstring == 0) jstring = ilujava_jEmptyString;
    cw = ilu_malloc(limit*sizeof(ilu_character)+2/*safety*/);
    /*javaString.h*/javaString2unicode(jstring, (unicode *) cw, len);
    ilu_OutputWStringVec(cCall, cw, 
    	(ilu_cardinal) limit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_V
}


EXPORTJAVA struct Hjava_lang_String *
xerox_ilu_IluCall_inChar16Array_old(JCALL jcall, long len)
    /* Fix-length array of character (WStringVec) */
{
    struct Hjava_lang_String * jstring;
    ilu_wstring data = 0;
    PROLOG_1(CALL_IN)
    ilu_InputWStringVec(cCall, &data, 
        (ilu_cardinal)len, &ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        OptionalPrintE(ioerrs, "IluCall");
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx110"); 
        return 0; 
    }
    jstring = IluJava_JString_from16((char *)data, len);
    ilu_free(data);
    return jstring;
}


EXPORTJAVA long
xerox_ilu_IluCall_szChar8Array(JCALL jcall, HArrayOfChar * ja, long lim)
    /* Fix-length array of short character (StringVec) */
{
    long len;
    char * cbuffer;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (ja == 0) {
	if (lim>0) {
	    ilujava_NullPointerException();
	    return 0;
	}
	len = 0;
	cbuffer = ilu_malloc(2); cbuffer[0] = 0;
    } else {
        len = obj_length(ja);
        if (len<lim) {
            OptionalPrintE(ioerrs, "szChar8Array");\
            SignalError(EE(), IluCallException, "xxx-59");
            return 0;
        }
        cbuffer = ilu_malloc(len + 2 /* 0C + safety */ );
        IluJava_JAoC_toC8(ja, cbuffer, len);
    }
    sz = ilu_SizeOfStringVec(cCall, (ilu_string) cbuffer, (ilu_cardinal) lim, &ioerrs);
    ilu_free(cbuffer);
    ERRCHECK_1
    return sz;
} /*szChar8Array*/


EXPORTJAVA void
xerox_ilu_IluCall_outChar8Arrayx(JCALL jcall, HArrayOfChar * ja, long lim)
{
    char * cbuffer;
    ilu_cardinal sz; 
    long len; 
    PROLOG_V(CALL_OUT)
    if (ja == 0) {
	if (lim>0) {
	    ilujava_NullPointerException();
	    return;
	}
	len = 0;
	cbuffer = ilu_malloc(2); cbuffer[0] = 0;
    } else {
        len = obj_length(ja);
        if (len<lim) {
            OptionalPrintE(ioerrs, "outChar8Arrayx");\
            SignalError(EE(), IluCallException, "xxx-59");
            return;
        }
        cbuffer = ilu_malloc(len + 2 /* 0C + safety */ );
        IluJava_JAoC_toC8(ja, cbuffer, len);
    }
    ilu_OutputStringVec(cCall, (ilu_string) cbuffer, (ilu_cardinal) lim, &ioerrs);
    ilu_free(cbuffer);
    ERRCHECK_V
} /*outChar8Arrayx*/


EXPORTJAVA HArrayOfChar *
xerox_ilu_IluCall_inChar8Array(JCALL jcall, long lim)
{
    HArrayOfChar * ja;
    char * cString;
    PROLOG_1(CALL_IN)
    ilu_InputStringVec(cCall, &cString, (ilu_cardinal) lim, &ioerrs);
    if (ILU_ERRNOK(ioerrs)) {
        ilu_free(cString);
        OptionalPrintE(ioerrs, "inChar8Array"); \
        _iluJava_IluErrorToJavaException(ioerrs, "xxx-58");
        return 0;
    }
    ja = IluJava_JAoC_from8(cString, lim);
    ilu_free(cString);
    return ja;
} /*inChar8Array*/


EXPORTJAVA long
xerox_ilu_IluCall_szChar16Array(JCALL jcall, 
    HArrayOfChar * ja, 
    long limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    long len;
    ilu_cardinal sz; 
    PROLOG_1(CALL_SZ)
    if (ja == 0) {
	if (limit>0) {
	    ilujava_NullPointerException();
	    return 0;
	}
	len = 0;
	cw = ilu_malloc(4); cw[0] = 0; cw[1] = 0;
    } else {
        len = obj_length(ja);
        if (len<limit) {
            OptionalPrintE(ioerrs, "szChar8Array");\
            SignalError(EE(), IluCallException, "xxx-59");
            return 0;
        }
        cw = ilu_malloc(limit*sizeof(ilu_character)+4 /*safety+terminator*/ );
        IluJava_JAoC_toC16(ja, (char *)cw, len);
    }
    sz = ilu_SizeOfWStringVec(cCall, cw, 
    	(ilu_cardinal)limit,  
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_1
    return sz;
} /*szChar16Array*/


EXPORTJAVA void
xerox_ilu_IluCall_outChar16Arrayx(JCALL jcall, 
    HArrayOfChar * ja, 
    long limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    long len;
    ilu_cardinal sz; 
    PROLOG_V(CALL_OUT)
    if (ja == 0) {
	if (limit>0) {
	    ilujava_NullPointerException();
	    return;
	}
	len = 0;
	cw = ilu_malloc(4); cw[0] = 0; cw[1] = 0;
    } else {
        len = obj_length(ja);
        if (len<limit) {
            OptionalPrintE(ioerrs, "outChar16Arrayx");\
            SignalError(EE(), IluCallException, "xxx-59");
            return;
        }
        cw = ilu_malloc(limit*sizeof(ilu_character)+4 /*safety+terminator*/ );
        IluJava_JAoC_toC16(ja, (char *)cw, len);
    }
    ilu_OutputWStringVec(cCall, cw, 
    	(ilu_cardinal) limit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK_V
} /*outChar16Array*/
 

EXPORTJAVA HArrayOfChar *
xerox_ilu_IluCall_inChar16Array(JCALL jcall, long len)
    /* Fix-length array of character (WStringVec) */
{
    HArrayOfChar * ja;
    ilu_wstring data = 0;
    PROLOG_1(CALL_IN)
    ilu_InputWStringVec(cCall, &data, 
        (ilu_cardinal)len, &ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        OptionalPrintE(ioerrs, "IluCall");
        _iluJava_IluErrorToJavaException(ioerrs, "xxx_xxx110"); 
        return 0; 
    }
    ja = IluJava_JAoC_from16((char *)data, len);
    ilu_free(data);
    return ja;
} /*inChar16Array*/



EXPORTJAVA void
xerox_ilu_IluCall_nativeMarkCallFailure(JCALL jcall, long protocolException)
{
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    if (cCall) {
        cCall->ca_pe = (ilu_ProtocolException) protocolException;
    }
} /*nativeMarkCallFailure*/




/* the end */
