/* IluJava_IluClassRep.c */
/* Chris Jacobi, December 20, 1996 10:37 am PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluClassRep.c,v 1.7 1996/12/20 18:42:03 jacobi Exp $ */
 
/*
 * C side of IluClassRep.java
 * also accessing IluMethodRep.java
 */


#include <StubPreamble.h>
#include <javaString.h>
#include "xerox_ilu_IluClassRep.h"
#include "xerox_ilu_IluMethodRep.h"
#include "xerox_ilu_IluExceptionRep.h"

#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"

#define JToCStringKeep(obj, field)	allocCString( unhand(obj) -> field )
    /* Access a String field of an object, convert it to an
     * ilu_string and permanently allocate the ilu_string.
     * (javaString.h)allocCString does unicode to char conversion
     */
     
#define JToCStringGC(obj, field)	makeCString( unhand(obj) -> field )
    /* Access a String field of an object, convert it to an
     * ilu_string and make it subject to java garbage collection.
     * (javaString.h)allocCString does unicode to char conversion
     */
     
#define JToBoolean(obj, field)	(( unhand(obj) -> field ) != 0)
    /* Access a field of an object as a boolean */

/* Compare to /tilde/jacobi/ilu/runtime/python/iluclobject.c */

static void
freeSuperClasses(ilu_string *cSuperClassArray, int sccnt)
    /* specifically used in reallyFinish */
{
    if (cSuperClassArray) {
        int sx;
        for (sx = 0; sx < sccnt; sx = sx+1) {
            /* use java's free because strings were allocated by java */
            java_free(cSuperClassArray[sx]);
        }
        /* ilu's free as this was allocated with ilu */
        ilu_free(cSuperClassArray);
    }
}

EXPORTJAVA void 
xerox_ilu_IluClassRep_reallyFinish(
		struct Hxerox_ilu_IluClassRep * jobj
		)
{
    int mx, sx, ex; /*loop indices*/
    ilu_Error err = ILU_INIT_NO_ERR;
    ilu_Error err2 = ILU_INIT_NO_ERR;
    ilu_Class cIluClass;
    ilu_string *cSuperClassArray;
    ilu_string className;
    /*struct Hxerox_ilu_IluClassRep * jSuper;*/
    struct HArrayOfObject * jSuper;
    struct HArrayOfObject * jMethods;
    int sccnt = 0;
    ilu_string cSingleton;
    
    className = JToCStringKeep(jobj, jIluClassName);
    jMethods = unhand(jobj)->jMethods;
    
    /* superclass array */
    jSuper = unhand(jobj)->jIluSuperClasses;
    if (jSuper) sccnt = obj_length(jSuper);
    if (_ilujava_definitionsFlag>0) {
        ilu_DebugPrintf("$ DefineClass %s sccnt: %d\n", className, sccnt);
    }
    cSuperClassArray = ilu_must_malloc(sccnt * sizeof(ilu_string));
    for (sx = 0; sx < sccnt; sx++) 
    {
        ilu_string sClassId; 
        struct Hxerox_ilu_IluClassRep* jThisSuper;
        jThisSuper = (struct Hxerox_ilu_IluClassRep*) unhand(jSuper)->body[sx];
        sClassId = /*javaString.h*/ allocCString(
            (Hjava_lang_String *) unhand(jThisSuper)->uuid
            ); /* C strings in C heap ! */
        cSuperClassArray[sx] = sClassId;
        if (_ilujava_definitionsFlag>0) {
           ilu_DebugPrintf("$ Superclass id [%s] \n", sClassId);
        }
    }

    /* Start the kernel registrations */
    ilu_EnterMutex(ilu_otmu, &err);
    if (ILU_ERRNOK(err)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _iluJava_IluErrorToJavaException(err, completedNo "reallyFinish0");
        return;
    }
    cSingleton = ( unhand(jobj)->singleton==0 ? 
    	0 : JToCStringKeep(jobj, singleton)
    	);
    cIluClass = ilu_DefineObjectType(
        className,
	JToCStringKeep(jobj, brand),
	JToCStringKeep(jobj, uuid),
    	cSingleton,
    	JToBoolean(jobj, optional),
    	JToBoolean(jobj, collectible),
    	JToCStringKeep(jobj, authentication),
    	(ilu_cardinal) unhand(jobj)->method_count, 
    	(ilu_cardinal) sccnt, 
	cSuperClassArray, 
	&err
    	);
    Access(jobj, yIluClass, ilu_Class) = cIluClass;
    if (ILU_ERRNOK(err)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _iluJava_IluErrorToJavaException(err, "reallyFinish1");
        return;
    }

    /* loop to register methods */
    for (mx = 0; mx < unhand(jobj)->method_count; mx = mx+1) 
    {
        int exceptCnt;
        ilu_Exception *cExceptionVec;
        ilu_Method cIluMethod;
        struct Hxerox_ilu_IluMethodRep* jIluMethodRep
        	= (struct Hxerox_ilu_IluMethodRep*) unhand(jMethods)->body[mx];
        struct HArrayOfObject * jExceptions 
        	= unhand(jIluMethodRep)->jExceptions;
        
        /* build exception array for this method */
        if (jExceptions) {
            exceptCnt = obj_length(jExceptions);
            cExceptionVec = 
        	ilu_must_malloc(exceptCnt * sizeof(ilu_Exception));
            for (ex = 0; ex < exceptCnt; ex = ex+1) 
            {
                struct Hxerox_ilu_IluExceptionRep* jExc;
                jExc = (struct Hxerox_ilu_IluExceptionRep*) unhand(jExceptions)->body[ex];
                cExceptionVec[ex] = Access(
                	jExc,
                	yIluException, /*field*/
                	ilu_Exception  /*type*/
                	);
                	/* check against nil in java */
            }
        } else {
            exceptCnt = 0;
            cExceptionVec = 0;
        }
        /* register method */
        cIluMethod = ilu_DefineMethod(
        	cIluClass,
        	(ilu_cardinal) mx,  /* i */
        	JToCStringGC(jIluMethodRep, jIluMethodName), 
        	unhand(jIluMethodRep)->id,
                JToBoolean(jIluMethodRep, cacheable),
		JToBoolean(jIluMethodRep, asynchronous),
		(ilu_cardinal) exceptCnt, 
		cExceptionVec, 
		&err
        	);
        if (ILU_ERRNOK(err)) {
            ilu_free(cExceptionVec);
            freeSuperClasses(cSuperClassArray, sccnt);
            _iluJava_IluErrorToJavaException(err, "reallyFinish2");
            return;
        }
        Access(jIluMethodRep, yIluMethod, ilu_Method) = cIluMethod;
        
        /* 
         * TERRIBLE LOOPHOLE 
         * Assigns a java object into a field declared as procedure 
         */
        ilu_SetMethodStubProc(cIluMethod, (ilu_StubProc) jIluMethodRep);
        ilu_free(cExceptionVec);
    }
    
    _iluJava_registerJavaClass(cIluClass, jobj); /*IluJava.h*/
    
    ilu_ObjectTypeDefined(cIluClass, &err2);

    ilu_ExitMutex(ilu_otmu, ilu_FALSE, &err);
    if (ILU_ERRNOK(err)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _iluJava_IluErrorToJavaException(err, "reallyFinish99");
        return;
    }
    if (ILU_ERRNOK(err2)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _iluJava_IluErrorToJavaException(err2, "reallyFinish98");
        return;
    }
    /* Done with kernel registrations */

    /* Clean up */
    freeSuperClasses(cSuperClassArray, sccnt);
    
} /* reallyFinish */
