/* IluJava_IluConnOps.c */
/* Chris Jacobi, December 22, 1996 3:12 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluConnOps.c,v 1.8 1996/12/22 23:25:00 jacobi Exp $ */
 
/* 
 * C side for IluSurrogateConnection.java, IluServerConnection.java
 */
  
#include <StubPreamble.h>
#include <interpreter.h>

#include "xerox_ilu_IluSurrogateConnection.h"
#include "xerox_ilu_IluServerConnection.h"
#include "xerox_ilu_IluPort.h"
#include "xerox_ilu_IluCall.h"
#include "xerox_ilu_IluClassRep.h"

#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"



static struct Hxerox_ilu_IluSurrogateConnection * jProtoCFInstance = 0;
    /* Some good instance from initialization because we failed 
     * calling static methods from native code.
     */


#define PORTWATCHER_LOOP 0
#define PORTWATCHER_FORK 1
#define PORTWATCHER_CLOSE 2
#define PORTWATCHER_DONE 3

EXPORTJAVA long 
xerox_ilu_IluPort_nativeWatchPort(
    	struct Hxerox_ilu_IluPort * jPort
    	)
{
    ilu_boolean     ret;
    ilu_boolean     closed = 0;
    ILU_ERRS((IoErrs, bad_locks, broken_locks, no_resources, interrupted)) err;
    ilu_Connection  conn;
    ilu_Port cPort = Access(jPort, yPort, ilu_Port);
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluConnOps_nativeWatchPort\n");
        ilu_DebugPrintf("   jPort: %x, cPort: %x \n", (int)jPort, (int)cPort);
    }
    ILU_CLER(err);
    ret = ilu_WaitForPortConnectionRequest(cPort, &err);
    if (ret==0) {
        Access(jPort, yConn, ilu_Connection) = 0;
        ILU_ERR_SWITCH(err) {
            ILU_SUCCESS_CASE
	  	/* port is closed */
	  	if (_ilujava_pctFlag>0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort port closed\n");
                }
                return PORTWATCHER_DONE;
            ILU_ERR_CASE(interrupted, v)
	  	/* interrupted; retry */
	  	if (_ilujava_pctFlag>0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort interrupted\n");
                }
                return PORTWATCHER_LOOP;
            ILU_ERR_ELSE
	  	/* error; give up */
                if (_ilujava_pctFlag>0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort ERR1\n");
                }
                return PORTWATCHER_DONE;
        } ILU_ERR_ENDSWITCH;
    }
    ILU_CLER(err);
    conn = ilu_HandleNewConnection(cPort, &closed, &err);
    if (closed) {
        if (_ilujava_pctFlag>0) {
            ilu_DebugPrintf("$ .._nativeWatchPort: closed \n");
        }
        Access(jPort, yConn, ilu_Connection) = 0;
        return PORTWATCHER_DONE;
    }
    if (conn==0) {
        ILU_ERR_SWITCH(err) {
            ILU_SUCCESS_CASE
	  	/* spurious wakeup; retry */
	  	if (_ilujava_pctFlag>0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort sprs\n");
                }
                Access(jPort, yConn, ilu_Connection) = 0;
                return PORTWATCHER_LOOP;
            ILU_ERR_ELSE
	  	/* error; give up */
                if (_ilujava_pctFlag>0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort ERR2\n");
                }
                if (Access(jPort, yConn, int) == -1) {
                    /*prevent infinite loop*/
                    return PORTWATCHER_DONE;
                }
                Access(jPort, yConn, int) = -1;
                return PORTWATCHER_CLOSE;
        } ILU_ERR_ENDSWITCH;
    } else {
        if (_ilujava_pctFlag>0) {
            ilu_DebugPrintf("$ .._nativeWatchPort: scss:<%p>\n",  conn);
        }
        Access(jPort, yConn, ilu_Connection) = conn;
        /* Requirement: ilu_DoneServingConnection must called eventually.
         * Loop reading requests will do so.
         */
        return PORTWATCHER_FORK;
    }
    /* notreached */
} /* nativeWatchPort */


EXPORTJAVA long 
xerox_ilu_IluSurrogateConnection_nativeOtherNewConnection(
	struct Hxerox_ilu_IluSurrogateConnection * jcf
	)
{
    ilu_Connection cConn; 
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluConnOps_nativeOtherNewConnection");
    }
    cConn = ilu_OtherNewConnection(&err);
    if (ILU_ERRNOK(err)) {
        _iluJava_IluErrorToJavaException(err, "OtherNewConnection");
    }
    Access(jcf, ySurrConn, ilu_Connection) = cConn;
    return (cConn != 0);
} /* nativeOtherNewConnection */


EXPORTJAVA void 
xerox_ilu_IluSurrogateConnection_nativeNewConnGetterForked(
	struct Hxerox_ilu_IluSurrogateConnection * jcf
	)
{
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluConnOps_nativeNewConnGetterForked");
    }
    jProtoCFInstance = jcf;
    ilu_NewConnectionGetterForked(&err);
    ILU_MUST_BE_SUCCESS(err);
} /* nativeNewConnGetterForked */


EXPORTJAVA void 
xerox_ilu_IluSurrogateConnection_nativeOutgoingConnectionThreadProc(
	struct Hxerox_ilu_IluSurrogateConnection * jcf
	)
/* Called through forked Java when ilu_StartCall did require  
 * forking a connection reader handling one surrogate connection.
 * Uses "this.ySurrCon"
 */ 
{
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    ilu_Connection cConn;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluConnOps_nativeRun jcf %x \n", (int) jcf);
    }
    cConn = Access(jcf, ySurrConn, ilu_Connection);
    ilu_OutgoingConnectionThreadProc(cConn, &err);
    ILU_MUST_BE_SUCCESS(err);
} /* nativeOutgoingConnectionThreadProc */


EXPORTJAVA long
xerox_ilu_IluServerConnection_receiveRequest(
    	struct Hxerox_ilu_IluServerConnection * jcr,
    	struct Hxerox_ilu_IluCall * jcall
    	)
/* servicing requests for a true servers connection */
{
    ilu_boolean     initted;
    ilu_Call cCall = Access(jcall, yCall, ilu_Call);
    ilu_Class       cIluClass;
    ilu_Method      cIluMethod;
    ilu_RcvReqStat  stat;
    ilu_cardinal    SN;
    ILU_ERRS((bad_locks, IoErrs, bad_param, broken_locks, internal)) lerr;
    ilu_boolean     closed = ilu_FALSE;
    ilu_Connection cConn = Access(jcr, yConn, ilu_Connection);
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluConnOps_receiveRequest cConn %p \n", cConn);
    }
    if (unhand(jcall)->needFinish) {
        BUGPRINT("$ error with call");
    }
    unhand(jcall)->jMethodRep = 0;
    if (!ilu_BlockingWaitForInputOnConnection(cConn, ILU_NIL)) {
        /*connection should be abandoned*/
        if (_ilujava_pctFlag>0) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest abandon conn\n");
        }
        ilu_DoneServingConnection(cConn, &lerr);
        return 0;
    }
    if (_ilujava_pctFlag>1) {
        ilu_DebugPrintf("$ IluConnOps_receiveRequest 1 \n");
    }
    stat = ilu_ReceiveRequest(cCall, &initted, cConn, &cIluClass, &cIluMethod,
			    &SN, &lerr);
    if (stat == ilu_RcvReqStat_request) {
        /*initted is true*/
        if (_ilujava_pctFlag>1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest request \n");
        }
        unhand(jcall)->jMustFork = (long) ilu_ThreadPerRequest(cConn);
        unhand(jcall)->jClassRep = (struct Hxerox_ilu_IluClassRep *)
        	 _ilujava_findJavaClass(cIluClass);
        unhand(jcall)->jMethodRep = (struct Hxerox_ilu_IluMethodRep *) 
        	ilu_GetMethodStubProc(cIluMethod);
        unhand(jcall)->jNeedsSizing = (long) ilu_CallNeedsSizing(cCall);
        if (_ilujava_pctFlag>1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest got request\n");
        }
        unhand(jcall)->needFinish = 1; /* enforces ilu_FinishCall */
        return 1; /*keep looping*/
    } else if (stat == ilu_RcvReqStat_quit) {
        if (_ilujava_pctFlag>1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest quit \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        ilu_DoneServingConnection(cConn, &lerr);
        return 0; /*exit loop*/
    } else if (stat == ilu_RcvReqStat_noop) {
        if (_ilujava_pctFlag>1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest noop \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        return 1; /*keep looping*/
    } else {
        if (_ilujava_pctFlag>1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest other \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        return 1; /*keep looping*/
    }
} /* receiveRequest */


