/* IluJava_IluDebug.c */
/* Chris Jacobi, December 20, 1996 1:28 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluDebug.c,v 1.8 1996/12/20 23:37:39 jacobi Exp $ */
 
/*
 * Native impl for IluDebug.java, IluDebugStream.java, IluDebugHook.java
 * including a few native utilities...
 */


#include "IluJava.h"
#include <javaString.h>
#include "xerox_ilu_IluDebugHooks.h"
#include "xerox_ilu_IluDebug.h"
#include "xerox_ilu_IluDebugStream.h"
#include "xerox_ilu_IluRT0.h"
#include "xerox_ilu_IluOInt.h"
#include "xerox_ilu_IluClassRep.h"
#include "xerox_ilu_IluServer.h"

static ilu_boolean _ilujava_langRegistered = FALSE;
static ilu_LanguageIndex _ilujava_langId = 999;

ilu_LanguageIndex _ilujava_LangIdx(void)
{
    if (! _ilujava_langRegistered) {
        _ilujava_langId = ilu_RegisterLanguage("Java");
        _ilujava_langRegistered = ilu_TRUE;
    }
    return _ilujava_langId;
}


EXPORTLIMITEDVAR char * _iluJava_notUsed1;
EXPORTLIMITEDVAR char * _iluJava_notUsed2;
/* Used to fool C compiler into believing a pointer is still in
 * use:  Assign to this to preven C compiler from discarding 
 * pointer -> so Java garbage collector will find the reference on the 
 * stack...
 */


EXPORTLIMITED void
_iluJava_BUG(char * f)
{
    ilu_DebugPrintf("$** Ilu_Java BUG: %s \n", f);
    if (_ilujava_iluDebuggingFlag>1) {
        _iluJava_PANIC(f);
    }
    SignalError(EE(), "xerox/ilu/IluSomeSystemException", "**_iluJava_BUG");
}


EXPORTLIMITED void
_iluJava_PANIC(char * f)
{
    ilu_DebugPrintf("$** Ilu_Java PANIC: %s \n", f);
    _ilu_Assert(0, f);
}


/*special immutable instances from set up*/
static struct Hxerox_ilu_IluDebug * _ilujava_IluDebugInstance = 0;
static struct Hxerox_ilu_IluDebugHooks * _ilujava_jIluDebugHooksInstance = 0;

/*special, dynamicly choosen instances*/
static struct Hxerox_ilu_IluDebugHooks * _ilujava_jAssertFail = 0;
static struct Hxerox_ilu_IluDebugHooks * _ilujava_jCheckFailure = 0;
static struct Hxerox_ilu_IluDebugHooks * _ilujava_jDebugHook = 0;


EXPORTJAVA void 
xerox_ilu_IluDebugHooks_reportInstance(struct Hxerox_ilu_IluDebugHooks * jinst)
/* Initialization help */
{
    _ilujava_jIluDebugHooksInstance = jinst;
}


EXPORTJAVA void 
xerox_ilu_IluDebug_reportIluDebugInstance(struct Hxerox_ilu_IluDebug * jinst)
/* Initialization help */
{
    _ilujava_IluDebugInstance = jinst;
}


INTERNALONLY void 
_ilujava_RaiseDebugHook(ilu_ErrorType et, const char *file, int line)
/* Register this with Ilu */
{
    struct Hxerox_ilu_IluDebugHooks * jinst = _ilujava_jIluDebugHooksInstance;
    if (jinst!=0) {
        struct Hjava_lang_String * js
        	= /*javaString.h*/ makeJavaString((char *) file, strlen(file));
        execute_java_dynamic_method(0,
            (struct Hjava_lang_Object *) jinst,
    	    "raiseDebugHook", 
    	    "(ILjava/lang/String;I)V",
    	    (long) et,
    	    js, 
    	    (long) line);
	}
}

EXPORTJAVA void 
xerox_ilu_IluDebugHooks_defineDebugHook(
    struct Hxerox_ilu_IluDebugHooks * unused, 
    struct Hxerox_ilu_IluDebugHooks * jdbh
    )
/* Java client registers an IluDebugHooks instance (or subclass) */
{
    struct Hxerox_ilu_IluDebugHooks * jold = _ilujava_jDebugHook;
    if (jdbh==0) jdbh = _ilujava_jIluDebugHooksInstance;
    _ilujava_jDebugHook = jdbh;
    if (jold==0) {
        ilu_RaiseDebugHook rdh = & _ilujava_RaiseDebugHook;
        ilu_SetRaiseDebugHook(rdh);
    }
}



INTERNALONLY void 
_ilujava_FailureConsumer(const char *file, int line)
/* Register this with Ilu */
{
    struct Hxerox_ilu_IluDebugHooks * jaf = _ilujava_jAssertFail;
    if (jaf!=0) {
        struct Hjava_lang_String * js
        	= /*javaString.h*/ makeJavaString((char *) file, strlen(file));
        execute_java_dynamic_method(0,
            (struct Hjava_lang_Object *) jaf,
    	    "raiseAssertionFailure", 
    	    "(Ljava/lang/String;I)V",
    	    js, 
    	    (long) line);
    } else {
        _iluJava_BUG("ASSERTION FAILURE");
    }
}

EXPORTJAVA void 
xerox_ilu_IluDebugHooks_defineAssertionFailure(
    struct Hxerox_ilu_IluDebugHooks * unused, 
    struct Hxerox_ilu_IluDebugHooks * jaf
    )
/* Java client registers an IluDebugHooks instance (or subclass) */
{
    struct Hxerox_ilu_IluDebugHooks * jold = _ilujava_jAssertFail;
    if (jaf==0) jaf = _ilujava_jIluDebugHooksInstance;
    _ilujava_jAssertFail = jaf;
    if (jold==0) {
        ilu_FailureConsumer cfc = & _ilujava_FailureConsumer;
        ilu_SetAssertionFailConsumer(cfc);
    }
}




INTERNALONLY void 
_ilujava_CheckFailureConsumer(const char *file, int line)
/* Register this with Ilu */
{
    struct Hxerox_ilu_IluDebugHooks * jcfc = _ilujava_jCheckFailure;
    if (jcfc!=0) {
        struct Hjava_lang_String * js
        	= /*javaString.h*/ makeJavaString((char *) file, strlen(file));
        execute_java_dynamic_method(0,
            (struct Hjava_lang_Object *) jcfc,
    	    "raiseCheckFailure", 
    	    "(Ljava/lang/String;I)V",
    	    js, 
    	    (long) line);
    } else {
        _iluJava_BUG("CONSISTENCY CHECK");
    }
}

EXPORTJAVA void 
xerox_ilu_IluDebugHooks_defineCheckFailure(
    struct Hxerox_ilu_IluDebugHooks * unused, 
    struct Hxerox_ilu_IluDebugHooks * jcf
    )
/* Java client registers an IluDebugHooks instance (or subclass) */
{
    struct Hxerox_ilu_IluDebugHooks * jold = _ilujava_jCheckFailure;
    if (jcf==0) jcf = _ilujava_jIluDebugHooksInstance;
    _ilujava_jCheckFailure = jcf;
    if (jold==0) {
        ilu_CheckFailureConsumer cfc = & _ilujava_CheckFailureConsumer;
        ilu_SetCheckFailureConsumer(cfc);
    }
}

EXPORTLIMITEDVAR int _ilujava_iluDebuggingFlag = 99;
EXPORTLIMITEDVAR int _ilujava_iluSpare1Flag = 99;
EXPORTLIMITEDVAR int _ilujava_iluSpare2Flag = 99;
EXPORTLIMITEDVAR int _ilujava_objectsFlag = 99;
EXPORTLIMITEDVAR int _ilujava_objectTransportFlag = 99;
EXPORTLIMITEDVAR int _ilujava_callsFlag = 99;
EXPORTLIMITEDVAR int _ilujava_definitionsFlag = 99;
EXPORTLIMITEDVAR int _ilujava_pctFlag = 99;
EXPORTLIMITEDVAR int _ilujava_traceFlag = 99;
EXPORTLIMITEDVAR int _ilujava_gcFlag = 99;
EXPORTLIMITEDVAR int _ilujava_lockFlag = 0;
EXPORTLIMITEDVAR int _ilujava_alarmFlag = 0;


EXPORTJAVA void
xerox_ilu_IluDebug_reportFlags(
    struct Hxerox_ilu_IluDebug * unused, 
    long iluDebuggingFlag,
    long iluSpare1Flag,
    long iluSpare2Flag,
    long objectsFlag,
    long objectTransportFlag,
    long callsFlag,
    long definitionsFlag,
    long pctFlag,
    long gcFlag,
    long lockFlag,
    long alarmFlag
    )
{
	_ilujava_iluDebuggingFlag =	iluDebuggingFlag;
	_ilujava_iluSpare1Flag =	iluSpare1Flag;
	_ilujava_iluSpare2Flag =	iluSpare2Flag;
	_ilujava_objectsFlag =  	objectsFlag;
	_ilujava_objectTransportFlag =	objectTransportFlag;
	_ilujava_callsFlag =    	callsFlag;
	_ilujava_definitionsFlag =	definitionsFlag;
	_ilujava_pctFlag =		pctFlag;
	_ilujava_gcFlag =		gcFlag;
	_ilujava_lockFlag =		lockFlag;
	_ilujava_alarmFlag =		alarmFlag;
}



EXPORTJAVA void
xerox_ilu_IluDebugStream_debugWrite1(
    struct Hxerox_ilu_IluDebugStream * unused, 
    long b
    )
{
    ilu_DebugPrintf("%c", b);
}


EXPORTJAVA void
xerox_ilu_IluDebugStream_debugWriteN(
    struct Hxerox_ilu_IluDebugStream * unused, 
    HArrayOfByte * b,
    long off,
    long len
    )
{
    long sz;
    char * ptr;
    if (b) {
        sz = obj_length(b);
        ptr = (char *) unhand(b)->body;
        while (len>0 & off<sz) {
            ilu_DebugPrintf("%c", ptr[off]);
            off = off+1;
            len = len-1;
        }
    }
}


EXPORTLIMITED char * 
_iluJava_captureJavaStack()
{
    struct Hxerox_ilu_IluDebug * jcfc = _ilujava_IluDebugInstance;
    if (jcfc!=0) {
        struct Hjava_lang_String * js;
        js = (struct Hjava_lang_String *) execute_java_dynamic_method(0,
            (struct Hjava_lang_Object *) jcfc,
    	    "dynamicCaptureStackTrace", 
    	    "()Ljava/lang/String;"
    	    );
        if (js) {
            return /*javaString*/ allocCString(js);
        }
    } 
    return 0;
}


EXPORTLIMITED int 
_iluJava_getIntProp(char * key, int deflt)
/* Read a java property value */
{
    int val = deflt;
    struct Hxerox_ilu_IluDebug * jcfc = _ilujava_IluDebugInstance;
    if (jcfc!=0 && key!=0) {
        struct Hjava_lang_String * js = makeJavaString(key, strlen(key));
        val = (int) execute_java_dynamic_method(0,
            (struct Hjava_lang_Object *) jcfc,
    	    "dynamicGetIntProp", 
    	    "(Ljava/lang/String;I)I",
    	    js,
    	    deflt
    	    );
    } 
    return val;
}

