/* IluJava_IluOInt.c */
/* Chris Jacobi, February 1, 1997 7:32 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* 
 * $Id: IluJava_IluOInt.c,v 1.11 1997/02/02 07:00:27 jacobi Exp $ 
 *
 *
 * This class helps in the mapping from real Java objects to kernel object 
 * Native side for IluOInt.java
 */

 
#include <StubPreamble.h>
#include <interpreter.h>
#include "xerox_ilu_IluOInt.h"
#include "xerox_ilu_IluOIntContainer.h"
#include "xerox_ilu_IluObject0.h"
#include "xerox_ilu_IluObject.h"
#include "xerox_ilu_IluClassRep.h"
#include "xerox_ilu_IluCall.h"
#include "IluJava.h"
#include <iluhash.h>


#include <sys_api.h>
/*
 *  WARNING
 *  *******
 * 
 *  sys_api.h as we got it here at xerox, in
 *  /project/java/jdk-1.0.2/include/sys_api.h
 *  has a bug:  It imports sysmacros_md.h
 *  which is not available.  We have made a
 *  a private copy of sysmacros_md to fix this
 *  java bug.  Here at Xerox we do not use 
 *  Sun's version which might be available in a 
 *  source release.
 *
 *  See sysmacros_md.h
 */


static HashTable _cjOb2jOI; /* key any java object; val is IluOInt */
static sys_mon_t * _ilujava_peekTabLock; 


static struct Hxerox_ilu_IluOInt * 
_ilujava_internalPeek(struct Hxerox_ilu_IluObject * jobj) 
{
     struct Hxerox_ilu_IluOInt * joi;
     joi = (struct Hxerox_ilu_IluOInt *)
     		_ilu_hash_FindInTable(_cjOb2jOI, jobj);
     return joi;
}


EXPORTJAVA long
xerox_ilu_IluOInt_setupForPeek(
	struct Hxerox_ilu_IluOInt * joi,
	struct Hxerox_ilu_IluObject * jobj
	) 
{
    if (joi==0 | jobj==0) return 1;
    sysMonitorEnter(_ilujava_peekTabLock);
    if (_ilujava_internalPeek(jobj) != 0) {
        sysMonitorExit(_ilujava_peekTabLock);
        return 1;
    }
    _ilu_hash_AddToTable(_cjOb2jOI, (ilu_refany) jobj, (ilu_refany) joi);
    unhand(joi)->jJavaObject = jobj;
    sysMonitorExit(_ilujava_peekTabLock);
    return 0;
}



static void 
internalDisablePeek(struct Hxerox_ilu_IluOInt * joi) 
{
     struct Hxerox_ilu_IluObject * jobj = unhand(joi)->jJavaObject;
     if (jobj != 0) {
         _ilu_hash_RemoveFromTable(_cjOb2jOI, jobj);
     }
}

static void
disablePeek(struct Hxerox_ilu_IluOInt * joi) 
{
     sysMonitorEnter(_ilujava_peekTabLock);
     internalDisablePeek(joi);
     sysMonitorExit(_ilujava_peekTabLock);
}


EXPORTJAVA void 
xerox_ilu_IluOInt_nPeek(
	struct Hxerox_ilu_IluOInt * unused,
	struct Hxerox_ilu_IluObject * jobj,
	struct Hxerox_ilu_IluOIntContainer * jCont)
/* private procedure; we trust jCont */ 
{
    struct Hxerox_ilu_IluOInt * joi;
    sysMonitorEnter(_ilujava_peekTabLock);
    joi = _ilujava_internalPeek(jobj);
    if (joi) {
        if (unhand(joi)->destroyed) {joi = 0;}
    }
    unhand(jCont)->oi = joi;
    sysMonitorExit(_ilujava_peekTabLock);
    if (joi) {
        if (unhand(joi)->jJavaObject!=jobj) {
            BUGPRINT("$ IluOInt_nPeek HUH?");
        }
    }
}


EXPORTJAVA struct Hxerox_ilu_IluOInt * 
xerox_ilu_IluOInt_nPeekWithCall(
	struct Hxerox_ilu_IluOInt * unused,
	struct Hxerox_ilu_IluObject * jobj,
	struct Hxerox_ilu_IluCall * jCall)
/* 
 * private procedure; we trust jCall.
 * Very similar to nPeek but uses an IluCall for traceability 
 */ 
{
    struct Hxerox_ilu_IluOInt * joi;
    sysMonitorEnter(_ilujava_peekTabLock);
    joi = _ilujava_internalPeek(jobj);
    if (joi) {
         if (unhand(joi)->destroyed) joi = 0;
    }
    unhand(jCall)->jObjTracable = (HObject*) joi;
    sysMonitorExit(_ilujava_peekTabLock);
    if (joi) {
        if (unhand(joi)->jJavaObject!=jobj) {
            BUGPRINT("$ IluOInt_nPeekWithCall HUH?");
        }
    }
    return joi;
}


EXPORTJAVA long 
xerox_ilu_IluOInt_destroyGentleOI(
	struct Hxerox_ilu_IluOInt * joi) 
{
     struct Hxerox_ilu_IluClassRep* jClass = unhand(joi)->jClassRep;
     ilu_Class cClass = Access(jClass, yIluClass, ilu_Class);
     ilu_Server cServer = Access(joi, yServer, ilu_Server);
     ilu_Object cIluObject;
     if (_ilujava_gcFlag>0) {
         ilu_DebugPrintf("$ destroyGentleOI_0: looking at %x \n", joi);
     }
     if (cServer==0) {
         if (_ilujava_gcFlag>0) {
             ilu_DebugPrintf("$ destroyGentleOI_1: %x  no server \n", joi);   
         }
         return 0; /*go away with it*/
     }
     if (unhand(joi)->veryInterest | unhand(joi)->retained | unhand(joi)->ghost) {
         /*Not monitored, must be conservative*/
         if (_ilujava_gcFlag>0) {
             ilu_DebugPrintf("$ destroyGentleOI_2: re-finalize %x \n", joi);   
         }
         return 1; /*re-finalize*/
     }
     /* use the server lock */  
     ilu_EnterServer(cServer, cClass);
         cIluObject = Access(joi, yIluKernelObject, ilu_Object);
         if (cIluObject!=0) {
             if (ilu_VeryInterested(cIluObject) | unhand(joi)->ghost) {
                 ilu_ExitServer(cServer, cClass);
                 if (_ilujava_gcFlag>0) {
                     ilu_DebugPrintf("$ destroyGentleOI_3: re-finalize %x \n", joi);   
                 }
                 return 1; /*re-finalize*/
             }
         }
         sysMonitorEnter(_ilujava_peekTabLock);
         if (unhand(joi)->veryInterest | unhand(joi)->retained | unhand(joi)->ghost) {
            sysMonitorExit(_ilujava_peekTabLock);
            ilu_ExitServer(cServer, cClass);
            if (_ilujava_gcFlag>0) {
                     ilu_DebugPrintf("$ destroyGentleOI_4: re-finalize %x \n", joi);   
            }
            return 1; /*re-finalize*/
         }
         unhand(joi)->destroyed = 1;
         unhand(joi)->yIluKernelObject = 0;
         internalDisablePeek(joi);
         sysMonitorExit(_ilujava_peekTabLock);
         if (cIluObject) {
             ilu_RegisterLanguageSpecificObject(cIluObject, 0, JAVALANGIDX);
         }
     ilu_ExitServer(cServer, cClass);
     return 0; /*don't re-finalize*/
}

EXPORTJAVA void
xerox_ilu_IluOInt_destroyRudeOI(
	struct Hxerox_ilu_IluOInt * joi)
{
     struct Hxerox_ilu_IluClassRep* jClass = unhand(joi)->jClassRep;
     ilu_Class cClass = Access(jClass, yIluClass, ilu_Class);
     ilu_Server cServer = Access(joi, yServer, ilu_Server);
     ilu_Object cIluObject;
     if (_ilujava_gcFlag>0) {
         ilu_DebugPrintf("$ destroyRudeOI_0: looking at %x \n", joi);
     }
     if (unhand(joi)->destroyed) return;
     if (cServer==0) {
         if (_ilujava_gcFlag>0) {
             ilu_DebugPrintf("$ destroyRudeOI_1: %x  no server \n", joi);   
         }
         return; /*go away with it*/
     }
     ilu_EnterServer(cServer, cClass);
         cIluObject = Access(joi, yIluKernelObject, ilu_Object);
         sysMonitorEnter(_ilujava_peekTabLock);
         unhand(joi)->destroyed = 1;
         unhand(joi)->yIluKernelObject = 0;
         internalDisablePeek(joi);
         sysMonitorExit(_ilujava_peekTabLock);
         if (cIluObject) {
             ilu_RegisterLanguageSpecificObject(cIluObject, 0, JAVALANGIDX);
         }
     ilu_ExitServer(cServer, cClass);
}

EXPORTJAVA void
xerox_ilu_IluOInt_withdrawOI(
	struct Hxerox_ilu_IluOInt * joi)
{
    char * ckey;
    ilu_boolean huh;
    ilu_Object cIluObject;
    ilu_Server cServer = Access(joi, yServer, ilu_Server);
    struct Hxerox_ilu_IluClassRep * jClass = unhand(joi)->jClassRep;
    ilu_Class cClass = Access(jClass, yIluClass, ilu_Class);
    if (cServer==0 | cClass==0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "yyy-3");
        return;
    }
    ilu_EnterServer(cServer, cClass);
    cIluObject = Access(joi, yIluKernelObject, ilu_Object);
    if (cIluObject==0) {
        ilu_ExitServer(cServer, cClass);
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "yyy-3");
        return;
    }
    ckey = Access(joi, yOwnerKey, char *);
    if (ckey) {
        Access(joi, yOwnerKey, char *) = 0;
        huh = ilu_WithdrawObject(cIluObject, ckey);
    } else {
        ilu_ExitServer(cServer, cClass);
    }
}

EXPORTJAVA void
xerox_ilu_IluOInt_publishOI(
	struct Hxerox_ilu_IluOInt * joi)
{
    char * ckey;
    ilu_Class cClass;
    ilu_Server cServer;
    ilu_Object cIluObject;
    struct Hxerox_ilu_IluClassRep * jClass;
    cServer = Access(joi, yServer, ilu_Server);
    jClass = unhand(joi)->jClassRep;
    cClass = Access(jClass, yIluClass, ilu_Class);
    ilu_EnterServer(cServer, cClass);
    cIluObject = Access(joi, yIluKernelObject, ilu_Object);
    if (cIluObject==0) {
        ilu_ExitServer(cServer, cClass);
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "xxx-203"
        	" Object no more exported by ilu");
        return;
    }
    ckey = Access(joi, yOwnerKey, char *);
    if (ckey) {
        ilu_ExitServer(cServer, cClass);
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "xxx-204");
        return;
    }
    Access(joi, yOwnerKey, char *) = ckey;
    ckey = ilu_PublishObject(cIluObject);
}


/* called within server lock */
static void 
_ilujava_noter(ilu_Object cobj, int vi)
{
    struct Hxerox_ilu_IluOInt * joi;
    joi = (struct Hxerox_ilu_IluOInt *) 
    	ilu_GetLanguageSpecificObject(cobj, JAVALANGIDX);
    if (_ilujava_gcFlag>0) {
        ilu_DebugPrintf("$ IluOInt noter: %d c<%x> j<%x> \n", vi, cobj, joi);
    }
    if (joi!=ILU_NIL && ilu_TrueInstanceP(cobj)) {
        ilu_Class cl = ilu_ClassOfObject(cobj);
        if (_ilujava_gcFlag>1) {
            ilu_DebugPrintf("$ IluOInt noter is-true\n");
        }
        if ( ilu_CollectibleP(cl)) {
            struct execenv *ee = EE(); /*Java execution environment*/
            if (_ilujava_gcFlag>1) {
                ilu_DebugPrintf("$ IluOInt noter is-collectable\n");
            }
            if (vi) {
                execute_java_dynamic_method(ee,
                    (struct Hjava_lang_Object *) joi,
                    "showInterest", 
                    "()V");
            } else {
                execute_java_dynamic_method(ee,
                    (struct Hjava_lang_Object *) joi,
                    "removeInterest", 
                    "()V");
            }
            if (ee->exceptionKind !=0) {
                _iluJava_PANIC("$ ObjectAxillary noter: java exception\n");
            }
        }
    }
}				


EXPORTLIMITED void
_ilujava_IluOIntInit()
{
    _ilujava_peekTabLock = (sys_mon_t *) java_sysMalloc(sysMonitorSizeof());
    memset((char *)_ilujava_peekTabLock, 0, sysMonitorSizeof());
#ifdef JDK_11
    sysMonitorInit(_ilujava_peekTabLock);
#else
    sysMonitorInit(_ilujava_peekTabLock, FALSE);
#endif       
    _cjOb2jOI = _ilu_hash_MakeNewTable(111, 
    	_ilu_hash_HashPointer, 
    	_ilu_hash_PointerCompare
    	);
    
    ilu_SetNoter(_ilujava_noter, JAVALANGIDX);
}


EXPORTJAVA Hjava_lang_String *
xerox_ilu_IluOInt_sbhOfOI(
	struct Hxerox_ilu_IluOInt * joi)
{
    Hjava_lang_String * jsbh = 0;
    char * csbh;
    ilu_Class cClass;
    ilu_Server cServer = Access(joi, yServer, ilu_Server);
    ilu_Object cIluObject;
    struct Hxerox_ilu_IluClassRep * jClass = unhand(joi)->jClassRep;
    if (cServer==0 ) {
        return 0;
    }
    cClass = Access(jClass, yIluClass, ilu_Class);
    ilu_EnterServer(cServer, cClass);
    cIluObject = Access(joi, yIluKernelObject, ilu_Object);
    if (cIluObject) {
        csbh = ilu_SBHOfObject(cIluObject);
        if (csbh) {
            jsbh = makeJavaString(csbh, strlen(csbh));
        }
    }
    ilu_ExitServer(cServer, cClass);
    return jsbh;
}


