/* IluJava_IluPort.c */ 
/* Chris Jacobi, December 20, 1996 9:56 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluPort.c,v 1.10 1996/12/21 06:42:33 jacobi Exp $ */
 
/*
 * C side of IluPort.java and IluTransportInfo.java
 */

#include <StubPreamble.h>
#include <javaString.h>
#include "xerox_ilu_IluPort.h"
#include "xerox_ilu_IluTransportInfo.h"
#include "xerox_ilu_IluServer.h"

#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"
#include "IluJava_JStubs.h"

EXPORTJAVA void 
xerox_ilu_IluPort_nativeInitPort(
		struct Hxerox_ilu_IluPort * jport
		)
{
    struct Hxerox_ilu_IluServer * jServer 
    	= unhand(jport)->jServer;
    struct Hjava_lang_String * jProtocolInfo 
    	= unhand(jport)->jProtocolInfo;
    struct Hxerox_ilu_IluTransportInfo * jTransportInfo 
    	= unhand(jport)->jTransportInfo;
    struct Hxerox_ilu_IluPassport * jPassport 
    	= unhand(jport)->jPassport;
    ilu_string cProtocolInfo;
    ilu_Server cServer;
    ilu_TransportInfo cTransportInfo;
    ilu_Passport cPassport = 0; 
    ilu_Port cPort;
    ilu_Error errs;
    cProtocolInfo = /*javaString.h*/ makeCString(jProtocolInfo);
    cTransportInfo = Access(jTransportInfo, yIluTransportInfo, ilu_TransportInfo);
    cServer = Access(jServer, yIluServer, ilu_Server);

    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluPort_nativeInitPort (jport: %x  jsrv: %x) srv %x\n", 
            (int) jport, (int) jServer, (int) cServer);
    }
    if (jPassport) cPassport = Access(jPassport, yIluPassport, ilu_Passport);
    cPort = ilu_CreatePort(cServer, cProtocolInfo, cTransportInfo, cPassport, &errs);
    if (ILU_ERRNOK(errs)) {
        BUGPRINT("$ nativeInitPort: some error");
        _iluJava_IluErrorToJavaException(errs, "xxx-206");
        return;
    }
    if (cPort==0) {
        BUGPRINT("$ nativeInitPort: 0 port");
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "xxx-207");
        return;
    }
    Access(jport, yPort, ilu_Port) = cPort;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ Port created %x \n", (int) cPort);
    }
} /* nativeInitPort */


EXPORTJAVA void 
xerox_ilu_IluPort_nativeClosePort(
		struct Hxerox_ilu_IluPort * jport
		)
{
    ilu_Port cPort = Access(jport, yPort, ilu_Port);
    if (cPort == 0) return;
    ilu_ClosePort(cPort);
 }  /* nativeClosePort */


EXPORTJAVA void 
xerox_ilu_IluPort_nativeDonePort(
		struct Hxerox_ilu_IluPort * jport
		)
{
    ILU_ERRS((bad_param, bad_locks, internal)) err;
    ilu_boolean dontCare;
    ilu_Port cPort = Access(jport, yPort, ilu_Port);
    ILU_CLER(err);
    if (cPort == 0) return;
    Access(jport, yPort, ilu_Port) = 0;
    if (unhand(jport)->pstate > 3) return;
    unhand(jport)->pstate = 99;
    dontCare = ilu_DoneWithPort(cPort, &err);
}  /* nativeDonePort */


EXPORTJAVA void 
xerox_ilu_IluTransportInfo_nativeInitTransportInfo(
		struct Hxerox_ilu_IluTransportInfo * jti,
		HArrayOfString * jinfo
		)
{
    char ** cObject;
    char * rover; /* the next un-assigned byte */
    struct Hjava_lang_String * jString;
    int totalLen = 0;
    int thisLen = 0;
    int i;
    int cnt = obj_length(jinfo);
    /* compute length */
    for (i = 0; i < cnt; i = i+1)
    {
        jString = unhand(jinfo)->body[i];
        thisLen = javaStringLength(jString);/*javaString.h*/
        totalLen = totalLen+thisLen;
    }
    totalLen = totalLen + cnt; /* add padding per string */
    totalLen = totalLen + cnt; /* add extra padding per string; see below */
    totalLen = totalLen + (cnt+2)*sizeof(char *); /*pointer array, stopper and final padding*/
    /* allocate data */
    cObject = (char **) ilu_must_malloc(totalLen);
    Access(jti, yIluTransportInfo, char **) = cObject;
    /* copy data */
    cObject[cnt] = 0;
    rover = (char *) & (cObject[cnt+1]);
    for (i = 0; i < cnt; i = i+1)
    {
        cObject[i] = rover;
        jString = unhand(jinfo)->body[i];
        thisLen = javaStringLength(jString);/*javaString.h*/
        /* 
         * ? Must add 1 otherwise it overwrites the last character.   
         * I do not understand this, but the java rt does the same thing.
         */
        javaString2CString(jString, rover, thisLen+1); /*javaString.h*/
        rover = rover + thisLen+1;
        rover[0] = '\0';
        rover = rover + 1; 
    }
    
} /* nativeInitTransportInfo */


