/* IluJava_IluRT0.c */
/* Chris Jacobi, January 9, 1997 5:13 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluRT0.c,v 1.12 1997/01/10 01:15:39 jacobi Exp $ */

/*
 * C side of IluRT0.java
 */

#include <StubPreamble.h>
#include <javaString.h>
#include "xerox_ilu_IluRT0.h"
#include "xerox_ilu_IluOInt.h"
#include "xerox_ilu_IluClassRep.h"
#include "xerox_ilu_IluServer.h"

#include <iluxport.h>
#include <iluerror.h>
#include <iluhash.h>
#include <iluntrnl.h>
#include <iluerrs.h>

#include "IluJava_CJ.h"
#include "IluJava.h"


IMPORTVAR char * _iluJava_notUsed1;
IMPORTVAR char * _iluJava_notUsed2;

/* ================ Registry ================ */


static HashTable _c2j; /* key ilu_Class; val is java IluClassRep */


EXPORTLIMITED void
_ilujava_IluRt0InitHashTable()
{
    _c2j = _ilu_hash_MakeNewTable(111, 
    	_ilu_hash_HashPointer, 
    	_ilu_hash_PointerCompare
    	);
}

 
/* locking by caller */
EXPORTLIMITED void 
_iluJava_registerJavaClass (
	ilu_Class cIluClass, 
	/* struct Hxerox_ilu_IluClassRep * */ ilu_refany jClassRep
	)
/* Register java class necessary to CREATE java surrogate object or TRUE */
{
  _ilu_hash_AddToTable(_c2j, 
  	(ilu_refany) cIluClass, 
  	(ilu_refany) jClassRep
  	);
}


EXPORTLIMITED ilu_refany   /* struct Hxerox_ilu_IluClassRep * */
_ilujava_findJavaClass(ilu_Class cIluClass)
/* Get java class necessary to CREATE java surrogate object */
{
    return _ilu_hash_FindInTable(_c2j, cIluClass);
}

/* ================ ======== ================ */


  
/* Inside(cIluObject->server, cIluObject->class) */
/* oi must be finished */
EXPORTLIMITED ilu_refany /* struct Hxerox_ilu_IluOInt* */
_iluJava_creatSurrOIFromRegistry(
	ilu_Object cIluObject, 
	ilu_refany jClassRep   /* struct Hxerox_ilu_IluClassRep* */
	)
{
    struct Hxerox_ilu_IluClassRep* jExpectedClass 
    	= (struct Hxerox_ilu_IluClassRep*) jClassRep;
    HObject * jobj;
    struct Hxerox_ilu_IluOInt * joi;
    ilu_Class cActualIluClass;
    struct Hxerox_ilu_IluClassRep * jActualClass;
    
    cActualIluClass = ilu_ClassOfObject(cIluObject);
    jActualClass = (Hxerox_ilu_IluClassRep *)
        _ilujava_findJavaClass(cActualIluClass);
    if (jActualClass==0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", 
            "surrogate class not registered"
            );
        BUGPRINT("$ IluJava failure: surrogate class not found... \n");
        return 0;
    }
/*
    if (jActualClass!=jExpectedClass) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "ilu-class missmatch");
        BUGPRINT("$ IluJava failure: class missmatch... \n");
        return 0;
    }
*/
    jobj = execute_java_constructor(
    	PRIVILEGED_EE,
    	"xerox/ilu/IluOInt", 
    	0, /* ClassClass * or 0 if named */
    	"()"
    	);
    if (jobj==0) 
    {
       /* failed somehow in java */
       _iluJava_PANIC("$ IluJava: IluOInt creation failed... \n");
       return 0;
    }
    joi = (struct Hxerox_ilu_IluOInt *) jobj;
    unhand(joi)->jClassRep = jActualClass;
    Access(joi, yServer, ilu_Server) = ilu_ServerOfObject(cIluObject);
    Access(joi, yIluKernelObject, ilu_Object) = cIluObject;
    ilu_RegisterLanguageSpecificObject(cIluObject, joi, JAVALANGIDX);
    /* We just created joi. Don't declare it a ghost; if it is found
     * by the collector, so be it.
     */
    return (ilu_refany) joi;
}


/* compare this to /tilde/jacobi/ilu/runtime/c/ilu.c 
 * and look at /tilde/jacobi/ilu/runtime/kernel/iluerrs.h
 */ 

#define CORBAPREFIX "CORBA/corba_"

/* Shorthand used in IluErrorToJavaException */
#define E_SWITCH(err, exc) 				\
    ILU_ERR_CASE(err, v) {				\
        SignalError(EE(), exc, t); 			\
        return;						\
    }

/* Shorthand used in IluErrorToJavaException 
 * IN CASE THE XCEPTION IS NOT YET IMPLEMENTED
 */
#define UNIMPL_SWITCH(err, ignored) 			\
    ILU_ERR_CASE(err, v) {				\
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", t); 	\
        return;						\
    }


EXPORTLIMITED void 
_iluJava_IluErrorToJavaException(ilu_Error err, char * t)
{
    ILU_ERR_SWITCH(err) {
        ILU_SUCCESS_CASE return;

        /*.../ilu/runtime/kernel/iluerrs.h calls these public */
        E_SWITCH(unknown,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(bad_param,"xerox/ilu/IluBadParameterException")
        UNIMPL_SWITCH(no_memory,CORBAPREFIX "NO_MEMORY")
        UNIMPL_SWITCH(imp_limit,CORBAPREFIX "IMP_LIMIT")
        UNIMPL_SWITCH(comm_failure,CORBAPREFIX "COMM_FAILURE")
        UNIMPL_SWITCH(inv_objref,CORBAPREFIX "INV_OBJREF")
        UNIMPL_SWITCH(no_permission,CORBAPREFIX "NO_PERMISSION")
        E_SWITCH(internal,"xerox/ilu/IluInternalException")
        E_SWITCH(marshal,CORBAPREFIX "MARSHAL")
        UNIMPL_SWITCH(initialize,CORBAPREFIX "INITIALIZE")
        UNIMPL_SWITCH(no_implement,CORBAPREFIX "NO_IMPLEMENT")
        UNIMPL_SWITCH(bad_typecode,CORBAPREFIX "BAD_TYPECODE")
        UNIMPL_SWITCH(bad_operation,CORBAPREFIX "BAD_OPERATION")
        UNIMPL_SWITCH(no_resources,CORBAPREFIX "NO_RESOURCES")
        UNIMPL_SWITCH(no_response,CORBAPREFIX "NO_RESPONSE")
        UNIMPL_SWITCH(persist_store,CORBAPREFIX "PERSIST_STORE")
        UNIMPL_SWITCH(bad_inv_order,CORBAPREFIX "BAD_INV_ORDER")
        UNIMPL_SWITCH(transient,CORBAPREFIX "TRANSIENT")
        UNIMPL_SWITCH(free_mem,CORBAPREFIX "FREE_MEM")
        UNIMPL_SWITCH(inv_ident,CORBAPREFIX "INV_IDENT")
        UNIMPL_SWITCH(inv_flag,CORBAPREFIX "INV_FLAG")
        UNIMPL_SWITCH(intf_repos,CORBAPREFIX "INTF_REPOS")
        UNIMPL_SWITCH(bad_context,CORBAPREFIX "BAD_CONTEXT")
        UNIMPL_SWITCH(obj_adapter,CORBAPREFIX "OBJ_ADAPTER")
        UNIMPL_SWITCH(data_conversion,CORBAPREFIX "DATA_CONVERSION")
        UNIMPL_SWITCH(bad_locks,CORBAPREFIX "BAD_LOCKS")
        UNIMPL_SWITCH(broken_locks,CORBAPREFIX "BROKEN_LOCKS")
        UNIMPL_SWITCH(interrupted,CORBAPREFIX "INTERRUPTED")
        UNIMPL_SWITCH(gss_security,CORBAPREFIX "GSS_SECURITY")
        
        /*.../ilu/runtime/kernel/iluerrs.h calls these unresolved */
        E_SWITCH(KernelBroken,"xerox/ilu/IluInternalException")
        E_SWITCH(BadDataStructure,"xerox/ilu/IluInternalException")
        E_SWITCH(MallocFailure,"xerox/ilu/IluInternalException")
        E_SWITCH(MalformedKernelObject,"xerox/ilu/IluInternalException")
        E_SWITCH(MaxCountExceeded,"xerox/ilu/IluInternalException")
        E_SWITCH(ProtocolAlreadyRegistered,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(TransportAlreadyRegistered,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(WrongLocks,"xerox/ilu/IluInternalException")
        E_SWITCH(BadProtocolInfo,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(BadArguments,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(GcRegFailed,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(NoConnection,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(ClosedServer,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(ObjectNotFoundInServer,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(NoObjectForSBH,"xerox/ilu/IluSomeSystemException")
        E_SWITCH(CantCondition,"xerox/ilu/IluSomeSystemException")
        
        ILU_ERR_ELSE {
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", t);
            return;
        }
    } ILU_ERR_ENDSWITCH;
} 


EXPORTJAVA struct Hxerox_ilu_IluOInt *
xerox_ilu_IluRT0_oiFromSBH(
	struct Hxerox_ilu_IluRT0 * unused,
	struct Hjava_lang_String * jSBH,
	struct Hxerox_ilu_IluClassRep * jClassRep
	)
{
    struct Hxerox_ilu_IluOInt * joi; 
    ilu_Object cIluObject;
    ilu_Error err = ILU_INIT_NO_ERR;
    char * cSBH = makeCString(jSBH); /*<javaString.h> subject to GC*/
    
    cIluObject = ilu_ObjectOfSBH(
    	cSBH, 
    	Access(jClassRep, yIluClass, ilu_Class),
    	&err
    	);
    if (ILU_ERRNOK(err)) {
        _iluJava_IluErrorToJavaException(err, "ilu_ObjectOfSBH");
        return 0;
    }
    if (cIluObject == 0) {
        return 0;
    }
    joi = (struct Hxerox_ilu_IluOInt*) ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    /* true or surrogate at this point */
    if (joi == 0) {
        /* but we do not create true objects here */
        joi = (struct Hxerox_ilu_IluOInt*) _iluJava_creatSurrOIFromRegistry(
        	cIluObject, 
        	(ilu_refany) jClassRep
        	);
    } else {
        if (unhand(joi)->retained) {unhand(joi)->ghost = 1;}
    }
    ilu_ExitServer(
    	ilu_ServerOfObject(cIluObject),
    	Access(jClassRep, yIluClass, ilu_Class)
    	);
    _iluJava_notUsed1 = cSBH;
    return joi; 
}


EXPORTJAVA struct Hxerox_ilu_IluOInt *
xerox_ilu_IluRT0_oiFromLookup(
	struct Hxerox_ilu_IluRT0 * unused,
	struct Hjava_lang_String * jsid,
	struct Hjava_lang_String * jih,
	struct Hxerox_ilu_IluClassRep * jClassRep
	)
{
    struct Hxerox_ilu_IluOInt * joi; 
    char * csid = makeCString(jsid); /* <javaString.h> subject to GC */
    char * cih = makeCString(jih); /* <javaString.h> subject to GC */
    ilu_Object cIluObject = ILU_NIL;  
    ilu_Class cIluClass = Access(jClassRep, yIluClass, ilu_Class);
    int cchange;
    cchange = ilu_ReLookupObject(csid, cih, cIluClass, &cIluObject);
    if (cIluObject == 0) {
        return 0;
    }
    joi = (struct Hxerox_ilu_IluOInt *) ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    if (joi == ILU_NIL) {
        joi = (struct Hxerox_ilu_IluOInt *) _iluJava_creatSurrOIFromRegistry(
        	cIluObject, 
        	(ilu_refany) jClassRep
        	);
    } else {
        if (unhand(joi)->retained) {unhand(joi)->ghost = 1;}
    }
    ilu_ExitServer(
    	ilu_ServerOfObject(cIluObject),
    	Access(jClassRep, yIluClass, ilu_Class)
    	);
    _iluJava_notUsed1 = csid;
    _iluJava_notUsed2 = cih;
    return joi;
}


EXPORTJAVA struct Hjava_lang_String *
xerox_ilu_IluRT0_inventID(struct Hxerox_ilu_IluRT0 * unused)
{
    Hjava_lang_String * jstring;
    ilu_string cstring = ilu_InventID();
    jstring = /*javaString.h*/ makeJavaString(cstring, strlen(cstring));
    ilu_free(cstring);
    return jstring;
}

IMPORTVAR char * _iluJava_notUsed1;

EXPORTJAVA void
xerox_ilu_IluRT0_registerTrue(
	struct Hxerox_ilu_IluRT0 * unused,
	struct Hxerox_ilu_IluOInt * joi,
	Hjava_lang_String * jih,
	struct Hxerox_ilu_IluServer * jserver
	)
/* NOTE: 
 * similar code in (IluJava_IluServer.c)_ilujava_ObjectTableCreateObj
 */
{
    ilu_Object cObject;
    ilu_string cih = 0;
    ilu_Server cServer;
    ilu_Class cClass;
    struct Hxerox_ilu_IluClassRep * jClass;
    jClass = unhand(joi)->jClassRep;
    if (jih==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR no instance handle\n");
    }
    if (jClass==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR jClass==0\n");
    }
    if (jserver==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR jserver==0\n");
    }
    cServer = Access(jserver, yIluServer, ilu_Server);
    if (cServer==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR cServer==0\n");
    }
    Access(joi, yServer, ilu_Server) = cServer;
    cClass = Access(jClass, yIluClass, ilu_Class);
    if (cClass==0) {
        BUGPRINT("$ ** IluRT0_registerTrue ERROR cClass==0\n");
    }
    cih = allocCString(jih); /*javaString.h using malloc*/ 
    ilu_EnterServer(cServer, cClass);
    cObject = ilu_FindOrCreateTrueObject(cih, cServer, cClass, joi);
    java_free(cih);
    
    if (cObject) {
        ilu_refany jalt;
        if (_ilujava_gcFlag>1) {
            cObject->ob_timeout = 120; /*for debugging: gc after two minutes*/
        }
        jalt = ilu_GetLanguageSpecificObject(cObject, JAVALANGIDX);
        if (jalt != joi) {
            /* This can happen: 
             * joi has been created outside of the server lock.... 
             */
            if (_ilujava_objectsFlag>1) {
                ilu_DebugPrintf("$ registerTrue: object creation conflict\n");   
            }
            cObject = 0;
        }
    }
    if (_ilujava_objectsFlag>1) {
        ilu_DebugPrintf("$ xerox_ilu_IluRT0_registerTrue: found %x \n", cObject);   
    }
    Access(joi, yIluKernelObject, ilu_Object) = cObject;
    ilu_ExitServer(cServer, cClass);
    _iluJava_notUsed1 = cih;
}



