/* IluJava_IluServer.c */
/* Chris Jacobi, December 18, 1996 7:14 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluServer.c,v 1.6 1996/12/19 05:17:15 jacobi Exp $ */
 
/* 
 * C side for IluServer.java 
 * See also IluObjectTable.java
 */
 
#include <StubPreamble.h>
#include <javaString.h>
#include <interpreter.h>

#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"
#include "IluJava_JStubs.h"

/* for debugging only */
#include <iluntrnl.h>

/* L1 >= {server}; L2, Main unconstrained */
INTERNALONLY ilu_Object
_ilujava_ObjectTableCreateObj(ilu_ObjectTable cot, ilu_string cih)
/* Upcalled by kernel to request creation of an object */
{
    ILU_ERRS((no_memory)) err;
    ilu_Server cServer;
    ilu_Class cClass;
    ilu_Object cObject;
    struct Hxerox_ilu_IluClassRep * jClass;
    struct Hxerox_ilu_IluObjectTable * jot;
    struct Hxerox_ilu_IluServer * jServer;
    struct Hxerox_ilu_IluOInt * joi;
    struct Hjava_lang_String * jih;
    jot = (struct Hxerox_ilu_IluObjectTable *) cot->ot_rock;
    jServer = unhand(jot)->jServer;
    jih = /*javaString.h*/ makeJavaString((char *) cih, strlen(cih));
    if (_ilujava_objectsFlag>1) {
        ilu_DebugPrintf("$ ObjectTableCreateObj for %s jServer %x\n", cih, jServer);
    }
    joi = (struct Hxerox_ilu_IluOInt *)
			execute_java_dynamic_method(EE(),
        (struct Hjava_lang_Object *) jot,
        "doCreateTrueObject", 
        "(Ljava/lang/String;)Lxerox/ilu/IluOInt;",
        jih
        );
    if (_ilujava_objectsFlag>1) {
        ilu_DebugPrintf("$ ObjectTableCreateObj joi=<%x>\n", joi);
    }
    /* conceptionally similar (locking would be wrong) to 
     * xerox_ilu_IluRT0_registerTrue(0, joi, jih, jServer);
     */
    if (joi) {
        jClass = unhand(joi)->jClassRep;
        cClass = Access(jClass, yIluClass, ilu_Class);
        cServer = Access(jServer, yIluServer, ilu_Server);
        Access(joi, yServer, ilu_Server) = cServer;
        if (_ilujava_objectsFlag>1) {
            ilu_DebugPrintf("$ ObjectTableCreateObj before FindOrCreateTrueObject cServer=<%x>\n", cServer);
        }
        cObject = ilu_FindOrCreateTrueObject(cih, cServer, cClass, joi);
        Access(joi, yIluKernelObject, ilu_Object) = cObject;
        if (_ilujava_objectsFlag>1) {
            ilu_DebugPrintf("$ ObjectTableCreateObj cObject=<%x>\n", cObject);
        }
        return cObject;
    }
    _iluJava_notUsed1 = (char *) jih;
    return 0;
}


INTERNALONLY void
_ilujava_ObjectTableFreeSelf(ilu_ObjectTable cot)
/* Upcalled by kernel to request deletion of table */
{
    execute_java_dynamic_method(EE(),
        (struct Hjava_lang_Object *) cot->ot_rock,
        "doObjectTableFreed", 
        "()V"
        );
    ilu_free(cot);
}


INTERNALONLY ilu_ObjectTable _ilujava_createOT(
	struct Hxerox_ilu_IluObjectTable * jot
	)
/* Create the native side of an object table */
{
    ilu_ObjectTable cot;
    if (jot==0) return 0;
    cot = ilu_must_malloc(sizeof(ilu_ObjectTable_s));
    cot->ot_object_of_ih = _ilujava_ObjectTableCreateObj;
    cot->ot_free_self = _ilujava_ObjectTableFreeSelf;
    cot->ot_rock = jot; /*not GC'ed because the server references it*/
    return cot;
}


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfCreateServer(
	struct Hxerox_ilu_IluServer * jServ, 
	struct Hjava_lang_String * jID
	)    
{
    ilu_ObjectTable cObjtab = 0; 
    ilu_Server cIluServer;
    char * cId = /* javaString.h [malloc] */ allocCString(jID); 
    struct Hxerox_ilu_IluObjectTable * jot = unhand(jServ)->jot;
    if (jot) {
        cObjtab = _ilujava_createOT(jot);
    }
    cIluServer = ilu_CreateTrueServer(cId, cObjtab, JAVALANGIDX);
    if (cIluServer==0) BUGPRINT("$ Couldn't create true server");
    Access(jServ, yIluServer, ilu_Server) = cIluServer;
    Access(jServ, yPtr, char *) = cId;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluServer_nativeOfCreateServer: true server created; j<%x> c<%x>  \n", jServ, cIluServer);
    }
} /*nativeOfCreateServer*/


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfDestroyServer(struct Hxerox_ilu_IluServer * jServ)
{
    ilu_Server cIluServer = Access(jServ, yIluServer, ilu_Server);
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluServer_nativeOfDestroyServer; j<%x> c<%x>\n", jServ, cIluServer);
    }
    ilu_BankServer(cIluServer);
} /*nativeOfDestroyServer*/


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfFreeServer(struct Hxerox_ilu_IluServer * jServ)
{
    ilu_Server cIluServer = Access(jServ, yIluServer, ilu_Server);
    char * cId = Access(jServ, yPtr, char *);
    Access(jServ, yPtr, char *) = 0;
    ilu_free(cId);
} /*nativeOfFreeServer*/

