/* IluJava_JOps.c */
/* Chris Jacobi, December 12, 1996 1:11 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_JOps.c,v 1.4 1996/12/12 21:30:39 jacobi Exp $ */

/*
 * Handy java access operations.
 * Implementation makes disgusting use of java implementation functionality
 */

#include <StubPreamble.h>
#include <interpreter.h>
#include <javaString.h>
#include "IluJava_JOps.h"


extern HArrayOfByte * 
IluJava_JAoB_from8(char * data, unsigned len) 
    /* Allocates and initializes a Java array of byte.
     * data: initial data interpreted as 8 bit bytes 
     * len: length of data (or array) in 8 bit units
     * Caller owns data. 
     */
{
    extern HArrayOfByte *MakeByteString(char *str, long len);
    HArrayOfByte * ja;
    /* MakeByteString from the java interpreter (classruntime.c) */
    ja = MakeByteString(data, len);
    if (ja == 0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "huh x ?");
    }
    return ja;
}


extern HArrayOfChar * 
IluJava_JAoC_from8(char * data, unsigned len) 
    /* Allocates and initializes a Java array of char
     * data: initial data interpreted as 8 bit characters 
     * len: length of data in 8 bit units
     * Caller owns data 
     */
{
    HArrayOfChar * ja;
    /* MakeString from interpreter.h (classruntime.c) */    
    ja = MakeString(data, len);
    if (ja == 0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "huh y ?");
    }
    return ja;
}


extern HArrayOfChar * 
IluJava_JAoC_from16(char * data, unsigned len) 
    /* Allocates and initializes a Java array of char.
     * data: initial data interpreted as 16 bit characters 
     * len: length of data in 16 bit units
     * Caller owns data 
     */
{
    extern HArrayOfChar *MakeUniString(unicode *str, long len);
    HArrayOfChar * ja;
    /* MakeUniString from the java interpreter (classruntime.c) */    
    ja = MakeUniString((unicode *) data, len);
    if (ja == 0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "huh z ?");
    }
    return ja;
}


extern void 
IluJava_JString_toC16(Hjava_lang_String * js, char * buffer, unsigned len) 
    /* Copies 16 bit characters from a java String.
     * buffer: "array" of 16 bit character units 
     * len: length of buffer in 16 bit units
     * Caller owns buffer  
     */
{
    /* javaString.h */
    javaString2unicode(js, (unicode *) buffer, len);
}
  
     
extern void 
IluJava_JString_toC8(Hjava_lang_String * js, char * buffer, unsigned len) 
    /* Copies characters from a java String.
     * Characters must fit in 8 bit units. 
     * buffer: array" of 8 bit character units 
     * len: length of buffer in 8 bit units
     * Caller owns buffer 
     */
{	
    /* javaString.h */
    javaString2CString(js, buffer, len);
}


extern void 
IluJava_JAoC_toC16(HArrayOfChar * ja, char * buffer, unsigned cnt) 
    /* Copies java array of char into a C 16-bit buffer.
     * buffer: "array" of 16 bit character units; must be long 
     * enough to receive all cnt chars plus terminating 0C unit. 
     * cnt: Number of 8 bit units to copy  (Must be >= 1 )
     * Caller owns buffer  
     * Ugly but checked November 26, 1996 11:07:36 pm PST 
     */
{
    /* compare to string.c::javaString2unicode */
    memcpy(buffer, unhand(ja)->body, cnt*2);
    buffer[cnt*2] = 0;
    buffer[cnt*2+1] = 0;
}
  
     
extern void 
IluJava_JAoC_toC8(HArrayOfChar * ja, char * buffer, unsigned cnt) 
    /* Copies java array of char into a C string buffer.
     * Restricted to 8 bit units of the java chars. 
     * buffer: string buffer; must be long enough to receive 
     * all cnt chars plus terminating 0C char.
     * cnt: Number of 8 bit units to copy  (Must be >= 1 )
     * Caller owns buffer; 
     * Ugly but checked November 26, 1996 11:04:38 pm PST 
     */
{
    /* compare to string.c::javaString2CString */
    unicode *src = unhand(ja)->body;
    unicode2str(src, buffer, cnt);  /* interpreter.h */
}	


extern struct Hjava_lang_String * 
IluJava_JString_from16(char * data, unsigned len) 
    /* Allocates and initializes a Java String 
     * data: initial data interpreted as 16 bit characters 
     * len: length of data (or string) in 16 bit units
     * Caller owns data. 
     */

{
    HArrayOfChar * ja;
    struct Hjava_lang_String * js;
    ja = IluJava_JAoC_from16(data, len);    
    /* classJavaLangString from interpreter.h */
    js = (struct Hjava_lang_String *) 
        execute_java_constructor(EE(), 0, classJavaLangString, "([C)", ja);
    if (js == 0) {
        SignalError(EE(), "xerox/ilu/IluSomeSystemException", "huh q ?");
    }
    return js;
}



