/* IluJava_JOps.h */
/* Chris Jacobi, November 26, 1996 11:09 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_JOps.h,v 1.2 1996/11/27 07:22:29 jacobi Exp $ */

/*
 * Handy java access operations.
 */

#include <StubPreamble.h>
#include <interpreter.h>
#include <javaString.h>


extern HArrayOfByte * 
IluJava_JAoB_from8(char * data, unsigned len); 
    /* Allocates and initializes a Java array of byte.
     * data: initial data interpreted as 8 bit bytes 
     * len: length of data (or array) in 8 bit units
     * Caller owns data. 
     */


extern HArrayOfChar * 
IluJava_JAoC_from8(char * data, unsigned len); 
    /* Allocates and initializes a Java array of char
     * data: initial data interpreted as 8 bit characters 
     * len: length of data in 8 bit units
     * Caller owns data 
     */


extern HArrayOfChar * 
IluJava_JAoC_from16(char * data, unsigned len); 
    /* Allocates and initializes a Java array of char.
     * data: initial data interpreted as 16 bit characters 
     * len: length of data in 16 bit units
     * Caller owns data 
     */


extern void 
IluJava_JString_toC16(Hjava_lang_String * js, char * buffer, unsigned len); 
    /* Copies 16 bit characters from a java String.
     * buffer: "array" of 16 bit character units 
     * len: length of buffer in 16 bit units
     * Caller owns buffer  
     */
  
     
extern void 
IluJava_JString_toC8(Hjava_lang_String * js, char * buffer, unsigned len); 
    /* Copies characters from a java String.
     * Characters must fit in 8 bit units. 
     * buffer: array" of 8 bit character units 
     * len: length of buffer in 8 bit units
     * Caller owns buffer 
     */


extern void 
IluJava_JAoC_toC16(HArrayOfChar * ja, char * buffer, unsigned len); 
    /* Copies java array of char into a C 16-bit buffer.
     * buffer: "array" of 16 bit character units; must be long 
     * enough to receive all cnt chars plus terminating 0C unit. 
     * cnt: Number of 8 bit units to copy  (Must be >= 1 )
     * Caller owns buffer  
     */
  
     
extern void 
IluJava_JAoC_toC8(HArrayOfChar * ja, char * buffer, unsigned cnt); 
    /* Copies java array of char into a C string buffer.
     * Restricted to 8 bit units of the java chars. 
     * buffer: string buffer; must be long enough to receive 
     * all cnt chars plus terminating 0C char.
     * cnt: Number of 8 bit units to copy  (Must be >= 1 )
     * Caller owns buffer; 
     */


extern struct Hjava_lang_String * 
IluJava_JString_from16(char * data, unsigned len); 
    /* Allocates and initializes a Java String 
     * data: initial data interpreted as 16 bit characters 
     * len: length of data (or string) in 16 bit units
     * Caller owns data. 
     */




