/* IluJava_pollwt.c */
/* Chris Jacobi, February 5, 1997 4:13 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_pollwt.c,v 1.3 1997/02/06 00:31:29 jacobi Exp $ */
 
/* 
 * WaitTech implementation with poll
 */


#include <iluxport.h>
#include <poll.h>
#include <errno.h>
#include "IluJava.h"

static int debug = 0;
static ilu_FineTime _ilujava_z = {0, 0};


static int pendingJavaException (void)
{
  struct execenv *ee = EE();
  return ee->exceptionKind;
}


INTERNALONLY int 
_ilujava_TimoutToMilli(ilu_FineTime * limit)
{
    if (limit) {
        ilu_FineTime now = ilu_FineTime_Now();
        ilu_FineTime wait;
        wait = ilu_FineTime_Sub(*limit, now);
        if (ilu_FineTime_Cmp(wait, _ilujava_z) > 0) {
            if (wait.ft_s < 1000000) {
                return wait.ft_s * 1000 +
                         (int) ilu_rescale(wait.ft_t, ilu_FineTimeRate, 1000);
            } else {
                return 1000000000;
            }
        } else {
            return 0;
        }
    } else {
        return -1;
    }
}


INTERNALONLY void 
_ilujava_Waiter(int fd, 
       int auxfd,
       ilu_boolean * sure,
       ilu_FineTime * limit, 
       ilu_boolean input,
       ILU_ERRS((interrupt)) * err)
{
    struct pollfd   p[2];
    int             timeout, res; 
    int             theerr = EAGAIN;
    unsigned long   nfds = 1;

    initialize_monitors(fd);
    if (auxfd != -1) {
        nfds = 2;
        initialize_monitors(auxfd);
    }
        /* 
         * Ignore the result; it is a fatal error if it returns FALSE
         * but the type for FALSE is not exported by java.
         *
         * If somebody knows how to fix the compile time warning correctly
         * please tell me.
         */

    if (pendingJavaException() != 0) {
        ilu_DebugPrintf("$ _ilujava_poll: pendingJavaException !! \n");
        *sure = ilu_FALSE;
        ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void)0);
        return;
    }
    
    while (theerr == EAGAIN) {
        p[0].fd = fd;
        p[0].events = (input ? POLLIN : POLLOUT) | POLLPRI;
        p[0].revents = 0;
        if (auxfd != -1) {
            p[1].fd = auxfd;
            p[1].events = POLLIN | POLLPRI;
            p[1].revents = 0;
        }
        timeout = _ilujava_TimoutToMilli(limit);
        
        if (debug>0) {
            ilu_DebugPrintf("$ _ilujava_poll: timeout %d  fd %d  pendingException()=>%d\n", timeout, p[0].fd,
			pendingJavaException());
	}
			
	/*
	while (pendingJavaException() != 0) {
            sysThreadYield();
            timeout = _ilujava_TimoutToMilli(limit);
	}
	*/

        res = poll(p, nfds, timeout);

        if (debug>0) {
            ilu_DebugPrintf("$ _ilujava_poll: returns %d\n", res);
        }
        
        if (res >= 0) {
            *sure = res > 0;
            ILU_CLER(*err);
            return;
        }
        theerr = errno;
        ilu_DebugPrintf("$ _ilujava_poll: errno %d\n", theerr);
        perror("_ilujava_poll ERROR IS");
    }
    if (!ilu_Check(theerr == EINTR, err)) {
        ilu_DebugPrintf("$ _ilujava_poll: errno=EINTR\n");
        return;
    }
    *sure = ilu_FALSE;
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void)0);
    if (debug>0) {
        ilu_DebugPrintf("$ _ilujava_poll: interrupted\n");
    }
    return;
}


INTERNALONLY void 
_ilujava_WaitIn(int fd, int auxfd, ilu_boolean * sure,
	ilu_FineTime * limit,
	ILU_ERRS((interrupt)) * err)
/* This is an ilu_FDWaitProc */
{
    if (debug) {
        ilu_DebugPrintf("$ _ilujava_poll: (in) %d\n", fd);
    }
    _ilujava_Waiter(fd, auxfd, sure, limit, ilu_TRUE, err);
}


INTERNALONLY void 
_ilujava_WaitOut(int fd, int auxfd, ilu_boolean * sure,
	 ilu_FineTime * limit,
	 ILU_ERRS((interrupt)) * err)
/* This is an ilu_FDWaitProc */
{
    if (debug) {
       ilu_DebugPrintf("$ _ilujava_poll: (out) %d\n", fd);
    }
    _ilujava_Waiter(fd, auxfd, sure, limit, ilu_FALSE, err);
}


INTERNALONLY ilu_WaitTech _iluJava_pollWT 
	= {_ilujava_WaitIn, _ilujava_WaitOut};


EXPORTLIMITED void
_ilujava_WaitTechInit()
{
    debug = _iluJava_getIntProp("ilu.debug.poll", 0);
    ilu_SetWaitTech(&_iluJava_pollWT);
}
