/* IluJava_selectwt.c */
/* Chris Jacobi, December 17, 1996 12:27 pm PST */


/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_selectwt.c,v 1.3 1996/12/17 21:13:11 jacobi Exp $ */
 
/*
 * WaitTech implementation with select
 * Alternative to IluJava_pollwt.c
 */


#include <iluxport.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#include <errno.h>
#include "IluJava.h"

static ilu_FineTime _ilujava_z0 = {0, 0};

INTERNALONLY void 
_ilujava_sWaiter(int fd, 
       int auxfd,
       ilu_boolean * sure,
       ilu_FineTime * limit, 
       ilu_boolean input,
       ILU_ERRS((interrupt)) * err)
{
    struct fd_set   rwfds; /*feels like lots of memory ??*/
    struct fd_set   excnfds;
    struct timeval* timeoutPtr = 0;
    struct timeval  timeout;
    int             res; 
    {

ilu_DebugPrintf("$ _ilujava_select THE AUXFD PARAMETER IS STLL IGNORED\n");
SYNTAX ERROR

        if (fd>=FD_SETSIZE) {
            _ilu_Assert(fd<FD_SETSIZE, "IluJava_selectwt: fd too large");
        }
        ILU_CLER(*err);
        FD_ZERO(&rwfds);
        FD_ZERO(&excnfds);
        FD_SET(fd, &excnfds);
        FD_SET(fd, &rwfds);
        if (limit) {
             ilu_FineTime    now = ilu_FineTime_Now();
             ilu_FineTime    wait = ilu_FineTime_Sub(*limit, now);
             ilu_DebugPrintf("$ _ilujava_select: USES TIMEOUT");
             if (ilu_FineTime_Cmp(wait, _ilujava_z0) > 0) {
                 timeout.tv_sec = wait.ft_s;
                 timeout.tv_usec = (int) ilu_rescale(wait.ft_t, ilu_FineTimeRate, 1000000);
                 timeoutPtr = &timeout;
             } 
        }
        ilu_DebugPrintf("$ _ilujava_select: calling fd=%d, input=%d\n", fd, input);
        res = select(fd+1, 
        	(input ? &rwfds : 0), 
        	(input ? 0 : &rwfds), 
        	&excnfds, 
        	timeoutPtr);
        if (res<0) {
            /* error */
            int theerr = errno;
            if (theerr == EINTR) {
            	/* thread has been asked to interrupt */
            	*sure = ilu_FALSE;
            	ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
            	return;
            } else {
            	/* other error */
            	*sure = ilu_FALSE;
            	ilu_Check(theerr == EINTR, err);
            	return;
            }
        }
        if (res=0) {
            /* time limit expired */
            *sure = ilu_FALSE;
            return;
        }
        if (res > 0) {
            /* somefile descriptor is ready */
            *sure = ilu_TRUE;
            return;
        }
    }
}


INTERNALONLY void 
_ilujava_SWaitIn(int fd, int auxfd, ilu_boolean * sure,
	ilu_FineTime * limit,
	ILU_ERRS((interrupt)) * err)
/* This is an ilu_FDWaitProc */
{
    ilu_DebugPrintf("$ _ilujava_poll: (in) %d\n", fd);
    _ilujava_sWaiter(fd, auxfd, sure, limit, ilu_TRUE, err);
}


INTERNALONLY void 
_ilujava_SWaitOut(int fd, int auxfd, ilu_boolean * sure,
	 ilu_FineTime * limit,
	 ILU_ERRS((interrupt)) * err)
/* This is an ilu_FDWaitProc */
{
    ilu_DebugPrintf("$ _ilujava_poll: (out) %d\n", fd);
    _ilujava_sWaiter(fd, auxfd, sure, limit, ilu_FALSE, err);
}


INTERNALONLY ilu_WaitTech _iluJava_selectWT 
	= {_ilujava_SWaitIn, _ilujava_SWaitOut};


EXPORTLIMITED void
_iluJava_WaitTechInit()
{
    ilu_SetWaitTech(&_iluJava_selectWT);
}

