/* IluMethodRep.java */
/* Chris Jacobi, January 3, 1997 2:08 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluMethodRep.java,v 1.7 1997/01/03 22:30:41 jacobi Exp $ */
 
/*
 * The registration of the method in the kernel is made
 * in IluJava_IluClassRep.c
 */

package xerox.ilu;

//import xerox.ilu.IluClassRep;
//import xerox.ilu.IluExceptionRep;
//import xerox.ilu.IluSkeleton

/**
 * Definition of a method.<br>
 * Used by stubs; unlikely to be useful for applications.
 * @see IluClassRep
 */
public class IluMethodRep {
    private long yIluMethod = 0;
    /*friends*/ xerox.ilu.IluClassRep jClassRep;
    public int methodIdx; //accessed by stub
    /*friends*/ String jIluMethodName;
    /*friends*/ int id;
    /*friends*/ boolean cacheable;
    /*friends*/ boolean asynchronous;
    /*friends*/ xerox.ilu.IluExceptionRep[] jExceptions;
    /*friends*/ xerox.ilu.IluSkeleton jSkeleton;
    	/* Implementation trick: IluMethodRep is stored
    	 * in the ilu_StubProc field of the _ilu_Method_s.
    	 * Requirement: IluMethodRep must not be garbage collected
    	 * met by stuffing the  IluMethodRep into the IluClassRep
    	 */
    
    /** 
     * Disable unspecified methods (reduces creation of bogus instances) 
     */
    private IluMethodRep() {};

    /** 
     * Stub visible creation of an IluMethodRep, except it doesn't yet 
     * perform any ilu kernel call 
     */
    public static IluMethodRep registerMethod (
            xerox.ilu.IluClassRep jClassRep,
            int methodIdx, 		//Internal indexing
            String iluMethodName,
            int id, 			//This is passed through the wire
            boolean cacheable,
            boolean asynchronous,
            xerox.ilu.IluExceptionRep[] exceptions,
            xerox.ilu.IluSkeleton skeleton
            ) {
        if (skeleton==null || jClassRep==null) {
            throw new java.lang.RuntimeException("missing arguments");
        }
        if (jClassRep.jMethods[methodIdx]!=null) {
            throw new java.lang.RuntimeException("ilu method redefined");
        }
        IluMethodRep m = new IluMethodRep();
        m.jClassRep = jClassRep;
        m.methodIdx = methodIdx;
        m.jIluMethodName = iluMethodName;
        m.id = id;
        m.cacheable = cacheable;
        m.asynchronous = asynchronous;
        //canonicalize exceptions (used by native code) 
        if (exceptions!=null) {
            if (exceptions.length==0) exceptions = null;
        }
        m.jExceptions = exceptions;
        m.jSkeleton = skeleton;
        jClassRep.jMethods[methodIdx] = m;
        return m;
    } //registerMethod
    
    /** useful for debugging */
    public java.lang.String toString() {
       return super.toString() + ":" + this.jIluMethodName;
    } //toString
    
} //IluMethodRep
