/* IluRT0.java */
/* Chris Jacobi, January 11, 1997 10:00 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluRT0.java,v 1.21 1997/01/12 06:04:42 jacobi Exp $ */ 
 
/* 
 * See also See also IluJava_IluRT0.c
 */

package xerox.ilu;
import xerox.ilu.IluClassRep;
import xerox.ilu.IluSystemException;
import xerox.ilu.IluObject;
import xerox.ilu.IluObject0;
import xerox.ilu.IluOInt;
import xerox.ilu.IluRuntimeError;

/**
 * Top level implementation of ilu features.  Applications rather use Ilu.java.
 */
public class IluRT0 {

    /** 
     * Initializes IluRT0 
     */
    public static void init() {
        IluInit.init();
    }
    
    
    /** 
     * Shortcut 
     */
    public static void throwErr(String s) {
        s = "**xerox.ilu.IluRT0." + s;
        IluDebug.clientPreError(s);
        throw new xerox.ilu.IluRuntimeError(s);
    } //throwErr
    
    
    /** 
     * Shortcut 
     */
    public static void panic(String s) {
        s = "**xerox.ilu.IluRT0." + s;
        IluDebug.panic(s);
    } //panic
    
    
    /** 
     * Importing an object.<p>
     * Object might be created as consequence of this call. 
     */
    public static xerox.ilu.IluObject  
    objectFromSBH (
    		String sbh, 
    		xerox.ilu.IluClassRep iluClass
    		) throws xerox.ilu.IluSystemException {
        xerox.ilu.IluOInt oi = oiFromSBH(sbh, iluClass);
        if (oi==null) {
            IluDebug.clientPreError("objectFromSBH finds no object");
            throw new xerox.ilu.IluNoObjectException(
                "no object for [sbh = '" + sbh + "']"
                );
        }
        return finishFromOI(oi);
    } //objectFromSBH
    
    
    /**
     * Native sub-part for objectFromSBH. <p> 
     * New "unfinished" IluOInt might be created.
     */
    /*friends*/ static native xerox.ilu.IluOInt  
    oiFromSBH (
        String sbh, 
        xerox.ilu.IluClassRep iluClass
    	) throws xerox.ilu.IluSystemException;
    

    /** 
     * Importing an object.<p>
     * Object might be created as consequence of this call. 
     */
    public static xerox.ilu.IluObject 
    lookupObject(
    		String sid, String ih, xerox.ilu.IluClassRep iluClass
 		) throws xerox.ilu.IluSystemException {
        xerox.ilu.IluOInt oi = oiFromLookup(sid, ih, iluClass);
        if (oi==null) {
            IluDebug.clientPreError("lookupObject finds no object");
            throw new xerox.ilu.IluNoObjectException(
                "no object for [sid = '" + sid + "' ih = '" + ih + "']"
                );
        }
        return finishFromOI(oi);
    } //lookupObject
    
    
    /**
     * Native sub-part for lookupObject <p>
     * New "unfinished" IluOInt might be created.
     */
    private static native xerox.ilu.IluOInt  
    oiFromLookup (
        String sid, 
        String ih, 
        xerox.ilu.IluClassRep iluClass
    	) throws xerox.ilu.IluSystemException;
    
    
    /**
     * Java sub-part of lookupObject <p>
     * Creates new IluOInt if necessary.
     */
    private static xerox.ilu.IluObject 
    finishFromOI (IluOInt oi) throws xerox.ilu.IluSystemException {
        xerox.ilu.IluObject obj = oi.getRealObject();
        //if the object exists it can be true or surrogate
        if (obj==null) {
            //but we only create surrogates here
            oi.createSurrogateForOI();
            obj = oi.getRealObject();
        }
        return obj;
    } //finishFromOI


    /**
     * Generates a string that's unique over space and time.
     * A server with nothing better to use might call this to 
     * get an ID.
     */
    public native static String  
    inventID();
    
    
    /** 
     * Literals for the lifetime argument of registerTrueObject.
     * Remember obj to prevent java garbage collection.
     * Forget obj to subject it to java garbage collection. 
     * Unspec requires ilu set a value depending on other arguments. 
     */
    public static final int unspec = 0;
    public static final int remember = 1;
    public static final int forget = 2;
    
    
    /**
     * Create true objects.
     * Life time is important and must be determined.  Normally the ilu
     * object stays alive until the intermediate object is garbage
     * collected.  Use any of a large number of schemes to prevent
     * garbage collection.
     * 
     * Return value is intermediate object.  Caller may prevent this
     *     from being garbage collected; See IluAnchorSupport.
     */
    public static Object 
    registerTrueObject(
    		String ih,
    		xerox.ilu.IluObject0 obj, 
    		xerox.ilu.IluServer server, 
    		IluClassRep iluClass,
    		int lifetime, //0 gives default behaviour,
    		xerox.ilu.IluAnchorSupport extraAnchor //null is ok
    		) throws IluSystemException {
        xerox.ilu.IluOInt joi;
        if (iluClass == null) {
            iluClass = IluClassAccess.iluClassFromJavaClass(obj.getClass());
            if (iluClass == null) {throwErr("IluClass of object unknown");}
        }
        if (ih == null) {
            if (obj instanceof IluIHProposer) {
                ih = ((IluIHProposer)obj).proposeIluIH();
            }
            if (ih == null) ih = inventID();
        }
        if (server == null) server = IluServer.defaultServer();
        server.finishInitialization(); 
	joi = xerox.ilu.IluOInt.newOI(obj, iluClass, server); 
	        //throws exception if object already known to Ilu
        if (joi.jClassRep == null) {
    	    throwErr("registerTrueObject jClassRep missing ");
        }
        registerTrue(joi, ih, server);
        if (joi.yIluKernelObject == 0) {
    	    throwErr("registerTrueObject: object already exists");
        }
        if (joi.jJavaObject == null) {
    	    panic("registerTrueObject jJavaObject not set");
        }
        if (joi.yServer == 0) {
    	    panic("registerTrueObject yServer not set");
        }
        if (extraAnchor != null) {
            extraAnchor.rememberAnchor(joi);
            if (lifetime==unspec) {lifetime = forget;}
        }
        if (extraAnchor != obj & obj instanceof IluAnchorSupport) {
            ((IluAnchorSupport) obj).rememberAnchor(joi);
            if (lifetime==unspec) {lifetime = forget;}
        }
        switch (lifetime) {
            case remember:
            case forget:
                break;
            default:
               if (iluClass.isCollectible()) {lifetime = forget;}
               else {lifetime = remember;}
        }
        switch (lifetime) {
            case remember:
                joi.retainOI();
                break;
            case forget:
                //joi.exposeOI(); not necessary as it never was retained
                break;
        }
        return joi;
    } //registerTrueObject
    
    
    /**
     * Native part for registerTrueObject
     */
    private static native void
    registerTrue(xerox.ilu.IluOInt joi, String ih, xerox.ilu.IluServer server);
    
    
    /** 
     * Publishes the SBH of the object in the local object domain
     */
    public static void
    publishTrueObject(xerox.ilu.IluObject0 obj0) 
            throws xerox.ilu.IluSystemException {
        xerox.ilu.IluObject obj = (IluObject) obj0;
        if (obj instanceof IluSurrogateObject) throwErr("is surrogate");
        IluOInt oi = IluOInt.peek(obj);
        if (oi == null) {
            //object unknown; try to register it right here and now...
            String ih = null;
            if (obj instanceof IluIHProposer) {
                ih = ((IluIHProposer)obj).proposeIluIH();
            }
            if (ih == null) {
                throwErr("publishTrueObject does not invent instance handles");
            }
            IluClassRep iluClass =
                IluClassAccess.iluClassFromJavaClass(obj.getClass());
            if (iluClass != null) {
                IluRT0.registerTrueObject(
                    ih, 	//ih
                    obj, 
                    null,	//server
                    iluClass,
                    0,  	//lifetime
                    null	//extraAnchor
                    );
                oi = IluOInt.peek(obj);
            }
            if (oi == null) {
                throwErr("object not registerd with Ilu " + obj);
            }
        }
        oi.publishOI();
    } //publishTrueObject
    
    
    /** 
     * withdrawObject is undo of publishTrueObject
     */
    public static void
    withdrawObject(xerox.ilu.IluObject0 obj) {
        if (obj instanceof IluSurrogateObject) throwErr("is surrogate");
        IluOInt oi = IluOInt.peek(obj);
        if (oi != null) oi.withdrawOI();
    } //withdrawObject
    
    
    /** 
     * May return NIL if the object's server isn't exported through any
     * port; may return an invalid SBH if the cached one references a
     * closed port.
     */
    public static String 
    sbhOfObject(xerox.ilu.IluObject0 obj) {
        IluOInt oi = IluOInt.peek((xerox.ilu.IluObject)obj);
        if (oi == null) return null;
        return oi.sbhOfOI();
    } //sbhOfObject
    
    
    /** 
     * Destroy ILU-ness of object.  Only for impatient applications,
     * normal applications leave this to the garbage collector.
     * Do NOT use from applets; error behavior is not yet defined.
     *
     * True objects only.
     */
    public static void
    destroyObject(xerox.ilu.IluObject0 obj) {
        if (obj instanceof IluSurrogateObject) throwErr("is surrogate");
        IluOInt oi = IluOInt.peek(obj);
        if (oi != null) oi.destroyRudeOI();
    } //destroyObject
    
} // IluRT0
