/* IluSystemException0.java */
/* Chris Jacobi, December 19, 1996 11:26 am PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */ 
 

/*
 * $Id: IluSystemException0.java,v 1.3 1996/12/19 20:35:20 jacobi Exp $
 */

package xerox.ilu;

/**
 * IluSystemException0 is an artificial superclass of IluSystemException
 * - Clients may use this to catch all ilu system exceptions,
 *   but since all system exceptions are also IluSystemException's,
 *   clients can as well totally ignore this class.
 * - Clients must not throw this exceptions. 
 * - Clients must not declare methods to throw this exceptions. 
 */
public class IluSystemException0 extends xerox.ilu.IluException {

    public IluSystemException0() {
        super();
        this.theExceptionType = 2;
    }

    public IluSystemException0(String s) {
        super(s);
        this.theExceptionType = 2;
    }
    
    static public final int completedYes = 0;
    static public final int completedNo = 1;
    static public final int completedMaybe = 2;
    
    private int theCompleted = 0;
    private int theMinor = 0;
    
    public int getCompleted() { 
        return theCompleted;
    }
    
    public xerox.ilu.IluSystemException setCompleted(int completed) {
        if (completed>completedMaybe) completed = completedMaybe;
        theCompleted = completed;
        return (xerox.ilu.IluSystemException) this;
    }

    public int getMinor() { 
        return theMinor;
    }
    
    public xerox.ilu.IluSystemException setMinor(int minor) { 
        theMinor = minor;
        return (xerox.ilu.IluSystemException) this;
    }
    
    public String toString() {
	String superString = super.toString();
	String minorCodeString = "  minor code: " + theMinor;
	String completedString = null;
	switch (theCompleted) {
	case completedYes:
	    completedString = "  completed: yes";
	    break;
	case completedNo:
	    completedString = "  completed: no";
	    break;
	case completedMaybe:
	default:
	    completedString = " completed: maybe";
	    break;
	}
	return superString + minorCodeString + completedString;
    }


} // IluSystemException0
