//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: bnell $
// $Date: 1996/11/13 16:57:30 $
// $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/ilu-scm.cc,v $
// $Revision: 1.2 $
//

#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
extern "C" {
#include <iluhash.h>
}
#include "ilu-scm-private.h"
#include "ilusrvr-scm.h"
#include "iluobj-scm.h"

#if 0
static int stupid = 0;
static SCM* stupid_array = 0;

SCM* alloc_scm_obj()
{
  if(!stupid_array) {
    stupid_array = scheme::make_array(1024);
  }
  assert(stupid < 1024);
  return &stupid_array[stupid++];
}
#else
#define alloc_scm_obj() scheme::make_array(1)
#endif

////////////////////////////////////////////////////////////////////////
//
// from class ilu
//
////////////////////////////////////////////////////////////////////////

static ilu_boolean scheme_registered = ilu_FALSE;
static ilu_cardinal scheme_lang = 2112; // init to some bogus number

// ok
ilu_cardinal ilu_scheme_lang_idx()
{
  if(!scheme_registered) {
    scheme_lang = ilu_RegisterLanguage("Scheme");
    scheme_registered = ilu_TRUE;
  }
  return scheme_lang;
}

// ok
SCM ilu__scheme_lang_idx()
{
  return scheme_value<ilu_cardinal>(ilu_scheme_lang_idx());
}

// ok
void ilu_full_prefer_success(ilu_Error* e, const char* atf, unsigned int atl)
{
  const char     *n, *f;
  int             l;

  if (ILU_ERROK(*e))
    return;
  n = ILU_ERR_NAME(*e);
  f = ilu_ErrorFile(e);
  l = ilu_ErrorLine(e);
  ILU_ERRPRINTF("ILU warning: error %s from %s:%d simplified at %s:%d\n",
		n, f, l, atf, atl);
}


// ok
SCM ilu__full_prefer_success(SCM _e, SCM _atf, SCM _atl)
{
  scheme_ptr<ilu_Error> e = _e;
  scheme_tmp_string atf = _atf;
  scheme_value<ilu_cardinal> atl = _atl;
  ilu_full_prefer_success(e, atf, atl);
  return scheme::True;
}

static ilu_Mutex otmu = ILU_NIL;

// ok
SCM ilu__enter_ot_mu()
{
  ILU_ERRS((internal, no_memory)) lerr;
  if (otmu == ILU_NIL)
    otmu = ilu_GetOTMutex();
  ilu_EnterMutex(otmu, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return scheme::True;
}

// ok
SCM ilu__exit_ot_mu()
{
#ifdef __GNUC__
  ILU_ERRS((internal, no_memory)) lerr = ILU_INIT_NO_ERR;
#else
  ILU_ERRS((internal, no_memory)) lerr = ILU_NO_ERR;
#endif

  if (otmu == ILU_NIL)
    otmu = ilu_GetOTMutex();
  ilu_ExitMutex(otmu, ilu_FALSE, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return scheme::True;
}


// ok
SCM ilu__object_type_defined(SCM _t)
{
  scheme_ilu_ptr<ilu_Class> t = _t;
  ILU_ERRS((internal, no_memory)) lerr;
  ilu_ObjectTypeDefined(t, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return scheme::True;
}

// caller owns strings
SCM ilu__define_exception(SCM _i, SCM _e)
{
  scheme_tmp_string i = _i;   // not sure about deleting this!
  scheme_tmp_string e = _e;   // not sure about deleting this!
  ILU_ERRS((internal, no_memory)) lerr;
  ilu_Exception   ans;
  ans = ilu_DefineException(i, e, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return scheme_tmp_string(ans);
}

SCM ilu__define_object_type(SCM _cl_name,
			    SCM _cl_brand,
			    SCM _cl_unique_id,
			    SCM _cl_singleton,
			    SCM _cl_optional,
			    SCM _cl_collectible,
			    SCM _cl_authentication,
			    SCM _cl_method_count,
			    SCM _cl_scls_count,
			    SCM _cl_scls_ids)
{
  scheme_string cl_name = _cl_name;
  scheme_string cl_brand = _cl_brand;
  scheme_string cl_unique_id = _cl_unique_id;
  scheme_string cl_singleton = _cl_singleton;
  scheme_ilu_boolean cl_optional = _cl_optional;
  scheme_ilu_boolean cl_collectible = _cl_collectible;
  scheme_string cl_authentication = _cl_authentication;
  scheme_value<ilu_cardinal> cl_method_count = _cl_method_count;
  scheme_value<ilu_cardinal> cl_scls_count = _cl_scls_count;
  scheme_vec<char*> cl_scls_ids = _cl_scls_ids;
  ILU_ERRS((internal, no_memory)) lerr;
  ilu_Class ans;
  ans = ilu_DefineObjectType(cl_name, cl_brand, cl_unique_id,
			     cl_singleton, cl_optional,
			     cl_collectible, cl_authentication,
			     cl_method_count, cl_scls_count,
			     cl_scls_ids, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);

  // must delete these??
  for(unsigned int i = 0; i < cl_scls_ids.length(); i++)
    delete[] cl_scls_ids[i];

  return scheme_ilu_ptr<ilu_Class>(ans);
}


// caller owns me_name and me_exceptionVector
SCM ilu__define_method(SCM _c,
		       SCM _i,
		       SCM _me_name,
		       SCM _me_id,
		       SCM _me_cacheable,
		       SCM _me_asynchronous,
		       SCM _me_exceptionCount,
		       SCM _me_exceptionVector)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  scheme_value<ilu_cardinal> i = _i;
  scheme_string me_name = _me_name;
  scheme_value<ilu_cardinal> me_id = _me_id;
  scheme_ilu_boolean me_cacheable = _me_cacheable;
  scheme_ilu_boolean me_asynchronous = _me_asynchronous;
  scheme_value<ilu_cardinal> me_exceptionCount = _me_exceptionCount;
  scheme_vec<ilu_Exception> me_exceptionVector = _me_exceptionVector;

  ILU_ERRS((internal, no_memory)) lerr;
  (void) ilu_DefineMethod(c, i, me_name, me_id, me_cacheable,
			  me_asynchronous, me_exceptionCount,
			  me_exceptionVector, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);

  return scheme::True;
}


// call is an output parameter (must allocate call object!)
// returns either a pair (error . call)
SCM ilu__start_call(SCM _call, SCM _s, SCM _intro_type, SCM _method)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_ptr<ilu_Server> s = _s;
  scheme_ilu_ptr<ilu_Class> intro_type = _intro_type;
  scheme_ilu_ptr<ilu_Method> method = _method;
  ilu_Connection newconn;

  ilu_StartCall(scm_call_call(call), s, intro_type, method, ilu_scheme_lang_idx(),
		(ilu_Passport) 0, &newconn, scm_call_err(call));
  if (newconn != NULL)
    ilu_server_monitor_conn(newconn);

  return scheme::cons(scheme_ilu_ptr<ilu_SCMCall>(call),
		      scheme_ilu_boolean((ilu_boolean)ILU_ERROK(*scm_call_err(call))));
}

/*Main Invariant holds, L2 otherwise unconstrained*/
SCM ilu__begin_sizing_exception(SCM _call, SCM _eIndex)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_integer> eIndex = _eIndex;
  ilu_cardinal ans;
  ans = ilu_BeginSizingException(scm_call_call(call), eIndex, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


/*L1_sup < cmu*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
   after: L2     >=   {call's conn's callmu, iomu} if result is true,
	  L2 disjoint {call's conn's callmu, iomu} if result is false. */
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest not registered.*/
SCM ilu__begin_exception(SCM _call, SCM _eIndex, SCM _argSize)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_integer> eIndex = _eIndex;
  scheme_value<ilu_cardinal> argSize = _argSize;
  ilu_Connection  conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean     ans;
  if (ilu_ThreadPerRequest(conn))
    ilu_server_disable_requests(call, conn);
  ans = ilu_BeginException(scm_call_call(call), eIndex, argSize, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean((ilu_boolean)(ans != 0));
}

SCM ilu__start_request(SCM _call, SCM _argSize)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> argSize = _argSize;
  ilu_boolean     ans;
  ans = ilu_StartRequest(scm_call_call(call), argSize, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean((ilu_boolean)(ans != 0));
}

/*Main Invariant holds, L2 otherwise unconstrained*/
SCM ilu__begin_sizing_reply(SCM _call, SCM _exns_possible)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean exns_possible = _exns_possible;
  ilu_cardinal    ans;
  ans = ilu_BeginSizingReply(scm_call_call(call), exns_possible, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

/*L1_sup < cmu*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
   after: L2     >=   {call's conn's callmu, iomu} if result is true,
	  L2 disjoint {call's conn's callmu, iomu} if result is false. */
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest not registered.*/
SCM ilu__begin_reply(SCM _call, SCM _exceptions, SCM _argSize)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean exceptions = _exceptions;
  scheme_value<ilu_cardinal> argSize = _argSize;
  ilu_Connection  conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean     ans;
  if (ilu_ThreadPerRequest(conn))
    ilu_server_disable_requests(call, conn);
  ans = ilu_BeginReply(scm_call_call(call), exceptions, argSize, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean((ilu_boolean)(ans != 0));
}


SCM ilu__end_sequence(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_EndSequence(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__end_union(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_EndUnion(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__end_record(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_EndRecord(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__end_array(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_EndArray(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

// exceptions are strings...
SCM ilu__exception_of_method(SCM _method, SCM _index)
{
  scheme_ilu_ptr<ilu_Method> method = _method;
  scheme_value<ilu_cardinal> index = _index;
  ilu_Exception ex = ilu_ExceptionOfMethod(method, index);
  return scheme_tmp_string(ex);
}

SCM ilu__exception_count_of_method(SCM _m)
{
  scheme_ilu_ptr<ilu_Method> m = _m;
  ilu_cardinal c = m->me_exceptionCount;
  return scheme_value<ilu_cardinal>(c);
}

SCM ilu__get_object_class(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  return scheme_ilu_ptr<ilu_Class>(ilu_ClassOfObject(obj));
}

ilu_Server ilu_get_object_server(ilu_Object obj)
{
  return ilu_ServerOfObject(obj);
}


SCM ilu__get_object_server(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  return scheme_ilu_ptr<ilu_Server>(ilu_get_object_server(obj));
}


/*Main Invariant holds*/
/*before: L2 >= {call's conn's callmu, iomu};
  after: success <=> L2 >= {call's conn's callmu, iomu}*/
SCM ilu__reply_read(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_ReplyRead(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  ilu_boolean ok = ILU_ERROK(*scm_call_err(call)) ? ilu_TRUE : ilu_FALSE;
  return scheme_ilu_boolean(ok);
}

// free call object
SCM ilu__finish_call(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_FinishCall(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  if (scm_call_call(call)->ca_incoming && !scm_call_call(call)->ca_reqs_enabled)
    ilu_server_enable_requests(call, ilu_ConnectionOfCall(scm_call_call(call)));
  ilu_boolean ok = ILU_ERROK(*scm_call_err(call)) ? ilu_TRUE : ilu_FALSE;
  return scheme_ilu_boolean(ok);
}


/*Main Invariant holds*/
/*L2    >=    {call's conn's callmu, iomu} before,
 *L2 disjoint {call's conn's callmu, iomu} after*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest not registered;
  after:  _ILU_C_ReadServiceRequest registered.*/
SCM ilu__finish_exception(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Connection conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean ans = ilu_FinishException(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  ilu_server_enable_requests(call, conn);
  return scheme_ilu_boolean(ans);
}


SCM ilu__finish_request(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_boolean ans = ilu_FinishRequest(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean(ans);
}

/*Main Invariant holds*/
/*L2    >=    {call's conn's callmu, iomu} before,
 *L2 disjoint {call's conn's callmu, iomu} after*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest not registered;
  after:  _ILU_C_ReadServiceRequest registered.*/
SCM ilu__finish_reply(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Connection conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean ans = ilu_FinishReply(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  ilu_server_enable_requests(call, conn);
  return scheme_ilu_boolean(ans);
}


SCM ilu__no_reply(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Connection  conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean ans = ilu_NoReply(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  if (!ilu_ThreadPerRequest(conn))
    ilu_server_enable_requests(call, conn);
  return scheme::True;
}


/*Main invariant holds
 *before: L2     >=  {call's conn's callmu, iomu}
 *after:  L2 not >= {call's conn's         iomu},
 *after:  L2     >= {call's conn's callmu} iff protocol not concurrent*/
/*If not threaded:
  before: (iluServer)ReadServiceRequest not registered;
  after:  (iluServer)ReadServiceRequest registered iff protocol concurrent.*/
SCM ilu__finish_parameters(SCM _call, SCM _obj)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Connection  conn = ilu_ConnectionOfCall(scm_call_call(call));
  ilu_boolean ans = ilu_RequestRead(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  if (ilu_ThreadPerRequest(conn))
    ilu_server_enable_requests(call, conn);
  if (_obj == scheme::False) {
    if (ilu_BeginException(scm_call_call(call), 0,
			   ilu_ProtocolException_GarbageArguments,
			   scm_call_err(call)))
      ilu_FinishException(scm_call_call(call), scm_call_err(call));
    ilu_prefer_success(scm_call_err(call));
    return scheme::False;
  }
  return scheme_ilu_boolean(ans);
}


SCM ilu__input_byte(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_byte b2;
  ilu_InputByte(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_byte>(b2);
}

SCM ilu__input_boolean(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_boolean b2;
  ilu_InputBoolean(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean(b2);
}

SCM ilu__input_optional(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_boolean b2;
  ilu_InputOptional(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_boolean(b2);
}


SCM ilu__input_cardinal(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_cardinal b2;
  ilu_InputCardinal(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(b2);
}


SCM ilu__input_character(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_character b2;
  ilu_InputCharacter(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_character>(b2);
}


SCM ilu__input_enum(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_shortcardinal b2;
  ilu_InputEnum(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_shortcardinal>(b2);
}


SCM ilu__input_integer(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_integer b2;
  ilu_InputInteger(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_integer>(b2);
}


SCM ilu__input_real(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  double b;
  ilu_InputReal(scm_call_call(call), &b, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  ilu_real b2 = b;
  return scheme_value<ilu_real>(b2);
}

SCM ilu__input_short_cardinal(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_shortcardinal b2;
  ilu_InputShortCardinal(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_shortcardinal>(b2);
}

SCM ilu__input_short_integer(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_shortinteger b2;
  ilu_InputShortInteger(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_shortinteger>(b2);
}

SCM ilu__input_short_real(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  float b;
  ilu_InputShortReal(scm_call_call(call), &b, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  ilu_shortreal b2 = b;
  return scheme_value<ilu_shortreal>(b2);
}

SCM ilu__input_long_cardinal(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_longcardinal b2;
  ilu_InputLongCardinal(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_longcardinal(b2);
}


SCM ilu__input_long_integer(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_longinteger b2;
  ilu_InputLongInteger(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_longinteger(b2);
}


SCM ilu__input_long_real(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_longreal b2;
  ilu_InputLongReal(scm_call_call(call), &b2, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_longreal(b2);
}

SCM ilu__input_opaque(SCM _call, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> len = _len;
  ilu_opaque buf = 0;
  ilu_InputOpaque(scm_call_call(call), &buf, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  SCM ret = scheme_vec<ilu_byte>((ilu_byte*)buf, len);
  ilu_free(buf);
  return ret;
}

SCM ilu__input_bytes(SCM _call, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_byte* buf = 0;
  ilu_cardinal len = 0;
  ilu_InputBytes(scm_call_call(call), &buf, &len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  SCM ret = scheme_vec<ilu_byte>(buf, len);
  ilu_free(buf);
  return ret;
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__input_string(SCM _call, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal len = 0;
  ilu_string buf = 0;
  ilu_InputString(scm_call_call(call), &buf, &len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  SCM ret = scheme_tmp_string(buf);
  ilu_free(buf);  // must be deleted by ilu
  return ret;
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__input_string_vec(SCM _call, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> len = _len;
  ilu_string buf = 0;
  ilu_InputStringVec(scm_call_call(call), &buf, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  SCM tmp = scheme_vec<ilu_shortcharacter>((unsigned char*)buf, len);
  ilu_free(buf);
  return tmp;
}



// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__input_w_string(SCM _call, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal len = 0;
  ilu_wstring buf = 0;
  ilu_InputWString(scm_call_call(call), &buf, &len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_ilu_wstring(buf);
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__input_w_string_vec(SCM _call, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> len = _len;
  ilu_wstring buf = 0;
  ilu_InputWStringVec(scm_call_call(call), &buf, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  SCM tmp = scheme_vec<ilu_character>(buf, len);
  ilu_free(buf);
  return tmp;
}

ilu_Object ilu_input_object_id(ilu_SCMCall call,
			       ilu_boolean discriminator_p,
			       ilu_Class putative_class)
{
  ilu_Object obj = NULL;
  ilu_InputObjectID(scm_call_call(call), &obj,
		    discriminator_p ? ilu_TRUE : ilu_FALSE,
		    putative_class, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  /* now obj != NULL => Inside(obj->server, putative_class) */
  return obj;
}



SCM ilu__input_object_id(SCM _call, SCM _discriminator_p, SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean discriminator_p = _discriminator_p;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_Object obj = ilu_input_object_id(call, discriminator_p, putative_class);
  return scheme_ilu_ptr<ilu_Object>(obj);
}

SCM ilu__input_sequence(SCM _call, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal c2 = 0;
  ilu_InputSequence(scm_call_call(call), &c2, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(c2);
}


SCM ilu__input_union(SCM _call, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal d2 = 0;
  ilu_InputUnion(scm_call_call(call), &d2, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(d2);
}


SCM ilu__input_record(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_InputRecord(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__input_array(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_InputArray(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


/*after: Inside(ks, c) */
SCM ilu__enter_server(SCM _ks, SCM _c)
{
  scheme_ilu_ptr<ilu_Server> ks = _ks;
  scheme_ilu_ptr<ilu_Class> c = _c;
  ilu_EnterServer(ks, c);
  return scheme::True;
}

/*before: Inside(obj->server, obj->class) */
void ilu_exit_server(ilu_Server ks, ilu_Class c)
{
  ilu_ExitServer(ks, c);
}

/*before: Inside(obj->server, obj->class) */
SCM ilu__exit_server(SCM _ks, SCM _c)
{
  scheme_ilu_ptr<ilu_Server> ks = _ks;
  scheme_ilu_ptr<ilu_Class> c = _c;
  ilu_ExitServer(ks, c);
  return scheme::True;
}


SCM ilu__sbh_to_object(SCM _sbh, SCM _c)
{
  scheme_tmp_string sbh = _sbh;
  scheme_ilu_ptr<ilu_Class> c = _c;
  ilu_Object obj;
  ILU_ERRS((bad_locks, broken_locks, inv_objref, no_memory, internal)) lerr;
  SCM o;

  obj = ilu_ObjectOfSBH(sbh, c, &lerr);
  ilu_prefer_success(&lerr);
  if (obj == NULL)
    return scheme::False;

  /* now Inside(obj->server, c) */
  o = (SCM)ilu_GetLanguageSpecificObject(obj, ilu_scheme_lang_idx());

  if (o == scheme::False)
    o = ilu_object_create_from_registry(ilu_ClassOfObject(obj), obj);

  ilu_ExitServer(ilu_ServerOfObject(obj), c);

  return o;
}


/* before: Inside(obj->server, obj->class) */
SCM ilu__sbh_of_object(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  ilu_Server s = ilu_ServerOfObject(obj);
  ilu_Class c = ilu_ClassOfObject(obj);
  ilu_string result;
  result = ilu_SBHOfObject(obj);
  ilu_ExitServer(s, c);
  return scheme_tmp_string(result);
}


SCM ilu__output_byte(SCM _call, SCM _byte)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_byte> byte = _byte;
  ilu_OutputByte(scm_call_call(call), byte, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_boolean(SCM _call, SCM _b)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean b = _b;
  ilu_OutputBoolean(scm_call_call(call), b, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_optional(SCM _call, SCM _b)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean b = _b;
  ilu_OutputOptional(scm_call_call(call), b, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_cardinal(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> v = _v;
  ilu_OutputCardinal(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_character(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_character> v = _v;
  ilu_OutputCharacter(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_enum(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortcardinal> v = _v;
  ilu_OutputEnum(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_integer(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_integer> v = _v;
  ilu_OutputInteger(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_real(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_real> v = _v;
  ilu_OutputReal(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_short_integer(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortinteger> v = _v;
  ilu_OutputShortInteger(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_short_real(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortreal> v = _v;
  ilu_OutputShortReal(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_long_cardinal(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longcardinal v = _v;
  ilu_OutputLongCardinal(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_long_integer(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longinteger v = _v;
  ilu_OutputLongInteger(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_long_real(SCM _call, SCM _v)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longreal v = _v;
  ilu_OutputLongReal(scm_call_call(call), v, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__output_bytes(SCM _call, SCM _buff, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_byte> bytes = _buff;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_OutputBytes(scm_call_call(call), bytes, len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__output_opaque(SCM _call, SCM _buff, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_byte> bytes = _buff;
  scheme_value<ilu_cardinal> len = _len;
  ilu_OutputOpaque(scm_call_call(call), bytes, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__output_string(SCM _call, SCM _buff, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_string buff = _buff;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_OutputString(scm_call_call(call), buff, len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__output_string_vec(SCM _call, SCM _buff, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_shortcharacter> buff = _buff;
  scheme_value<ilu_cardinal> len = _len;
  ilu_OutputStringVec(scm_call_call(call), (ilu_string)(unsigned char*)buff, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


// XXXXXXXXXXXXXXXXXXX
SCM ilu__output_w_string(SCM _call, SCM _buf, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_wstring buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_string buf2 = 0;
  ilu_cardinal buf2len;
#ifdef __linux__
  ilu_OutputWString(scm_call_call(call), buf, len, limit, scm_call_err(call));
#else
  ilu_OutputWString(scm_call_call(call), buf, len, limit, &buf2, &buf2len, scm_call_err(call));
  ilu_free(buf2);
#endif
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

// XXXXXXXXXXXXXXXXXXX
SCM ilu__output_w_string_vec(SCM _call, SCM _buf, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_character> buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  ilu_string buf2;
  ilu_cardinal buf2len;
#ifdef __linux__
  ilu_OutputWStringVec(scm_call_call(call), buf, len, scm_call_err(call));
#else
  ilu_OutputWStringVec(scm_call_call(call), buf, len, &buf2, &buf2len, scm_call_err(call));
  ilu_free(buf2);
#endif
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


/* before: obj != NULL implies Inside(obj->server, obj->class) */
ilu_boolean ilu_output_object_id(ilu_SCMCall call, ilu_Object obj,
				 ilu_boolean discriminator_p,
				 ilu_Class putative_class)
{
  ilu_OutputObjectID(scm_call_call(call), obj,
		     discriminator_p ? ilu_TRUE : ilu_FALSE,
		     putative_class, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return ilu_TRUE;
}

/* before: obj != NULL implies Inside(obj->server, obj->class) */
SCM ilu__output_object_id(SCM _call, SCM _obj,
			  SCM _discriminator_p,
			  SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  scheme_ilu_boolean discriminator_p = _discriminator_p;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_boolean b;
  b = ilu_output_object_id(call, obj, discriminator_p, putative_class);
  return scheme_ilu_boolean(b);
}


SCM ilu__output_sequence(SCM _call, SCM _length, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> length = _length;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_OutputSequence(scm_call_call(call), length, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_union(SCM _call, SCM _discriminator, SCM _discriminator_size)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> discriminator = _discriminator;
  scheme_value<ilu_cardinal> discriminator_size = _discriminator_size;
  ilu_OutputUnion(scm_call_call(call), discriminator, discriminator_size, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__output_array(SCM _call)
{
  /***** this 0 needs to change to the actual number of elements to output */
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_OutputArray(scm_call_call(call), 0, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}

SCM ilu__output_record(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_OutputRecord(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::True;
}


SCM ilu__size_of_byte(SCM _call, SCM _byte)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<unsigned char> byte = _byte;
  ilu_cardinal ans = ilu_SizeOfByte(scm_call_call(call), byte, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_boolean(SCM _call, SCM _b)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean b = _b;
  ilu_cardinal ans = ilu_SizeOfBoolean(scm_call_call(call), b, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_optional(SCM _call, SCM _opt)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean opt = _opt;
  ilu_cardinal ans = ilu_SizeOfOptional(scm_call_call(call), opt, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_cardinal(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> val = _val;
  ilu_cardinal ans = ilu_SizeOfCardinal(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_character(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_character> val = _val;
  ilu_cardinal ans = ilu_SizeOfCharacter(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_enum(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortcardinal> val = _val;
  ilu_cardinal ans = ilu_SizeOfEnum(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_integer(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_integer> val = _val;
  ilu_cardinal ans = ilu_SizeOfInteger(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_real(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_real> val = _val;
  ilu_cardinal ans = ilu_SizeOfReal(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_short_cardinal(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortcardinal> val = _val;
  ilu_cardinal ans = ilu_SizeOfShortCardinal(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_short_integer(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortinteger> val = _val;
  ilu_cardinal ans = ilu_SizeOfShortInteger(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_short_real(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_shortreal> val = _val;
  ilu_cardinal ans = ilu_SizeOfShortReal(scm_call_call(call), (float)val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_long_cardinal(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longcardinal val = _val;
  ilu_cardinal ans = ilu_SizeOfLongCardinal(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_long_integer(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longinteger val = _val;
  ilu_cardinal ans = ilu_SizeOfLongInteger(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_long_real(SCM _call, SCM _val)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_longreal val = _val;
  ilu_cardinal ans = ilu_SizeOfLongReal(scm_call_call(call), val, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__size_of_bytes(SCM _call, SCM _buf, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_byte> buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal ans = ilu_SizeOfBytes(scm_call_call(call), buf, len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__size_of_opaque(SCM _call, SCM _buf, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_byte> buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  ilu_cardinal ans = ilu_SizeOfOpaque(scm_call_call(call), buf, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__size_of_string(SCM _call, SCM _buf, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_string buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> lim = _limit;
  ilu_cardinal ans = ilu_SizeOfString(scm_call_call(call), buf, len, lim, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
SCM ilu__size_of_string_vec(SCM _call, SCM _buf, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_shortcharacter> buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  ilu_cardinal ans;
  ans = ilu_SizeOfStringVec(scm_call_call(call), (ilu_string)(unsigned char*)buf, len, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

// XXXXXXXXXXXXXXXXXX
SCM ilu__size_of_w_string(SCM _call, SCM _buf, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_wstring buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_string buf2;
  ilu_cardinal buf2len;
  ilu_cardinal ans;
#ifdef __linux__
  ans = ilu_SizeOfWString(scm_call_call(call), buf, len, limit, scm_call_err(call));
#else
  ans = ilu_SizeOfWString(scm_call_call(call),buf,len,limit,&buf2,&buf2len,scm_call_err(call));
  ilu_free(buf2);
#endif
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

// XXXXXXXXXXXXXXXXXX
SCM ilu__size_of_w_string_vec(SCM _call, SCM _buf, SCM _len)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_vec<ilu_character> buf = _buf;
  scheme_value<ilu_cardinal> len = _len;
  ilu_string buf2;
  ilu_cardinal buf2len;
  ilu_cardinal ans;
#ifdef __linux__
  ans = ilu_SizeOfWStringVec(scm_call_call(call), buf, len, scm_call_err(call));
#else
  ans = ilu_SizeOfWStringVec(scm_call_call(call),buf,len,&buf2,&buf2len,scm_call_err(call));
  ilu_free(buf2);
#endif
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}



/* before: obj != NULL implies Inside(obj->server, obj->class) */
ilu_cardinal ilu_size_of_object_id(ilu_SCMCall call, ilu_Object obj,
				   ilu_boolean discriminator_p,
				   ilu_Class putative_class)
{
  ilu_cardinal size;
  size = ilu_SizeOfObjectID(scm_call_call(call), obj,
			    discriminator_p ? ilu_TRUE : ilu_FALSE,
			    putative_class, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  if (obj != NULL)
    ilu_ExitServer(ilu_ServerOfObject(obj), ilu_ClassOfObject(obj));

  return size;
}


/* before: obj != NULL implies Inside(obj->server, obj->class) */
SCM ilu__size_of_object_id(SCM _call, SCM _obj, SCM _discriminator_p,
			   SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  scheme_ilu_boolean discriminator_p = _discriminator_p;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_cardinal size;
  size = ilu_size_of_object_id(call, obj, discriminator_p, putative_class);
  return scheme_value<ilu_cardinal>(size);
}

SCM ilu__size_of_sequence(SCM _call, SCM _len, SCM _limit)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> len = _len;
  scheme_value<ilu_cardinal> limit = _limit;
  ilu_cardinal ans;
  ans = ilu_SizeOfSequence(scm_call_call(call), len, limit, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

SCM ilu__size_of_union(SCM _call, SCM _discriminator, SCM _discriminator_size)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_value<ilu_cardinal> discriminator = _discriminator;
  scheme_value<ilu_cardinal> discriminator_size = _discriminator_size;
  ilu_cardinal ans;
  ans = ilu_SizeOfUnion(scm_call_call(call), discriminator, discriminator_size, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_array(SCM _call)
{
  /****** this 0 needs to change to the actual number of elements to output */
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_cardinal ans;
  ans = ilu_SizeOfArray(scm_call_call(call), 0, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}


SCM ilu__size_of_record(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_cardinal ans = ilu_SizeOfRecord(scm_call_call(call), scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<ilu_cardinal>(ans);
}

// returns success code.
SCM ilu__wait_for_reply(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_ProtocolException ans;
  ilu_cardinal sc;
  ans = ilu_GetReply(scm_call_call(call), &sc, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  return scheme::cons(scheme_value<int>((int)ans),
		      scheme_value<ilu_cardinal>(sc));
}

static ilu_Server DefaultServer = NULL;

SCM ilu__set_default_server(SCM _s)
{
  scheme_ilu_ptr<ilu_Server> s = _s;
  DefaultServer = s;
  return scheme::True;
}

SCM ilu__get_default_server()
{
  return scheme_ilu_ptr<ilu_Server>(DefaultServer);
}

static char *DefaultProtocol = "sunrpc";
static ilu_string DefaultTransport[] = { "sunrpcrm", "tcp_0_0", 0 };

ilu_Port ilu_create_port(ilu_Server server,
			 ilu_string protocolType,
			 ilu_TransportInfo transportType)
{
  char *p;
  ilu_TransportInfo t;
  if ((p = protocolType) == NULL)
    p = DefaultProtocol;
  if ((t = transportType) == NULL || !transportType[0])
    t = DefaultTransport;
#ifdef __linux__
  ilu_Error err;
  return ilu_CreatePort(server, p, t, 0, &err);
#else
  return ilu_CreatePort(server, p, t);
#endif
}

// transport info is a null terminated array of char*
SCM ilu__create_port(SCM _server, SCM _protocolType, SCM _transportType)
{
  scheme_ilu_ptr<ilu_Server> server = _server;
  scheme_string protocolType = _protocolType;
  scheme_vec0<char*> transportType = _transportType;
  ilu_Port tmp = ilu_create_port(server, protocolType, transportType);

  // not too sure about this
  for(unsigned int i = 0; i < transportType.length(); i++)
    delete[] transportType[i];

  return scheme_ilu_ptr<ilu_Port>(tmp);
}


SCM ilu__set_server_default_port(SCM _s, SCM _p)
{
  scheme_ilu_ptr<ilu_Server> s = _s;
  scheme_ilu_ptr<ilu_Port> p = _p;
  ilu_SetServerDefaultPort(s, p);
  return scheme::True;
}

SCM ilu__parse_sbh(SCM _sbh)
{
  scheme_string sbh = _sbh;
  char* plainInstH = 0;
  char* plainServerId = 0;
  char* plainMstid = 0;
  char* encodedContactInfo= 0;
  ilu_cardinal clen = 0;
  ilu_boolean ans;
  ilu_Error lerr;

  ans = ilu_ParseSBH(sbh, &plainInstH, &plainServerId, &plainMstid,
		     &encodedContactInfo, &clen, &lerr);
  ilu_prefer_success(&lerr);

  SCM tmp = scheme::list((SCM)scheme_tmp_string(plainInstH),
			 (SCM)scheme_tmp_string(plainServerId),
			 (SCM)scheme_tmp_string(plainMstid),
			 (SCM)scheme_tmp_string(encodedContactInfo),
			 (SCM)scheme_value<ilu_cardinal>(clen),
			 scheme::eol);

  ilu_free(plainInstH);
  ilu_free(plainServerId);
  ilu_free(plainMstid);
  ilu_free(encodedContactInfo);

  return tmp;
}


ilu_RcvReqStat ilu_receive_request(ilu_Connection conn, ilu_SCMCall call,
				   ilu_boolean *initted, ilu_Class *pclass,
				   ilu_Method *method,
				   ilu_cardinal *serial_number)
{
  ilu_RcvReqStat  ans;
  ilu_cardinal    sn = 0;
  ilu_boolean	initted2;
  ans = ilu_ReceiveRequest(scm_call_call(call), &initted2, conn, pclass, method, &sn, scm_call_err(call));
  ilu_prefer_success(scm_call_err(call));
  *initted = initted2;
  *serial_number = sn;
  return ans;
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXX
// is this ever called!!!!!!!?????????
// OUTPUT params = call, inited
// call object must be created by receive request
SCM ilu__receive_request(SCM _conn)
{
  scheme_ilu_ptr<ilu_Connection> conn = _conn;
  ilu_SCMCall call = new ilu_SCMCall_s;
  ilu_boolean initted;
  ilu_Class pclass;
  ilu_Method method;
  ilu_cardinal serial = 0;
  ilu_RcvReqStat ans;
  ans = ilu_receive_request(conn, call, &initted, &pclass, &method, &serial);
  ilu_prefer_success(scm_call_err(call));
  return scheme_value<int>((int)ans);
}

ilu_Connection ilu_handle_new_connection(ilu_Port port)
{
  ilu_Connection c;
  ilu_boolean closed;

  c = ilu_HandleNewConnection(port, &closed);
  if (closed)
    return NULL;
  else
    return c;
}

SCM ilu__handle_new_connection(SCM _port)
{
  scheme_ilu_ptr<ilu_Port> port = _port;
  ilu_Connection c;
  ilu_boolean closed;

  c = ilu_HandleNewConnection(port, &closed);
  if (closed)
    return scheme::False;
  else
    return scheme_ilu_ptr<ilu_Connection>(c);
}

SCM ilu__close_connection(SCM _conn)
{
  scheme_ilu_ptr<ilu_Connection> conn = _conn;
  ilu_CloseConnection(conn);
  return scheme::True;
}

SCM ilu__wait_for_port_connection_request(SCM _p)
{
  scheme_ilu_ptr<ilu_Port> p = _p;
  ilu_boolean b = ilu_WaitForPortConnectionRequest(p) ? ilu_FALSE : ilu_TRUE;
  return scheme_ilu_boolean(b);
}

SCM ilu__find_class_from_type_name(SCM type_name)
{
  scheme_string name = type_name;
  ilu_Class c = ilu_FindClassFromName(name);
  return scheme_ilu_ptr<ilu_Class>(c);
}

SCM ilu__find_class_from_id(SCM _id)
{
  scheme_string id = _id;
  ilu_Class c = ilu_FindClassFromID(id);
  return scheme_ilu_ptr<ilu_Class>(c);
}


// ?????????????
/* Inside(s, c) */
SCM ilu__create_true_kernel_object(SCM _ih, SCM _server, SCM _c, SCM _lspo)
{
  scheme_tmp_string ih = _ih;
  scheme_ilu_ptr<ilu_Server> server = _server;
  scheme_ilu_ptr<ilu_Class> c = _c;
  SCM* lo = alloc_scm_obj();
  *lo = _lspo;
  ilu_Object o = ilu_FindOrCreateTrueObject(ih, server, c, (void*)lo);
  return scheme_ilu_ptr<ilu_Object>(o);
}

// ????????????
SCM ilu_get_language_specific_object(ilu_Object obj)
{
  SCM* so = (SCM*)ilu_GetLanguageSpecificObject(obj, ilu_scheme_lang_idx());
  return so ? *so : scheme::False;
}

/* Inside(obj->server, obj->class) */
SCM ilu__get_language_specific_object(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  return ilu_get_language_specific_object(obj);
}

// ?????????????
/* Inside(obj->server, obj->class) */
SCM ilu__set_language_specific_object(SCM _obj, SCM _lspo)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;

  SCM* lo = (SCM*)ilu_GetLanguageSpecificObject(obj, ilu_scheme_lang_idx());
  if(!lo)
    lo = alloc_scm_obj();
  *lo = _lspo;

  ilu_RegisterLanguageSpecificObject(obj, (void*)lo, ilu_scheme_lang_idx());
  return scheme::True;
}

SCM ilu__publish_object(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  ilu_string proof = ilu_PublishObject(obj);
  SCM tmp = scheme_tmp_string(proof);
  ilu_free(proof);
  return tmp;
}


SCM ilu__withdraw_object(SCM _obj, SCM _proof)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  scheme_string proof = _proof;
  ilu_boolean b = ilu_WithdrawObject(obj, proof);
  return scheme_ilu_boolean(b);
}

SCM ilu__server_of_object(SCM _obj)
{
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  ilu_Server server = ilu_ServerOfObject(obj);
  return scheme_ilu_ptr<ilu_Server>(server);
}

SCM ilu_class__name(SCM _c)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  ilu_string name = (ilu_Class)c ? c->cl_name : 0;
  return scheme_tmp_string(name);
}

SCM ilu_class__unique_id(SCM _c)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  ilu_string uid = (ilu_Class)c ? c->cl_unique_id : 0;
  return scheme_tmp_string(uid);
}

////////////////////////////////////////////////////////////////////////

SCM ilu__run_main_loop(SCM _stop)
{
  scheme_value<int> stop = _stop;
  ilu_RunMainLoop(&stop);
  return scheme_value<int>(stop);
}

SCM ilu__exit_main_loop(SCM _stop)
{
  scheme_value<int> stop = _stop;
  ilu_ExitMainLoop(&stop);
  return scheme_value<int>(stop);
}

static HashTable proc_table = 0;
static HashTable stub_proc_table = 0;
static HashTable ot_table = 0;

void init_hash_tables()
{
  proc_table = _ilu_hash_MakeNewTable(100, _ilu_hash_HashPointer,
				      _ilu_hash_PointerCompare);
  stub_proc_table = _ilu_hash_MakeNewTable(100, _ilu_hash_HashPointer,
					   _ilu_hash_PointerCompare);
  ot_table = _ilu_hash_MakeNewTable(100, _ilu_hash_HashPointer,
				    _ilu_hash_PointerCompare);
}

static void handler_caller(int fd, void* rock)
{
  SCM* pp = (SCM*)_ilu_hash_FindInTable(proc_table, (ilu_refany)fd);
  scheme::call(*pp);
}

SCM ilu__register_input_handler(SCM _fd, SCM _handlerProc)
{
  FILE* f = scheme::stream(_fd);
  int fd = fileno(f);
  SCM* pp = scheme::make_array(1);
  *pp = _handlerProc;
  _ilu_hash_AddToTable(proc_table, (void*)fd, pp);
  return scheme_ilu_boolean(ilu_RegisterInputSource(fd, handler_caller, pp));
}

SCM ilu__register_output_handler(SCM _fd, SCM _handlerProc)
{
  FILE* f = scheme::stream(_fd);
  int fd = fileno(f);
  SCM* pp = scheme::make_array(1);
  *pp = _handlerProc;
  _ilu_hash_AddToTable(proc_table, (void*)fd, pp);
  return scheme_ilu_boolean(ilu_RegisterOutputSource(fd, handler_caller, pp));
}

SCM ilu__unregister_input_handler(SCM _fd)
{
  FILE* f = scheme::stream(_fd);
  int fd = fileno(f);
  SCM* pp = (SCM*)_ilu_hash_FindInTable(proc_table, (void*)fd);
  scheme::free_array(pp);
  return scheme_ilu_boolean(ilu_UnregisterInputSource(fd));
}

SCM ilu__unregister_output_handler(SCM _fd)
{
  FILE* f = scheme::stream(_fd);
  int fd = fileno(f);
  SCM* pp = (SCM*)_ilu_hash_FindInTable(proc_table, (void*)fd);
  scheme::free_array(pp);
  return scheme_ilu_boolean(ilu_UnregisterOutputSource(fd));
}

static void alarm_caller(void* rock)
{
  SCM* proc = (SCM*)rock;
  scheme::call(*proc);
  scheme::free_array(proc);
}

SCM ilu__create_alarm()
{
  ilu_refany alarm = ilu_CreateAlarm();
  return scheme::from_ptr(alarm);
}

SCM ilu__set_alarm(SCM _alarm, SCM _time, SCM _proc)
{
  ilu_refany alarm = scheme::to_ptr(_alarm);
  scheme_value<double> d = _time;
  ilu_FineTime t = ilu_FineTime_FromDouble(d);
  SCM* pp = scheme::make_array(1);
  *pp = _proc;
  ilu_SetAlarm(alarm, t, alarm_caller, (void*)pp);
  return scheme::True;
}

SCM ilu__clear_alarm(SCM _alarm)
{
  ilu_refany alarm = scheme::to_ptr(_alarm);
  ilu_UnsetAlarm(alarm);
  return scheme::True;
}

SCM ilu__time_now()
{
  ilu_FineTime now = ilu_FineTime_Now();
  double d =  now.ft_s + now.ft_t / (1.0 * ilu_FineTimeRate);
  return scheme_value<double>(d);
}


///////////////////////////////////////////////////////////////////////////

static SCM* theML = NULL;

static void CallRun(int *stop)
{
  SCM _stop = scheme::call("run", *theML, scheme_value<int>(*stop));
  *stop = scheme_value<int>(*stop);
  return;
}

static void CallExit(int *stop)
{
  SCM _stop = scheme::call("exit", *theML, scheme_value<int>(*stop));
  *stop = scheme_value<int>(*stop);
  return;
}

static ilu_boolean CallUnregisterInput(int fd)
{
  SCM b = scheme::call("unregister-input-handler", *theML,
		       scheme_value<int>(fd));
  return scheme_ilu_boolean(b);
}

static ilu_boolean CallUnregisterOutput(int fd)
{
  SCM b = scheme::call("unregister-output-handler", *theML,
		       scheme_value<int>(fd));
  return scheme_ilu_boolean(b);
}

static ilu_refany CallCreateAlarm (void)
{
  return scheme::to_ptr(scheme::call("create-alarm", *theML));
}

static void CallUnsetAlarm(ilu_refany alarm)
{
  scheme::call("clear-alarm", *theML, scheme::from_ptr(alarm));
}

static SCM c_input_handler(SCM _proc, SCM _fd, SCM _rock)
{
  void (*proc)(int, void*) = (void (*)(int, void*))scheme::to_ptr(_proc);
  int fd = scheme_value<int>(_fd);
  void* rock = scheme::to_ptr(_rock);
  proc(fd, rock);
  return scheme::True;
}


static ilu_boolean CallRegisterInput(int fd,
				     void (*proc)(int fd, void * rock),
				     void* rock)
{
  SCM _proc = scheme::lambda((SCM(*)())c_input_handler, 3, 0, 0, "");
  _proc = scheme::curry(_proc, scheme::from_ptr((void*)proc));
  _proc = scheme::curry(_proc, scheme_value<int>(fd));
  _proc = scheme::curry(_proc, scheme::from_ptr(rock));
  SCM b = scheme::call("register-input-handler", *theML, _proc);
  return scheme_ilu_boolean(b);
}

static ilu_boolean CallRegisterOutput(int fd,
				      void (*proc)(int fd, void * rock),
				      void * rock)
{
  SCM _proc = scheme::lambda((SCM(*)())c_input_handler, 3, 0, 0, "");
  _proc = scheme::curry(_proc, scheme::from_ptr((void*)proc));
  _proc = scheme::curry(_proc, scheme_value<int>(fd));
  _proc = scheme::curry(_proc, scheme::from_ptr(rock));
  SCM b = scheme::call("register-output-handler", *theML, _proc);
  return scheme_ilu_boolean(b);
}

static SCM c_alarm_handler(SCM _proc, SCM _rock)
{
  void (*proc)(void*) = (void (*)(void*))scheme::to_ptr(_proc);
  void* rock = scheme::to_ptr(_rock);
  proc(rock);
  return scheme::True;
}

static void CallSetAlarm(ilu_refany alarm, ilu_FineTime t,
			 void (*proc)(void * rock), void * rock)
{
  SCM _proc = scheme::lambda((SCM(*)())c_alarm_handler, 2, 0, 0, "");
  _proc = scheme::curry(_proc, scheme::from_ptr((void*)proc));
  _proc = scheme::curry(_proc, scheme::from_ptr(rock));
  SCM _alarm = scheme::from_ptr(alarm);
  double d = t.ft_s + t.ft_t / (1.0 * ilu_FineTimeRate);
  SCM _time = scheme_value<double>(d);
  scheme::call("set-alarm", *theML, _alarm, _time, _proc);
}

static ilu_MainLoop kml = {
  CallRun, CallExit,
  CallRegisterInput, CallUnregisterInput,
  CallRegisterOutput, CallUnregisterOutput,
  CallCreateAlarm, CallSetAlarm, CallUnsetAlarm
};

SCM ilu__set_main_loop(SCM ml)
{
  if(!theML)
    theML = scheme::make_array(1);
  *theML = ml;
  ilu_SetMainLoop(&kml);
  return scheme::True;
}

SCM ilu__err_nok(SCM _err)
{
  scheme_ptr<ilu_Error> err = _err;
  ilu_boolean b = ILU_ERRNOK(*(ilu_Error*)err) ? ilu_TRUE : ilu_FALSE;
  return scheme_ilu_boolean(b);
}

SCM ilu__get_gc_callback_class()
{
  ilu_Class c = ilu_GetGcCallbackClass();
  return scheme_ilu_ptr<ilu_Class>(c);
}

static void call_stub_proc(ilu_Call call)
{
  ilu_Method m = call->ca_method;
  SCM _call = scheme_ilu_ptr<ilu_Call>(call);
  SCM* pp = (SCM*)_ilu_hash_FindInTable(stub_proc_table, (ilu_refany)m);
  scheme::call(*pp, _call);
}

SCM ilu__set_method_stub_proc(SCM _m, SCM _proc)
{
  scheme_ilu_ptr<ilu_Method> m = _m;
  SCM* pp = scheme::make_array(1);
  *pp = _proc;
  _ilu_hash_AddToTable(stub_proc_table, (ilu_Method)m, pp);
  ilu_SetMethodStubProc(m, call_stub_proc);
  return scheme::True;
}

ilu_Object call_object_of_ih_proc(ilu_ObjectTable self, ilu_string ih)
{
  SCM _self = scheme_ilu_ptr<ilu_ObjectTable>(self);
  SCM _ih = scheme_tmp_string(ih);
  SCM* pp = (SCM*)_ilu_hash_FindInTable(ot_table, (ilu_refany)self);
  return scheme_ilu_ptr<ilu_Object>(scheme::call(pp[0], _self, _ih));
}

void call_free_self_fn_proc(ilu_ObjectTable self)
{
  SCM _self = scheme_ilu_ptr<ilu_ObjectTable>(self);
  SCM* pp = (SCM*)_ilu_hash_FindInTable(ot_table, (ilu_refany)self);
  scheme::call(pp[1], _self);
}

SCM ilu__create_object_table(SCM object_of_ih_fn, SCM free_self_fn)
{
  ilu_ObjectTable ot = new ilu_ObjectTable_s;
  SCM* pp = scheme::make_array(2);
  pp[0] = object_of_ih_fn;
  pp[1] = free_self_fn;
  _ilu_hash_AddToTable(ot_table, (ilu_refany)ot, pp);
  return scheme_ilu_ptr<ilu_ObjectTable>(ot);
}


// call stuff

SCM ilu_call__create()
{
  return scheme_ilu_ptr<ilu_SCMCall>(new ilu_SCMCall_s);
}

SCM ilu_call__destroy(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  delete (ilu_SCMCall)call;
  return scheme::True;
}

SCM ilu_call__caller(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Passport p = scm_call_call(call)->ca_caller;
  return scheme_ilu_ptr<ilu_Passport>(p);
}

SCM ilu_call__protocol_exception(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_ProtocolException pe = scm_call_call(call)->ca_pe;
  return scheme_value<int>((int)pe);
}

SCM ilu_call__error(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Error* e = scm_call_err(call);
  return scheme_ptr<ilu_Error>(e);
}

SCM ilu_call__method_of_call(SCM _call)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  ilu_Method m = scm_call_call(call)->ca_method;
  return scheme_ilu_ptr<ilu_Method>(m);
}

SCM ilu_class__method(SCM _c, SCM _num)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  scheme_value<ilu_cardinal> num = _num;
  ilu_Method m = c->cl_methods + (ilu_cardinal)num;
  return scheme_ilu_ptr<ilu_Method>(m);
}

