//  -*- Mode: C++; -*-
//
// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: bnell $
// $Date: 1996/11/13 16:57:36 $
// $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/ilu-type.cc,v $
// $Revision: 1.2 $
//

#include <assert.h>
#include <string.h>
#include "ilu-type.h"

//////////////////////////////////////////////////////////////////////

SCM scheme_ilu_longinteger::to_scheme()
{
  ilu_longinteger c = value();
  return scheme::cons(scheme::from_long(ILU_LONGINT_HIGH_WORD(&c)),
		      scheme::from_ulong(ILU_LONGINT_LOW_WORD(&c)));
}

scheme_ilu_longinteger::scheme_ilu_longinteger(SCM s)
{
  ilu_longinteger i;
  ILU_LONGINT_HIGH_WORD(&i) = scheme::to_long(scheme::car(s));
  ILU_LONGINT_LOW_WORD(&i) = scheme::to_ulong(scheme::cdr(s));
  set_value(i);
}

scheme_ilu_longinteger::scheme_ilu_longinteger(ilu_longinteger v)
: schemet<ilu_longinteger>(v)
{
}

//////////////////////////////////////////////////////////////////////

SCM scheme_ilu_longcardinal::to_scheme()
{
  ilu_longcardinal c = value();
  return scheme::cons(scheme::from_ulong(ILU_LONGCARD_HIGH_WORD(&c)),
		      scheme::from_ulong(ILU_LONGCARD_LOW_WORD(&c)));
}

scheme_ilu_longcardinal::scheme_ilu_longcardinal(SCM s)
{
  ilu_longcardinal c;
  ILU_LONGCARD_HIGH_WORD(&c) = scheme::to_ulong(scheme::car(s));
  ILU_LONGCARD_LOW_WORD(&c) = scheme::to_ulong(scheme::cdr(s));
  set_value(c);
}

scheme_ilu_longcardinal::scheme_ilu_longcardinal(ilu_longcardinal v)
: schemet<ilu_longcardinal>(v)
{
}

//////////////////////////////////////////////////////////////////////

// punt on longreals if not supported by the compiler!

SCM scheme_ilu_longreal::to_scheme()
{
  ilu_longreal r = value();
  SCM s;
#ifdef LONG_REAL_TYPE
  s = scheme::from_double(r);
#else
  s = scheme::from_double(0.0);
#endif
  return s;
}

scheme_ilu_longreal::scheme_ilu_longreal(SCM s)
{
  ilu_longreal r;
#ifdef LONG_REAL_TYPE
  r = scheme::to_double(s);
#else
  //  r = scheme::to_double(s);
#endif
  set_value(r);
}

scheme_ilu_longreal::scheme_ilu_longreal(ilu_longreal v)
: schemet<ilu_longreal>(v)
{
}

//////////////////////////////////////////////////////////////////////

scheme_ilu_wstring scheme_convert(SCM s, scheme_ilu_wstring*)
{
  return scheme_ilu_wstring(s);
}

scheme_ilu_longcardinal scheme_convert(SCM s, scheme_ilu_longcardinal*)
{
  return scheme_ilu_longcardinal(s);
}

scheme_ilu_longinteger scheme_convert(SCM s, scheme_ilu_longinteger*)
{
  return scheme_ilu_longinteger(s);
}

scheme_ilu_longreal scheme_convert(SCM s, scheme_ilu_longreal*)
{
  return scheme_ilu_longreal(s);
}

//////////////////////////////////////////////////////////////////////
