//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: bnell $
// $Date: 1996/11/08 19:08:56 $
// $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/ilu-type.h,v $
// $Revision: 1.2 $
//

#ifndef _ILU_TYPE_H_
#define _ILU_TYPE_H_

// mappings between ilu types and scheme types

#include <stdio.h>
#include "scheme-types.h"

extern "C" {
#include "iluxport.h"
}

#if 1
typedef struct ilu_SCMCall_s {
  ilu_Call_s call;
  ilu_Error  err;
}* ilu_SCMCall;
#define scm_call_call(c) (&(((ilu_SCMCall)(c))->call))
#define scm_call_err(c) (&(((ilu_SCMCall)(c))->err))
#else
typedef _ilu_Call_s ilu_SCMCall_s;
typedef ilu_Call ilu_SCMCall;
#define scm_call_call(c) ((ilu_Call)(c))
#define scm_call_err(c) (&(((ilu_Call)(c))->err))
#endif

template <class T> class scheme_ilu_ptr : public schemet<T> {
 public:
#ifdef __GNUG__
  operator bool() { return (bool)value(); }
#else
  operator int() { return (int)value(); }
#endif
  T operator->() { return value(); }
  SCM to_scheme()
  {
    return value() ? scheme::from_ulong((unsigned int)value()) : scheme::False;
  }
  scheme_ilu_ptr() : schemet<T>((T)0) { }
  scheme_ilu_ptr(T p) : schemet<T>(p) { }
  scheme_ilu_ptr(SCM s)
  : schemet<T>((T)0)
  {
    if(s != scheme::False)
      set_value((T)scheme::to_ulong(s));
  }
};

class scheme_ilu_boolean : public schemet<ilu_boolean> {
 public:
#ifdef __GNUG__
  operator bool() { return value() ? true : false; }
#else
  operator int() { return value() ? 1 : 0; }
#endif
  SCM to_scheme() { return value() ? scheme::True : scheme::False; }
  scheme_ilu_boolean()
    : schemet<ilu_boolean>(ilu_FALSE) { }
#ifdef __GNUG__
  scheme_ilu_boolean(bool v)
    : schemet<ilu_boolean>(v ? ilu_TRUE : ilu_FALSE) { }
#endif
  scheme_ilu_boolean(ilu_boolean v)
    : schemet<ilu_boolean>(v ? ilu_TRUE : ilu_FALSE) { }
  scheme_ilu_boolean(int v)
    : schemet<ilu_boolean>(v ? ilu_TRUE : ilu_FALSE) { }
  scheme_ilu_boolean(SCM s)
    : schemet<ilu_boolean>(s != scheme::False ? ilu_TRUE : ilu_FALSE) { }
};

class scheme_ilu_wstring : public scheme_vec0<ilu_shortcardinal> {
 public:
  scheme_ilu_wstring()
    : scheme_vec0<ilu_shortcardinal>() { }
  scheme_ilu_wstring(SCM s)
    : scheme_vec0<ilu_shortcardinal>(s) { }
  scheme_ilu_wstring(unsigned short* s)
    : scheme_vec0<ilu_shortcardinal>(s) { }
  scheme_ilu_wstring(const scheme_ilu_wstring& s)
    : scheme_vec0<ilu_shortcardinal>(s) { }
  ~scheme_ilu_wstring() { ilu_free(value()); }
};

class scheme_ilu_longcardinal : public schemet<ilu_longcardinal> {
 public:
  SCM to_scheme();
  scheme_ilu_longcardinal(SCM);
  scheme_ilu_longcardinal(ilu_longcardinal);
};

class scheme_ilu_longinteger : public schemet<ilu_longinteger> {
 public:
  SCM to_scheme();
  scheme_ilu_longinteger(SCM);
  scheme_ilu_longinteger(ilu_longinteger);
};

class scheme_ilu_longreal : public schemet<ilu_longreal> {
 public:
  SCM to_scheme();
  scheme_ilu_longreal(SCM);
  scheme_ilu_longreal(ilu_longreal);
};

extern scheme_ilu_wstring scheme_convert(SCM s, scheme_ilu_wstring*);
extern scheme_ilu_longcardinal scheme_convert(SCM s, scheme_ilu_longcardinal*);
extern scheme_ilu_longinteger scheme_convert(SCM s, scheme_ilu_longinteger*);
extern scheme_ilu_longreal scheme_convert(SCM s, scheme_ilu_longreal*);

template <class T>
inline scheme_ilu_ptr<T> scheme_convert(SCM s, scheme_ilu_ptr<T>*)
{
  return scheme_ilu_ptr<T>(s);
}

inline scheme_ilu_boolean scheme_convert(SCM s, scheme_ilu_boolean*)
{
  return scheme_ilu_boolean(s);
}

template <class T>
inline void delete_vec(T* vec, int num)
{
  while(num--)
    delete[] vec[num];
}

#endif
