//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: bnell $
// $Date: 1996/11/13 16:57:39 $
// $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/iluobj-scm.cc,v $
// $Revision: 1.2 $
//

#include "ilu-scm-private.h"
#include "iluobj-scm.h"

// use gscm_alloc_array or something....
struct registry_record {
  ilu_Class c;
  SCM* proc;
  struct registry_record *next;
};

static struct registry_record *RegisteredClasses = NULL;

static struct registry_record * findRegistryRecordByClass (ilu_Class c)
{
  struct registry_record *p;

  for (p = RegisteredClasses;  p != NULL;  p = p->next)
    if (p->c == c)
      return (p);
  return (NULL);
}

SCM ilu_object__register_surrogate_creator(SCM _c, SCM _proc)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  struct registry_record *p;

  if ((p = findRegistryRecordByClass(c)) == NULL)
    {
      p = new registry_record;
      p->c = c;
      p->proc = scheme::make_array(1);
      p->proc[0] = _proc;
      p->next = RegisteredClasses;
      RegisteredClasses = p;
    }
  return scheme::True;
}

/* Inside(obj->server, c) */
SCM ilu_object_create_from_registry(ilu_Class c, ilu_Object obj)
{
  struct registry_record *p;
  if((p = findRegistryRecordByClass(c)) == NULL) {
    return scheme::False;
  } else {
    SCM _obj = scheme_ilu_ptr<ilu_Object>(obj);
    return scheme::call(p->proc[0], _obj);
  }
}

/* Inside(obj->server, c) */
SCM ilu_object__create_from_registry(SCM _c, SCM _obj)
{
  scheme_ilu_ptr<ilu_Class> c = _c;
  scheme_ilu_ptr<ilu_Object> obj = _obj;
  return ilu_object_create_from_registry(c, obj);
}


SCM ilu_object__input_object(SCM _call,
			     SCM _discriminator_p,
			     SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_boolean discriminator_p = _discriminator_p;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_Object obj;
  ilu_Class c;
  SCM o;

#ifdef __GNUG__
  if (discriminator_p && putative_class->cl_singleton) {
#else
  if ((int)discriminator_p && putative_class->cl_singleton) {
#endif
    if ((obj = ilu_GetCallSingleton(scm_call_call(call), scm_call_err(call))) == NULL) {
      ilu_prefer_success(scm_call_err(call));
      return scheme::False;
    }
  } else {
    obj = ilu_input_object_id(call, discriminator_p, putative_class);
    if (obj == NULL)
      return scheme::False;
  }

  /* now Inside(obj->server, putative_class) */
  o = ilu_get_language_specific_object(obj);
  if (o == scheme::False) {
    if ((c = ilu_ClassOfObject(obj)) != NULL)
      o = ilu_object_create_from_registry(c, obj);
  }
  ilu_exit_server(ilu_get_object_server(obj), putative_class);

  return o;
}

SCM ilu_object__output_object(SCM _call, SCM _obj, SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_Object kobj;

  if(_obj == scheme::False) {
    kobj = NULL;
  } else {
    SCM _kobj = scheme::call("get-rpc-object", _obj);
    kobj = scheme_ilu_ptr<ilu_Object>(_kobj);
  }

  /* now kobj != NULL implies Inside(kobj->server, kobj->class) */
  ilu_boolean b = ilu_output_object_id(call, kobj, ilu_FALSE, putative_class);
  return scheme_ilu_boolean(b);
}

SCM ilu_object__size_of_object(SCM _call, SCM obj, SCM _putative_class)
{
  scheme_ilu_ptr<ilu_SCMCall> call = _call;
  scheme_ilu_ptr<ilu_Class> putative_class = _putative_class;
  ilu_Object kobj;

  if(obj == scheme::False) {
    kobj = NULL;
  } else {
    kobj = scheme_ilu_ptr<ilu_Object>(scheme::call("get-rpc-object", obj));
  }

  /* now kobj != NULL implies Inside(kobj->server, kobj->class) */
  ilu_cardinal s = ilu_size_of_object_id(call,kobj,ilu_FALSE,putative_class);
  return scheme_value<ilu_cardinal>(s);
}

SCM ilu_object__lookup(SCM _sid, SCM _ih, SCM _pclass)
{
  scheme_string sid = _sid;
  scheme_string ih = _ih;
  scheme_ilu_ptr<ilu_Class> pclass = _pclass;
  ilu_Object kobj;
  SCM o;

  if ((kobj = ilu_LookupObject (sid, ih, pclass)) == NULL)
    return scheme::False;
  /* now Inside the server of the kobj */
  if ((o = ilu_get_language_specific_object(kobj)) == scheme::False)
    {
      o = ilu_object_create_from_registry(pclass, kobj);
    }
  ilu_exit_server(ilu_get_object_server(kobj), pclass);

  return o;
}


SCM ilu_object__register_as_gc_callback(SCM obj)
{
  ilu_Object kobj;

  kobj = scheme_ilu_ptr<ilu_Object>(scheme::call("get-rpc-object", obj));

  if(kobj != NULL)
    {
      ilu_SetGcClient(kobj);  // register ilu_Object with GC handler in kernel
      SCM server = scheme::call("ilu-object:rpc-server", obj);
      SCM cr = scheme::call("ilu-object:instance-class-record", obj);
      ilu__exit_server(server, cr);
    }

  return scheme::True;
}

//////////////////////////////////////////////////////////////////////
// code to do the GC keepalive acknowedgements
//////////////////////////////////////////////////////////////////////

// this section is implemented in scheme!!!!!!!!!!!!
#if 0
class _ilu_GCCallback : public virtual iluObject {
  public:
    _ilu_GCCallback(class iluServer *server);
    virtual ~_ilu_GCCallback();

  static class _ilu_GCCallback * ILUCreateFromSBH(ilu_CString sbh);
  static class _ilu_GCCallback * ILUQuaT (class iluObject *from);
  static ilu_Class ILUClassRecord;

  virtual void * ILUCastDown (ilu_Class cast_to);

  virtual iluServer * ILUGetServer();

  private:
   class iluServer *ourServer;
};

ilu_Class _ilu_GCCallback::ILUClassRecord = NULL;

class _ilu_GCCallback * _ilu_GCCallback::ILUCreateFromSBH (ilu_CString sbh)
{
  return (_ilu_GCCallback *) ilu::SBHToObject(sbh, _ilu_GCCallback::ILUClassRecord);
}

class _ilu_GCCallback * _ilu_GCCallback::ILUQuaT (class iluObject *from)
{
  return((class _ilu_GCCallback *) (from->ILUCastDown (_ilu_GCCallback::ILUClassRecord)));
}

void * _ilu_GCCallback::ILUCastDown (ilu_Class cast_to)
{
  if (cast_to == NULL)
    return((void *)((class iluObject *) this));
  else if (cast_to == _ilu_GCCallback::ILUClassRecord)
    return ((void *) this);
  else return (NULL);
}

_ilu_GCCallback::_ilu_GCCallback (class iluServer *server)
{
  this->ILUInstanceClassRecord = ilu_GetGcCallbackClass();
  this->ILUSetMostSpecificObject((void *) this);
  this->ourServer = server;
}

_ilu_GCCallback::~_ilu_GCCallback ()
{
}

class iluServer *_ilu_GCCallback::ILUGetServer ()
{
  return (this->ourServer);
}

static class iluObject * Create__ilu_GCCallback(ilu_KernelObject obj)
{
  class _ilu_GCCallback *nobj = new _ilu_GCCallback(NULL);
  nobj->ILUSetRPCObject(obj);
  ilu::SetLanguageSpecificObject(obj, (class iluObject *) nobj);
  return ((class iluObject *) nobj);
}





class _ilu_GCCallbackInit {
  public:
    _ilu_GCCallbackInit();
};

static _ilu_GCCallbackInit do_init;
static _ilu_GCCallback * GCCallbackObject = NULL;

void ilu_GCInit()
{
  class iluServer *server;

  // Initialize GCCallback class

  _ilu_GCCallback::ILUClassRecord = ilu_GetGcCallbackClass();
  iluObject::RegisterSurrogateCreator (_ilu_GCCallback::ILUClassRecord, Create__ilu_GCCallback);

  // create instance of GCCallback on server

  server = new iluServer(NULL, NULL);
  server->AddPort (NULL, NULL, ilu_TRUE);
  GCCallbackObject = new _ilu_GCCallback(server);

  iluObject::_ILU_RegisterAsGCCallback(GCCallbackObject);
}

_ilu_GCCallbackInit::_ilu_GCCallbackInit ()
{
  if(!getenv("ILU_NO_AUTO_INIT_GC"))
    ilu_GCInit();
}
#endif
