//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: bnell $
// $Date: 1996/11/13 16:57:40 $
// $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/ilusrvr-scm.cc,v $
// $Revision: 1.2 $
//

#include <string.h>
#include "ilu-scm-private.h"
#include "ilusrvr-scm.h"

//
// this file has been checked for errors!
//

static ilu_boolean threaded    = ilu_FALSE;
static ilu_boolean threadedSet = ilu_FALSE;

static void (*Fork)(void (*proc)(void *arg), void *arg) = 0;

/*Main invariant holds */
void ilu_server_monitor_outgoing_connection(void *rock)
{
  ilu_Connection  conn = (ilu_Connection) rock;
  ILU_ERRS((IoErrs)) lerr;
  (void) ilu_OutgoingConnectionThreadProc(conn, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
}

/* Main invariant holds */
void ilu_server_pass_new_connections(void *rock)
{
  ilu_Connection  nu;
  ILU_ERRS((internal)) lerr;
  while (1) {
    nu = ilu_OtherNewConnection(&lerr);
    if (nu != NULL)
      (*Fork) (ilu_server_monitor_outgoing_connection, nu);
    ILU_MUST_BE_SUCCESS(lerr);
  }
}

// This method is used when someone outside of this module needs to use
// the Fork routine.
// Since it's declared static, it isn't visible (ilu.cpp needs this)

void ilu_server_monitor_conn(ilu_Connection conn)
{
    (*Fork) (ilu_server_monitor_outgoing_connection, conn);
}

/* Returns TRUE iff connection closed. */
/*Main Invariant holds
  before: L2 disjoint {arg's callmu, iomu}
  after:  L2     >=   {conn's callmu, iomu} if result==ilu_good_request,
  after:  L2 disjoint {conn's callmu, iomu} if result!=ilu_good_request */
/*If !threaded: ReadServiceRequest registered for fd of conn */
static ilu_boolean FinalServiceRequest (ilu_private arg)
{
  ilu_Connection  conn = (ilu_Connection) arg;
  ilu_RcvReqStat  s;
  ilu_SCMCall_s	  call;
  ilu_Class       ptype;
  ilu_Method      method;
  ilu_boolean	  initted;
  ilu_cardinal    serial_number;

  ilu_server_disable_requests((ilu_SCMCall) 0, conn);
  scm_call_call(&call)->ca_reqs_enabled = ilu_FALSE;
  s = ilu_receive_request(conn, &call, &initted,&ptype,&method,&serial_number);
  if (s == ilu_RcvReqStat_request) {
    ilu_refany stubproc = ilu_GetMethodStubProc(method);
    _ilu_Assert(stubproc != NULL, "ilusrvr.cpp:FinalServiceRequest");
    (*(void (*) (ilu_Call)) stubproc) (scm_call_call(&call));
  }
  if (initted) {
    ilu_FinishCall(scm_call_call(&call), scm_call_err(&call));
    ilu_prefer_success(scm_call_err(&call));
    ILU_HANDLED(*scm_call_err(&call));
  }
  if (s == ilu_RcvReqStat_quit)
    return ilu_TRUE;
  if (!scm_call_call(&call)->ca_reqs_enabled) {
    ilu_server_enable_requests(&call, conn);
  }
  return ilu_FALSE;
}

/* This routine handles the dispatching of an incoming request */

static void ReadServiceRequest (void *arg)
{
  (void) FinalServiceRequest(arg);
}

/*L1, L2, Main unconstrained*/
ilu_boolean ilu_server_enable_requests(ilu_SCMCall call, ilu_Connection conn)
{
  if (threaded)
    return ilu_TRUE;

  ilu_Server      s = ilu_ServerOfConnection(conn);
  ilu_string      sid = ilu_IDOfServer(s);
  ilu_Error       lerr;
  ilu_boolean ans;
  ilu_Error	*perr;

  if (call == ILU_NIL)
    perr = &lerr;
  else
    perr = scm_call_err(call);
  ans = ilu_SetConnectionInputHandler(conn, ReadServiceRequest, conn, perr);
  ilu_prefer_success(perr);
  if (!ans)
    fprintf(stderr,
	    "iluServer.cc:  Can't register ReadServiceRequest on conn %p server %s!\n",
	    conn, sid);
  else if (call != ILU_NIL)
    scm_call_call(call)->ca_reqs_enabled = ilu_TRUE;
  return ans;
}

/*L1, L2, Main unconstrained*/
ilu_boolean ilu_server_disable_requests(ilu_SCMCall call, ilu_Connection conn)
{
  if (threaded)
    return ilu_TRUE;

  ilu_Server   s = ilu_ServerOfConnection(conn);
  ilu_string   sid = ilu_IDOfServer(s);
  ilu_Error    lerr;
  ilu_boolean  ans;
  ilu_Error    *perr;

  if (call == ILU_NIL)
    perr = &lerr;
  else
    perr = scm_call_err(call);
  ans = ilu_SetConnectionInputHandler(conn, 0, NULL, perr);
  ilu_prefer_success(perr);
  if (!ans)
    fprintf(stderr, "iluServer.cc:  bug detected when unregistering input source for conn %p server %s!\n", conn, sid);
  else if (call != ILU_NIL)
    scm_call_call(call)->ca_reqs_enabled = ilu_FALSE;
  return ans;
}



/* When a request for a new connection to the server is received,
   we have to handle it. */

/*ARGSUSED*/
static void ReadConnectionRequest (void *arg)
{
  ilu_Port        p = (ilu_Port) arg;
  ilu_Error       lerr;
  ilu_Connection  conn;

  if ((conn = ilu_handle_new_connection(p)) == NULL) {
    fprintf(stderr, "iluServer::ReadConnectionRequest:  Error:  Unable to accept incoming connection.\n");
    return;
  }
  ilu_SetConnectionInputHandler(conn, ReadServiceRequest, conn, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
}


/* Main invariant holds */
static void RunConnection (void *arg)
{
  ilu_Connection conn   = (ilu_Connection) arg;

  for (;;) {
    if (!ilu_BlockingWaitForInputOnConnection(conn, (ilu_FineTime *) 0))
      break;
    if (FinalServiceRequest(conn))
      break;
  }
  ilu_CloseConnection(conn);
}

/* Main invariant holds */
static void ReadConnectionRequests (void *arg)
{
  ilu_Port        p = (ilu_Port) arg;
  ilu_Connection  conn;
  while (ilu_WaitForPortConnectionRequest(p)) {
    conn = ilu_handle_new_connection(p);
    if (conn != ILU_NIL)
      Fork(RunConnection, conn);
  }
}

static ilu_string inmemTinfo[] = { "inmem_", 0 };

ilu_boolean ilu_server_set_fork(void (*fork)(void (*proc)(void *arg), void *arg))
{
  ILU_ERRS((internal)) lerr;

  if (threadedSet)
    return ilu_FALSE;
  threadedSet = ilu_TRUE;
  Fork = fork;
  threaded = ilu_TRUE;
  (void) ilu__scheme_lang_idx();
  (void) ilu__get_default_server();
  (*fork)(ilu_server_pass_new_connections, NULL);
  (void) ilu_NewConnectionGetterForked(&lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return (ilu_TRUE);
}

ilu_boolean ilu_server_add_port(ilu_Server ks, ilu_string protocolType,
				ilu_TransportInfo transportType,
				ilu_boolean bdefault)
{
  ilu_Error       lerr;
  ilu_Port        nu;
  nu = ilu_create_port(ks, protocolType, transportType);

  if (nu == NULL) {
    fprintf(stderr, "iluServer::Start:  Error:  Couldn't open port.\n");
    return ilu_FALSE;
  }

  if (bdefault)
    ilu_SetServerDefaultPort(ks, nu);

  if (threaded && transportType != inmemTinfo) {
    Fork(ReadConnectionRequests, nu);
  } else {
    ilu_SetConnectionRequestHandler(nu, ReadConnectionRequest, nu, &lerr);
    ILU_MUST_BE_SUCCESS(lerr);
  }
  return ilu_TRUE;
}

SCM ilu_server__add_port(SCM _ks, SCM _protocolType,
			 SCM _transportType, SCM _bdefault)
{
  scheme_ilu_ptr<ilu_Server> ks = _ks;
  scheme_string protocolType = _protocolType;
  scheme_vec0<char*> transportType = _transportType;
  scheme_ilu_boolean bdefault = _bdefault;
  ilu_boolean b;

  b = ilu_server_add_port(ks, protocolType, transportType, bdefault);

  // must delete these
  for(unsigned int i = 0; i < transportType.length(); i++)
    delete[] transportType[i];

  return scheme_ilu_boolean(b);
}

SCM ilu_server__create(SCM _serviceID, SCM _objtab)
{
  scheme_tmp_string serviceID = _serviceID;  // don't delete
  scheme_ilu_ptr<ilu_ObjectTable> kot  = _objtab;
  char* sid = serviceID;

  threadedSet = ilu_TRUE;
  if ((sid == NULL) || (strlen(sid) == 0))
    sid = ilu_InventID();
  ilu_Server ks = ilu_CreateTrueServer(sid, kot, ilu_scheme_lang_idx());
  (void) ilu_server_add_port(ks, "sunrpc", inmemTinfo, ilu_FALSE);
  /* What to do if this fails?? */
  return scheme_ilu_ptr<ilu_Server>(ks);
}

SCM ilu_server__id(SCM _ks)
{
  scheme_ilu_ptr<ilu_Server> ks = _ks;
  return scheme_tmp_string(ilu_IDOfServer(ks));
}
