/* genenm.c */
/* Chris Jacobi, January 30, 1997 11:22 am PST */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* $Id: genenm.c,v 1.5 1997/01/30 20:40:03 jacobi Exp $ */


#include <stdio.h>
#include <string.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "io.h"
#include "name.h"
#include "util.h" 
#include "genstub.h"
#include "genenm.h"


PUBLIC boolean enm_is_a(Type t) {
    t = myUrType(t);
    return (type_kind(t)==enumeration_Type);
}


PUBLIC Type enm_assert(Type t) {
    t = myUrType(t);
    if (! enm_is_a(t)) fatal("should have been an enumeration");
    return t;
}


PRIVATE char * ennShortName(Type t) {
    t = enm_assert(t);
    return easyShortTypeNameCleaned(t);
}


PRIVATE char * ennName(Type t) {
    t = enm_assert(t);
    return packageDotString(t->interface, ennShortName(t));
}


PUBLIC char * enumerationJavaClass(Type t) {
    return ennName(t);
}


PUBLIC char * enm_typeDeclarator(Type t) {
    enm_assert(t);
    return "int";
}


PUBLIC char * enm_ioSzPiece(Type t, const char *arg) {
    t = enm_assert(t);
    return cat3("_call.szInt32(", arg, ")");
}


PUBLIC char * enm_ioOutPiece(Type t, const char *arg) {
    t = enm_assert(t);
    return cat5("_call.outInt32(", ennName(t), ".narrow(", arg, "))");
}


PUBLIC char * enm_ioInPiece(Type t) {
    t = enm_assert(t);
    return(cat3(ennName(t), ".narrow(", "_call.inInt32())"));
}


PRIVATE int
enumMin(list enum_field)
{
    int xval = ((EnumField)list_ref(enum_field, 0))->id;
    list l = list_cdr(enum_field);
    int i;
    for (i = 0; i < list_size(l); i++) 
        if (xval > ((EnumField)list_ref(l, i))->id)
            xval = ((EnumField)list_ref(l, i))->id;

    return xval;
}


PRIVATE int
enumMax(list enum_field)
{
    int xval = ((EnumField)list_ref(enum_field, 0))->id;
    list l = list_cdr(enum_field);
    int i;
    for (i = 0; i < list_size(l); i++) 
        if (xval < ((EnumField)list_ref(l, i))->id)
            xval = ((EnumField)list_ref(l, i))->id;

    return xval;
}


PRIVATE void
define1EnumLiteral(EnumField e, void *refany)
{
    printf("  public static final int %s = %d;\n", 
	   enumFieldName(e),  
	   e->id
	   );
}


PRIVATE void
defineEnumLiterals(Type t)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.enumeration;
    list_enumerate(fields, (EnumProc) define1EnumLiteral, NULL);
}


PUBLIC void
enm_defineMain(Type t)
{
    char* baseTypeName = ennShortName(t);
    
    NewJavaFile(t->interface, baseTypeName);
    printf("public class %s {\n\n", baseTypeName);
    printf("    public static final int narrow(int i)\n");
    printf("            throws xerox.ilu.IluEnumerationRangeException {\n");
    printf("        if (i >= %d && i <= %d) return i;\n",\
	   enumMin(type_description(t)->structuredDes.enumeration),\
	   enumMax(type_description(t)->structuredDes.enumeration));
    printf("        else throw");
    printf(" new xerox.ilu.IluEnumerationRangeException();\n");
    printf("    }\n\n");
    defineEnumLiterals(t);
    printf("} //%s\n", baseTypeName);

    /* no holder class */
}

