/* genenum.c */
/* Chris Jacobi, January 30, 1997 11:22 am PST */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* $Id: genenum.c,v 1.5 1997/01/30 20:39:29 jacobi Exp $ */


#include <stdio.h>
#include <string.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "io.h"
#include "name.h"
#include "util.h" 
#include "genstub.h"
#include "genenum.h"


PUBLIC boolean enm_is_a(Type t) {
    t = myUrType(t);
    return (type_kind(t)==enumeration_Type);
}


PUBLIC Type enm_assert(Type t) {
    t = myUrType(t);
    if (! enm_is_a(t)) fatal("should have been an enumeration");
    return t;
}


PRIVATE char * enmShortName(Type t) {
    t = enm_assert(t);
    return easyShortTypeNameCleaned(t);
}


PRIVATE char * enmName(Type t) {
    t = enm_assert(t);
    return packageDotString(t->interface, enmShortName(t));
}


PUBLIC char * enumerationJavaClass(Type t) {
    return enmName(t);
}


PUBLIC char * enm_typeDeclarator(Type t) {
    enm_assert(t);
    return enmName(t);
}


PUBLIC char * enumerationInt(Type t, ConstantValue v) {
    TypeKind vtk = v->type; /* bad name for a TypeKind */
    TypeKind ctk = type_ur_kind(t); 
    char* baseTypeName = enmName(t);
    enm_assert(t);
    switch (ctk) {
        case enumeration_Type:
            if (vtk != shortcharacter_Type) fatal("bad enumeration const");
            return cat3(baseTypeName, "._", v->val.s);
        default: fatal("bad enumeration const");
    }
}


PUBLIC char * enm_ioSzPiece(Type t, const char *arg) {
    t = enm_assert(t);
    return cat3("_call.szInt32(", arg, ".__toInt())");
}


PUBLIC char * enm_ioOutPiece(Type t, const char *arg) {
    t = enm_assert(t);
    return cat3("_call.outInt32(", arg, ".__toInt())");
}


PUBLIC char * enm_ioInPiece(Type t) {
    t = enm_assert(t);
    return cat2(enmName(t), ".__fromInt(_call.inInt32())");
}


PRIVATE int
enumMax(list enum_field)
{
    int xval = ((EnumField)list_ref(enum_field, 0))->id;
    list l = list_cdr(enum_field);
    cardinal i;
    for (i = 0; i < list_size(l); i++) 
        if (xval < ((EnumField)list_ref(l, i))->id)
            xval = ((EnumField)list_ref(l, i))->id;

    return xval;
}


PRIVATE void
define1EnumLiteral1(EnumField e, void *refany)
{
    printf("  public static final int _%s = %d;\n", 
	   enumFieldName(e),  
	   e->id
	   );
}

PRIVATE void
define1EnumLiteral2(EnumField e, char * baseTypeName)
{
    printf("  public static final %s %s = __fromInt(_%s);\n", 
	   baseTypeName,
	   enumFieldName(e),  
	   enumFieldName(e)
	   );
}


PRIVATE void
defineEnumLiterals1(Type t)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.enumeration;
    list_enumerate(fields, (EnumProc) define1EnumLiteral1, NULL);
}


PRIVATE void
defineEnumLiterals2(Type t)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.enumeration;
    list_enumerate(fields, (EnumProc) define1EnumLiteral2, enmShortName(t));
}


PUBLIC void
enm_defineMain(Type t)
{
    char* baseTypeName = enmShortName(t);
    int eMax = enumMax(type_description(t)->structuredDes.enumeration);
    
    NewJavaFile(t->interface, baseTypeName);
    printf("public final class %s {\n\n", baseTypeName);
    
    printf("    private int __val;\n");
    printf("    private static %s[] __a = new %s[%d];\n\n", 
        baseTypeName, baseTypeName, eMax+1);
    printf("    private %s(){}\n", baseTypeName);
    printf("    private %s(int __i){__val = __i;}\n\n", baseTypeName);

    printf("    public static final %s __fromInt(int __i)\n", baseTypeName);
    printf("            throws xerox.ilu.IluEnumerationRangeException {\n");
    printf("        if (__i < 0 || __i > %d) throw new xerox.ilu.IluEnumerationRangeException();\n", eMax);
    printf("        return __a[__i];\n");
    printf("    } //__fromInt\n\n");
    
    printf("    public final int __toInt() {\n");
    printf("        return __val;\n");
    printf("    } //__toInt\n\n");
    
    printf("    static {\n");
    printf("        for (int __i = 0; __i<=%d; __i++) {\n", eMax);
    printf("            __a[__i] = new %s(__i);\n", baseTypeName);
    printf("        }\n");
    printf("    }\n\n");
    
    defineEnumLiterals1(t);
    printf("\n");
    defineEnumLiterals2(t);
    printf("\n");
    printf("} //%s\n", baseTypeName);

    /* no holder class */
}

