/* genex.c */
/* Chris Jacobi, January 30, 1997 1:52 pm PST */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* $Id: genex.c,v 1.7 1997/01/30 22:29:25 jacobi Exp $ */


#include <stdio.h>
#include <string.h>

#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "name.h"
#include "util.h"
#include "context.h"
#include "genex.h"
#include "io.h"


PRIVATE char*
methodNameOfWriteProc(Exception e)
{
    char* shortname = exceptionName(e);
    return cat3("_", shortname, "_marshall"); 
}


PUBLIC char*
fullNameOfWriteProc(Exception e)
{
    if (e->type) {
        return cat5(
            packageName(e->interface),	/*Package*/
            ".",
            exceptionName(e),         	/*Class*/
            ".",
            methodNameOfWriteProc(e) 	/*Method*/
            ); 
    } else {
        return "xerox.ilu.IluUserException.writeException";
    }
            
}


PUBLIC void
map_exception(Exception e, void* x)
{
    char* shortname;
    if (e->import != 0) {
        return;
    }
    shortname = exceptionName(e);
    
    NewJavaFile(e->interface, shortname);

    if (e->doc_string != NULL) {
        printDocString("", e->doc_string);
    }
    printf("public class %s extends xerox.ilu.IluUserException {\n\n", 
          shortname);
    printf("    private static xerox.ilu.IluExceptionRep _%s_rep;\n", 
           shortname);
           
    printf("    public %s() {\n", shortname);
    printf("        super();\n");
    printf("    }\n\n");
        
    
    if (e->type) {
        char* declarator = typeDeclarator(e->type);
        printf("    public %s value;\n\n", declarator);
        printf("    public %s(%s _val) {\n", 
            shortname, 
            declarator
            );
        printf("        super();\n");
        printf("        this.value = _val;\n");
        printf("    }\n\n");
        
        
        printf("    //NOT really public; Must be accessible by ilu...\n");
        printf("    public void readException(xerox.ilu.IluCall _call)\n");
        printf("            throws xerox.ilu.IluSystemException {\n");
        printf("        this.value = %s;\n", ioInPiece(e->type));
        printf("    } //readException\n\n");
        
        
        /* This is static because the dynamic exception might be 
         * a subclass which is not supported in the ISL definition
         */
        printf("    //NOT really public; Must be accessible by ilu...\n");
        printf("    public static void %s(%s _ex, int index, xerox.ilu.IluCall _call)\n",
            methodNameOfWriteProc(e),
            shortname
            );
        printf("            throws xerox.ilu.IluSystemException {\n");
        printf("        int sz = 0;\n");
        printf("        if (_call.needsSizing()) {\n");
        printf("            sz = _call.beginSizingException(index);\n");
        if (e->type) {
            printf("            sz += %s;\n", ioSzPiece(e->type, "_ex.value"));
        }
        printf("        }\n");
        printf("        _call.startWriteException(index, sz);\n");
        if (e->type) {
            printf("        %s;\n", ioOutPiece(e->type, "_ex.value"));
        }
        printf("        _call.doneWriteException();\n");
        printf("    } //%s\n\n", methodNameOfWriteProc(e));
    }
    
    printf("    public static final xerox.ilu.IluExceptionRep _%s_getRep() {\n", \
           shortname);
    printf("        return _%s_rep;\n", shortname);
    printf("    }\n\n");
    
    printf("    static {\n");
    printf("        _%s_rep = xerox.ilu.IluExceptionRep.defineException(\n", 
        shortname
        );
    printf("            \"%s\", //java class name\n",
        dotCat(packageName(currentIfc), shortname)
        );
    printf("            \"%s\", //isl interface name\n",
        interface_name(currentIfc)
        );
    printf("            \"%s\"); //isl exception name\n", 
        exceptionShortName(e)
        );
    printf("    }\n\n");
    printf("} //%s\n", shortname);
}


/* end */



