/* genrecord.c */
/* Chris Jacobi, January 30, 1997 11:24 am PST */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* $Id: genrecord.c,v 1.16 1997/01/30 20:39:29 jacobi Exp $ */


#include <stdio.h>
#include <string.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "io.h"
#include "name.h"
#include "util.h"
#include "genstub.h"
#include "genrecord.h"


PUBLIC boolean rec_is_a(Type t) {
    t = myUrType(t);
    return (type_kind(t)==record_Type);
}


PUBLIC Type rec_assert(Type t) {
    t = myUrType(t);
    if (! rec_is_a(t)) fatal("should have been a record");
    return t;
}

PRIVATE char * recTypeShortName(Type t) {
    t = rec_assert(t);
    return easyShortTypeNameCleaned(t);
}


PRIVATE char * recTypeName(Type t) {
    t = rec_assert(t);
    return packageDotString(t->interface, recTypeShortName(t));
}


PUBLIC char * rec_typeDeclarator(Type t) {
    t = rec_assert(t);
    return recTypeName(t);
}


PUBLIC char * rec_ioSzPiece(Type t, const char *arg) {
    t = rec_assert(t);
    return cat4(recTypeName(t), "._szFunc(_call, ",  arg,  ")" );
}


PUBLIC char * rec_ioOutPiece(Type t, const char *arg) {
    t = rec_assert(t);
    return cat4(recTypeName(t), "._outFunc(_call, ",  arg,  ")" );
}


PUBLIC char * rec_ioInPiece(Type t) {
    t = rec_assert(t);
    return cat2(recTypeName(t), "._inFunc(_call)" );
}


PRIVATE void
printConstructorAssign1(Argument a, char* recName)
/* Initialize an instance var in the record class constructor.
 * Conceptually nested in printConstructorAssigns
 */
{
    char * name =  argumentName(a);
    printf("        %s.%s = %s;\n", recName, name, name);
}


PRIVATE void
printConstructorAssigns(Type t, char* recName)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    list_enumerate(fields, (EnumProc) printConstructorAssign1, recName);
}


PRIVATE void
printConstructorFormalArg1(Argument a, int * count)
/* Print the an argument to the record class constructor
 * Conceptually nested in printConstructorFormalArgs
 */
{
    if ( (*count)>0 ) printf(", ");
    (*count)++;
    printf("%s %s", typeDeclarator(a->type), argumentName(a));
}


PRIVATE void
printConstructorFormalArgs(Type t)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    int count = 0;
    list_enumerate(fields, (EnumProc) printConstructorFormalArg1, &count);
}

static boolean subclassable = 1;

PRIVATE void
printConstructors(Type t, char* rTypeName)
{
    if (subclassable) {
        printf("    //to force ILU to allocate subclasses\n");
        printf("    protected static java.lang.Class _theClass = null;\n\n");
    }
    
    printf("    //avoid unless corba compatibility required\n");
    printf("    public %s() {\n", rTypeName);
    printf("    } //%s\n\n", rTypeName);

    printf("    //not standard corba\n");
    printf("    public static %s alloc_%s() {\n", rTypeName, rTypeName);
    if (subclassable) {
        printf("        if (_theClass!=null) {\n");
        printf("            try {\n");
        printf("                return (%s) _theClass.newInstance();\n", rTypeName);
        printf("            } catch (Exception _e) {\n");
        printf("            }\n");
        printf("        }\n");
    }
    printf("        return new %s();\n", rTypeName);
    printf("    } //alloc_%s\n\n", rTypeName);
    
    printf("    //avoid unless corba compatibility required\n");
    printf("    public %s(", rTypeName);
    printConstructorFormalArgs(t);
    printf(") {\n");
    printConstructorAssigns(t, "this");
    printf("    } //%s\n\n", rTypeName);
    
    printf("    //not standard corba\n");
    printf("    public static %s alloc_%s(", rTypeName, rTypeName);
    printConstructorFormalArgs(t);
    printf(") {\n");
    printf("        %s _r = alloc_%s();\n", rTypeName, rTypeName);
    printConstructorAssigns(t, "_r");
    printf("        return _r;\n");
    printf("    } //alloc_%s\n\n", rTypeName);

}


PRIVATE void
printRecordFieldDecl1(Argument a, void *refany)
/* Print the declaration of a record member in the record class */
{
    printf("    public %s %s;\n", typeDeclarator(a->type), argumentName(a));
}


PRIVATE void
printRecordFieldDecls(Type t)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    list_enumerate(fields, (EnumProc) printRecordFieldDecl1, NULL);
    newline();
}


PRIVATE void
printSz1RecField(Argument a, void *refany)
{
    char * szPiece;
    char * argName;
    argName = cat2("_r.", argumentName(a));
    szPiece = ioSzPiece(a->type, argName);
    printf("        _sz = _sz + %s;\n", szPiece);
}


PRIVATE void
printSzRecords(Type t, char * rTypeName)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    printf("    //internal to ILU only\n");
    printf("    public static int _szFunc(xerox.ilu.IluCall _call, %s _r)\n", 
                rTypeName);
    printf("                throws xerox.ilu.IluSystemException {\n");
    printf("        int _sz = _call.szRecord();\n");
    list_enumerate(fields, (EnumProc) printSz1RecField, NULL);
    printf("        _call.endRecord();\n");
    printf("        return _sz;\n");
    printf("    } //_szFunc\n\n");
}


PRIVATE void
printOut1RecField(Argument a, void *refany)
{
    char * outPiece;
    char * argName;
    argName = cat2("_r.", argumentName(a));
    outPiece = ioOutPiece(a->type, argName);
    printf("        %s;\n", outPiece);
}


PRIVATE void
printOutRecords(Type t, char * rTypeName)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    printf("    //internal to ILU only\n");
    printf("    public static void _outFunc(xerox.ilu.IluCall _call, %s _r)\n", 
                rTypeName);
    printf("                throws xerox.ilu.IluSystemException {\n");
    printf("        _call.outRecord();\n");
    list_enumerate(fields, (EnumProc) printOut1RecField, NULL);
    printf("        _call.endRecord();\n");
    printf("    } //_outFunc\n\n");
}


PRIVATE void
printIn1RecField(Argument a, void *refany)
{
    char * recName = (char *) refany;
    char * inPiece;
    char * argName;
    argName = argumentName(a);
    inPiece = ioInPiece(a->type);
    printf("        %s.%s = %s;\n", recName, argName, inPiece);
}


PRIVATE void
printVARInRecords(Type t, char * rTypeName)
/* records are not subtyped; 
 * thats why a non static procedure is permitted 
 */
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    printf("    //internal to ILU only\n");
    printf("    void _inVar(xerox.ilu.IluCall _call)\n",
                    rTypeName);
    printf("                throws xerox.ilu.IluSystemException {\n");
    printf("        _call.inRecord();\n");
    list_enumerate(fields, (EnumProc) printIn1RecField, "this");
    printf("        _call.endRecord();\n");
    printf("    } //_inVar\n\n");
}


PRIVATE void
printInRecords(Type t, char * rTypeName)
{
    TypeDescription d = type_description(t);
    list fields = d->structuredDes.record;
    printf("    //internal to ILU only\n");
    printf("    public static %s _inFunc(xerox.ilu.IluCall _call)\n", rTypeName);
    printf("                throws xerox.ilu.IluSystemException {\n");
    printf("        %s _r = alloc_%s();\n", rTypeName, rTypeName);
    printf("        _r._inVar(_call);\n");
    printf("        return _r;\n");
    printf("    } //_inFunc\n\n");
}


PUBLIC void
rec_defineMain(Type t)
{
    char* rTypeName;
    t = myUrType(t);
    rTypeName = recTypeShortName(t);
    NewJavaFile(t->interface, rTypeName);
    printf("public class %s {\n", rTypeName);
    printRecordFieldDecls(t);
    printConstructors(t, rTypeName);
    printSzRecords(t, rTypeName);
    printOutRecords(t, rTypeName);
    printVARInRecords(t, rTypeName); /*like var argument*/
    printInRecords(t, rTypeName); /*the real thing*/
    printf("} //%s\n", rTypeName);
}

/* end */
