/* name.h */
/* Chris Jacobi, January 30, 1997 11:31 am PST */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */
 
/* $Id: name.h,v 1.33 1997/01/30 20:38:22 jacobi Exp $ */


/* never get confused
 * 4 cases for name strings occur frequently
 * a) Syntactactily and semanticly in isl(idl) space 
 *    E.g. Used in strings passed to the runtime
 * b) Syntactactily and semanticly in Java space with
 *    E.g. Names of reference interfaces, packages and methods
 * c) Syntactactily in Java and semanticly in isl space
 *    E.g. Type cases for unions.
 * d) Syntactactily in Java and semanticly unfinished but prepared for java
 *    E.g. Suffixed classes before the suffix is added
 *
 * E.g  package prefix and suffix is a semantic addition in java space
 * E.g  Keyword avoidance is a syntactic feature for java space
 */


/*
 * contains code that sets and gets all types of names (type, exceptions etc.).
 */

extern char* packageName(Interface i);
    /* name of java package for isl interface or idl module */
    
extern char* packageNameOptional(Interface i); 
    /* for all interfaces: package name or null string if local */
    
extern char* packageDotString(Interface i, char* name); 
    /* prefixex the string with the package name, if not local */


extern char* easyShortTypeNameCleaned(Type t);
/* Cleaned (syntax of a java id) original (short) 
 * name. Implemented for a subset of all types only.
 * Whether this has a semantic meaming in java for a type or not is only 
 * known by the implementation of that type.
 * Use when generating a type (known in right subset); don't use 
 * on components because suffixes etc are not added at this point.
 * This is NOT urtyped...
 * The name of the type is NOT the typename for instances !!!!!
 *
 */

extern char* typeNameUnresolvedButClean(Type t);
/* Semanticly in isl(idl) space, not java. However syntacticly a java identifier
 * Aliases are not resolved; forbidden identifiers not resolved
 * Used for ilu isl extensions of unions
 */

extern char* classNameCleaning(Interface i, char* mediumName);
/* Takes a mediumName (syntacticly correct java) and mangles
 * it if its name would be bad in the interface i.
 * Apply this after easy names...
 */


/* Declarators are context dependent; usefull for type declarations
 * new statements etc.  Syntactically maybe not identifiers; e.g. arrays.
 */ 
extern char* typeDeclarator(Type t);
extern char* holderTypeDeclarator(Type t); 

/* The following procedures return syntacticly javaized names */
extern char* procShortName(Procedure p);
extern char* argumentName(Argument a);
extern char* exceptionShortName(Exception e);
extern char* exceptionName(Exception e);
extern char* constantName(Constant c);
extern char* enumFieldName(EnumField e);

extern char* getInterfaceDirectory(Interface i);

extern void initName(); /*initialize the file name.c*/

/* end */
