/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

/*
$Id: type-basics.c,v 1.27 1997/01/29 01:52:40 janssen Exp $
*/

#include "lisp.h"

extern void GenerateClassCode (Type, Context);
extern void GenerateClassDefinition (Type, Context);

void OutputTypeName (Type type, Context context)
{
  enum PrimitiveTypes t = type_ur_kind(type);

  if (t == byte_Type) 
    fprintf (context->file, " (:primitive :byte)");
  else if (t == shortinteger_Type) 
    fprintf (context->file, " (:primitive :short-integer)");
  else if (t == longinteger_Type) 
    fprintf (context->file, " (:primitive :long-integer)");
  else if (t == shortcharacter_Type) 
    fprintf (context->file, " (:primitive :short-character)");
  else if (t == character_Type) 
    fprintf (context->file, " (:primitive :character)");
  else if (t == integer_Type) 
    fprintf (context->file, " (:primitive :integer)");
  else if (t == shortcardinal_Type) 
    fprintf (context->file, " (:primitive :short-cardinal)");
  else if (t == longcardinal_Type) 
    fprintf (context->file, " (:primitive :long-cardinal)");
  else if (t == cardinal_Type) 
    fprintf (context->file, " (:primitive :cardinal)");
  else if (t == boolean_Type) 
    fprintf (context->file, " (:primitive :boolean)");
  else if (t == real_Type) 
    fprintf (context->file, " (:primitive :real)");
  else if (t == shortreal_Type)
    fprintf (context->file, " (:primitive :short-real)");
  else if (t == longreal_Type)
    fprintf (context->file, " (:primitive :long-real)");
  else
    fprintf (context->file, " (%s %s)",
	     (t == object_Type) ? ":object" : ":constructed",
	     lisp_type_name(ur_type(type)));
}

typedef struct unioncontext_s {
  Context c;
  Type t;
  Argument default_arm;
} *UnionContext;

static void OutputUnionArmValue (ConstantValue val, Context context)
{
  switch (val->type)
    {
    case integer_Type:
    case shortinteger_Type:
    case cardinal_Type:
    case shortcardinal_Type:
    case byte_Type:
      fprintf (context->file, "%s%ld",
	       (val->val.i.sign < 0) ? "-" : "", val->val.i.value);
      break;
    case shortcharacter_Type:
      fprintf (context->file, ":%s", val->val.s);
      break;
    case boolean_Type:
      fprintf (context->file, "%s", val->val.b ? "cl:t" : "cl:nil");
      break;
    default:
      fprintf (stderr, "illegal discriminator value\n");
      return;
    }
}

static void OutputUnionArm (Argument arm, UnionContext s)
{
  if (arm != s->default_arm)
    {
      fprintf (s->c->file, " ((");
      list_enumerate (arm->values,
		      (iluparser_EnumProc) OutputUnionArmValue, s->c);
      fprintf (s->c->file, ")");
      OutputTypeName (arm->type, s->c);
      fprintf (s->c->file, ")");
    }
}

static void OutputEnumerationField (EnumField field, Context context)
{
  char buf[100];

  fprintf (context->file, " (%s %s)\n", lisp_string(field->name),
	   (field->id < 0) ? "-1" : (sprintf(buf, "%u", field->id), buf));
}

static void OutputFieldDescription (Argument field, Context context)
{
  fprintf (context->file, " (%s", lisp_argument_name(field));
  OutputTypeName (field->type, context);
  fprintf (context->file, ")\n");
}

static void OutputDimension (long int d, Context context)
{
  fprintf (context->file, " %lu", d);
}

static void GenerateTypeStubs (Type type, Context context)
{
  enum PrimitiveTypes t = type_kind(type);
  TypeDescription d = type_description(type);

  if (type->builtIn || t == invalid_Type || t == void_Type || type->importInterfaceName != NULL)
    return;

  switch (t)
    {
    case byte_Type:
    case shortinteger_Type:
    case shortcharacter_Type:
    case longinteger_Type:
    case character_Type:
    case integer_Type:
    case shortcardinal_Type:
    case longcardinal_Type:
    case cardinal_Type:
    case boolean_Type:
    case real_Type:
    case shortreal_Type:
    case longreal_Type:
      /* no methods defined on these types, no output needed here. */
      break;

    case alias_Type:
      {
	fprintf (context->file, "(ilu:define-alias %s \"%s\"",
		 lisp_type_name(type), type_uid(type));
	OutputTypeName (ur_type(type), context);
	fprintf (context->file, ")\n");
      }
      break;

    case union_Type:
      {
	struct unioncontext_s s;

	fprintf (context->file, "(ilu:define-union %s \"%s\"",
		 lisp_type_name(type), type_uid(type));
	OutputTypeName (d->structuredDes.uniond.discriminator_type, context);
	fprintf (context->file, " %s",
		 d->structuredDes.uniond.others_allowed ? "t" : "nil");
	s.default_arm = d->structuredDes.uniond.default_arm;
	if (s.default_arm != NULL)
	  OutputTypeName (s.default_arm->type, context);
	else
	  fprintf (context->file, " nil");
	s.t = type;
	s.c = context;
	list_enumerate (d->structuredDes.uniond.types, (EnumProc) OutputUnionArm, &s);
	fprintf (context->file, ")\n");
      }
      break;

    case sequence_Type:
      fprintf (context->file, "(ilu:define-sequence %s \"%s\"", lisp_type_name(type), type_uid(type));
      OutputTypeName (d->structuredDes.sequence.type, context);
      fprintf (context->file, " %lu)\n", d->structuredDes.sequence.limit);
      break;

    case optional_Type:
      fprintf (context->file, "(ilu:define-optional %s \"%s\"", lisp_type_name(type), type_uid(type));
      OutputTypeName (d->structuredDes.optional, context);
      fprintf (context->file, ")\n");
      break;

    case record_Type:
      fprintf (context->file, "(ilu:define-record %s \"%s\"\n", lisp_type_name(type), type_uid(type));
      list_enumerate (d->structuredDes.record, (EnumProc) OutputFieldDescription, context);
      fprintf (context->file, " )\n");
      break;

    case array_Type:
      fprintf (context->file, "(ilu:define-array %s \"%s\"", lisp_type_name(type), type_uid(type));
      OutputTypeName (d->structuredDes.array.type, context);
      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) OutputDimension, context);
      fprintf (context->file, ")\n");
      break;

    case enumeration_Type:
      fprintf (context->file, "(ilu:define-enumeration %s \"%s\"\n", lisp_type_name(type), type_uid(type));
      list_enumerate (d->structuredDes.enumeration, (EnumProc) OutputEnumerationField, context);
      fprintf (context->file, " )\n");
      break;      

    case object_Type:
      GenerateClassCode (type, context);
      break;

    default:
      break;
    };
  fprintf (context->file, "\n");
}

static void GenerateTypeDefinition (Type type, Context context)
{
  enum PrimitiveTypes t = type_ur_kind(type);
  TypeDescription d = type_description(type);

  if (type->builtIn || t == invalid_Type || t == void_Type || type->importInterfaceName != NULL)
    return;

  switch (t)
    {
    case byte_Type:
    case shortinteger_Type:
    case shortcharacter_Type:
    case longinteger_Type:
    case character_Type:
    case integer_Type:
    case shortcardinal_Type:
    case longcardinal_Type:
    case cardinal_Type:
    case boolean_Type:
    case real_Type:
    case shortreal_Type:
    case longreal_Type:
      fprintf (context->file, "(ilu:define-primitive-type %s ", lisp_type_name(type));
      OutputTypeName (type, context);
      fprintf (context->file, ")\n");
      break;

    case alias_Type:
      fprintf (context->file, "(ilu:define-alias-type %s ", lisp_type_name(type));
      OutputTypeName (ur_type(type), context);
      break;

    case union_Type:
      {
	struct unioncontext_s s;

	fprintf (context->file, "(ilu:define-union-type %s",
		 lisp_type_name(type));
	OutputTypeName (d->structuredDes.uniond.discriminator_type, context);
	fprintf (context->file, " %s",
		 d->structuredDes.uniond.others_allowed ? "t" : "nil");
	s.default_arm = d->structuredDes.uniond.default_arm;
	if (s.default_arm != NULL)
	  OutputTypeName (s.default_arm->type, context);
	else
	  fprintf (context->file, " nil");
	s.t = type;
	s.c = context;
	list_enumerate (d->structuredDes.uniond.types, (EnumProc) OutputUnionArm, &s);
	fprintf (context->file, ")\n");
      }
      break;

    case sequence_Type:
      fprintf (context->file, "(ilu:define-sequence-type %s", lisp_type_name(type));
      OutputTypeName (d->structuredDes.sequence.type, context);
      fprintf (context->file, " %lu)\n", d->structuredDes.sequence.limit);
      break;

    case optional_Type:
      fprintf (context->file, "(ilu:define-optional-type %s ", lisp_type_name(type));
      OutputTypeName (d->structuredDes.optional, context);
      fprintf (context->file, ")\n");
      break;

    case record_Type:
      fprintf (context->file, "(ilu:define-record-type %s\n", lisp_type_name(type));
      list_enumerate (d->structuredDes.record, (EnumProc) OutputFieldDescription, context);
      fprintf (context->file, " )\n");
      break;

    case array_Type:
      fprintf (context->file, "(ilu:define-array-type %s", lisp_type_name(type));
      OutputTypeName (d->structuredDes.array.type, context);
      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) OutputDimension, context);
      fprintf (context->file, ")\n");
      break;

    case enumeration_Type:
      fprintf (context->file, "(ilu:define-enumeration-type %s\n", lisp_type_name(type));
      list_enumerate (d->structuredDes.enumeration, (EnumProc) OutputEnumerationField, context);
      fprintf (context->file, " )\n");
      break;      

    case object_Type:
      GenerateClassDefinition (type, context);
      break;      

    default:
      break;
    };
  fprintf (context->file, "\n");
}

static void GenerateExceptionDef (Exception e, Context context)
{
  if (e->builtIn || e->importInterfaceName != NULL)
    return;

  if (e->corba_rep_id != NULL)
    fprintf (context->file, "(ilu:define-exception-type %s \"%s\"",
	     lisp_exception_name(e), e->corba_rep_id);
  else
    fprintf (context->file, "(ilu:define-exception-type %s \"ilu:%s.%s\"",
	     lisp_exception_name(e), interface_name(e->interface), exception_name(e));
  if (e->type == NULL)
    fprintf (context->file, " cl:nil");
  else
    OutputTypeName (e->type, context);
  if (e->doc_string == NULL)
    fprintf (context->file, " nil)\n\n");
  else
    fprintf (context->file, " \"%s\")\n\n", e->doc_string);
}

static void GenerateExceptionCode2 (Exception e, Context context)
{
  if (e->builtIn || e->importInterfaceName != NULL)
    return;

  fprintf (context->file, "(ilu:define-exception %s", lisp_exception_name(e));
  if (e->type == NULL)
    fprintf (context->file, " cl:nil");
  else
    OutputTypeName (e->type, context);
  fprintf (context->file, ")\n\n");
}

void GenerateTypeDefinitions (Interface interface, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = interface;

  list_enumerate (interface->types, (EnumProc) GenerateTypeDefinition, &context);
}

void GenerateTypeCode (Interface interface, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = interface;

  list_enumerate (interface->types, (EnumProc) GenerateTypeStubs, &context);
}

void GenerateExceptionDefinitions (Interface interface, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = interface;

  list_enumerate (interface->exceptions, (EnumProc) GenerateExceptionDef, &context);
}

void GenerateExceptionCode (Interface interface, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = interface;

  list_enumerate (interface->exceptions, (EnumProc) GenerateExceptionCode2, &context);
}

static void convertStringForm (char *buf, string str)
{
  register char *p;

  strcpy (buf, "#.(ilu::ascii-to-string #(");
  for (p = str;  *p != 0;  p++)
    sprintf (buf + strlen(buf), " #x%x", (unsigned char) *p);
  strcpy (buf + strlen(buf), "))");
}

static void GenerateConstant (Constant c, Context context)
{
  enum PrimitiveTypes t = type_ur_kind (c->type);

  if (c->interface != context->interface)
    return;

  fprintf (context->file, "(commonlisp:defconstant %s ", lisp_constant_name(c));

  if ((t == cardinal_Type || t == shortcardinal_Type || t == byte_Type)
      && (c->value->type == integer_Type))
    fprintf (context->file, "%lu)\n", c->value->val.i.value);
  else if ((t == integer_Type || t == shortinteger_Type)
	   && (c->value->type == integer_Type))
    fprintf (context->file, "%s%lu)\n",
	     (c->value->val.i.sign < 0) ? "-" : "", c->value->val.i.value);
  else if ((t == real_Type || t == shortreal_Type)
	   && (c->value->type == real_Type))
    fprintf (context->file, "%s%s.%s%s%ld)\n",
	     (c->value->val.r.sign < 0) ? "-" : "",
	     c->value->val.r.value,
	     (c->value->val.r.fraction == NULL) ? "0" : c->value->val.r.fraction,
	     (t == real_Type) ? "D" : "F",
	     c->value->val.r.exponent);
  else if (t == shortcharacter_Type && (c->value->type == integer_Type))
    fprintf (context->file, "#.(commonlisp:code-char #x%lx))\n",
	     c->value->val.i.value);
  else if (t == sequence_Type AND
	   type_ur_kind(type_description(c->type)->structuredDes.sequence.type)
	     == shortcharacter_Type AND
	   c->value->type == shortcharacter_Type)
    {
      char *buf;

      buf = iluparser_Malloc(100 + 4 * strlen(c->value->val.s));
      convertStringForm (buf, c->value->val.s);
      fprintf (context->file, "%s)\n", buf);
      iluparser_Free(buf);
    }
  else if (t == boolean_Type AND c->value->type == boolean_Type)
    {
      fprintf (context->file, "%s)\n", c->value->val.b ? "t" : "nil");
    }
  else
    {
      fprintf (stderr,
	       "Invalid constant, %s, encountered.  Type is %s, type of value is %d.\n",
	       name_base_name (c->name), lisp_type_name(c->type),
	       c->value->type);
    }
}

void GenerateConstants (Interface interface, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = interface;

  list_enumerate (interface->constants, (EnumProc) GenerateConstant, &context);
}
