//  -*- Mode: C++; -*-

// Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
//
// Unlimited use, reproduction, and distribution of this software is
// permitted.  Any copy of this software must include both the above
// copyright notice of Siemens Corporation and this paragraph.  Any
// distribution of this software must comply with all applicable United
// States export control laws.  This software is made available AS IS,
// and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
// PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
// THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
// CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
// SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// $Author: mlg $
// $Date: 1994/12/07 20:08:52 $
// $Source: /home/yugo/pat/repository/prt/src/utils/template.c,v $
// $Revision: 2.1.1.1 $
//

#include <assert.h>
#include <stdio.h>
#include <stream.h>
#include <string.h>
#include "ilu.hh"
#include "ilu-scm.h"
#include "Hook.hh"

#define HOOK_SERVER

#ifdef HOOK_SERVER

class HookImpl : public Hook_T_Hook {
 public:
  HookImpl(char *instanceHandle, iluServer *server);
  virtual ~HookImpl();
  virtual ilu_Boolean Invoke(HookStatus* status, ilu_T_CString data);
  virtual char * ILUGetInstanceHandle();
  virtual iluServer * ILUGetServer();
 private:
  char* _instance_handle;
  iluServer* _server;
};

// create new print hook
HookImpl::HookImpl(char *instance_handle, iluServer *server)
: _instance_handle(strdup(instance_handle)), _server(server)
{
}

char* HookImpl::ILUGetInstanceHandle()
{
  return _instance_handle;
}

iluServer* HookImpl::ILUGetServer()
{
  return _server;
}

// destroy a print hook (and it's buffer)
HookImpl::~HookImpl()
{
  free(_instance_handle);
}

ilu_Boolean
HookImpl::Invoke(HookStatus* status, ilu_T_CString data)
{
  if(!strcmp("dummy", data)) {
    cout << "C++ got string: " << data << endl;
    Hook_T_dummy* ex = new Hook_T_dummy;
    ex->stuff = strdup("dummy string");
    ex->more_stuff = 666;
    Hook_G::RaiseException(status, Hook_E_dummy, ex);
    return ilu_FALSE;
  } else {
    status->returnCode = HookReply_Success;
    cout << "C++ got string: " << data << endl;
    return ilu_TRUE;
  }
}
#else
typedef Hook_T_Hook Hook;
#endif

void test_hooks(void*)
{
  HookStatus status;
  Hook_T_Hook* hook1 = (Hook_T_Hook*)iluObject::Lookup("C++Server", "C++Hook", Hook_T_Hook::ILUClassRecord);
  Hook_T_Hook* hook2 = (Hook_T_Hook*)iluObject::Lookup("ScmServer", "ScmHook", Hook_T_Hook::ILUClassRecord);

  cout << "test_hooks()" << endl;

  cout << "using C++ Hook" << endl;
  hook1->Invoke(&status, "stuff1");

  cout << "using Scm Hook" << endl;
  hook2->Invoke(&status, "stuff2");
}

int main(int argc, char* argv[])
{
#ifdef HOOK_SERVER
  ilu_scheme_main(argc, argv, 0, "client.scm");

  iluServer server("C++Server", 0);
  server.AddPort(0, 0, ilu_TRUE);
  ilu::SetDefaultServer(&server);

  HookImpl* hook = new HookImpl("C++Hook", &server);
  hook->ILUPublish();

  void* alarm = ilu_CreateAlarm();
  ilu_FineTime delta = ilu_FineTime_FromDouble(5.0);
  ilu_FineTime alarm_time = ilu_FineTime_Add(ilu_FineTime_Now(), delta);
  ilu_SetAlarm(alarm, alarm_time, test_hooks, 0);

  server.Run();
#else
  HookStatus s;
  Hook_T_Hook* hook = 0;
  int num = argc == 1 ? 1 : atoi(argv[1]);

  hook = (Hook_T_Hook*)iluObject::Lookup("Server", "TheHook", Hook_T_Hook::ILUClassRecord);
  assert(hook);
  
  for(int i = 0; i < num; i++) {
    hook->Invoke(&s, "some string");
    hook->Invoke(&s, "dummy");
  }
#endif
  return 0;
}
